//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/config.rs: Static configuration, edit & recompile!
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

/// Path to /etc.
pub const PATH_ETC: &str = "/etc";

/// The default shell to execute.
/// Change this if your system doesn't have bash,
/// or set the environment variable SYD_SHELL.
pub const SYD_SH: &str = "/bin/bash";

/// The environment variable to read the default shell from
pub const ENV_SH: &str = "SYD_SH";
/// The environment variable to read the log level from.
pub const ENV_LOG: &str = "SYD_LOG";

/// Setting this environment variable is equivalent to the --unshare-mount CLI option.
pub const ENV_UNSHARE_MOUNT: &str = "SYD_UNSHARE_MOUNT";
/// Setting this environment variable is equivalent to the --unshare-uts CLI option.
pub const ENV_UNSHARE_UTS: &str = "SYD_UNSHARE_UTS";
/// Setting this environment variable is equivalent to the --unshare-ipc CLI option.
pub const ENV_UNSHARE_IPC: &str = "SYD_UNSHARE_IPC";
/// Setting this environment variable is equivalent to the --unshare-user CLI option.
pub const ENV_UNSHARE_USER: &str = "SYD_UNSHARE_USER";
/// Setting this environment variable is equivalent to the --unshare-pid CLI option.
pub const ENV_UNSHARE_PID: &str = "SYD_UNSHARE_PID";
/// Setting this environment variable is equivalent to the --unshare-net CLI option.
pub const ENV_UNSHARE_NET: &str = "SYD_UNSHARE_NET";
/// Setting this environment variable is equivalent to the --unshare-cgroup CLI option.
pub const ENV_UNSHARE_CGROUP: &str = "SYD_UNSHARE_CGROUP";

/// Setting this environment variable is equivalent to the --root CLI option.
pub const ENV_FAKEROOT: &str = "SYD_FAKEROOT";
/// Setting this environment variable defines the number of simultaneous threads.
pub const ENV_NPROC: &str = "SYD_NPROC";

/// The environment variable to set to disable logging to syslog.
pub const ENV_NO_SYSLOG: &str = "SYD_NO_SYSLOG";

/// The environment variable to set to disable cross memory attach, and fallback to /proc/pid/mem.
pub const ENV_NO_CROSS_MEMORY_ATTACH: &str = "SYD_NO_CROSS_MEMORY_ATTACH";

/// Path prefix for magic stat commands.
pub const MAGIC_PREFIX: &str = "/dev/syd";

/// Path to the file which will be used as file status for magic stat commands.
pub const MAGIC_FILE: &str = "/dev/null";

/// Api version of the configuration.
pub const API_VERSION: &str = env!("CARGO_PKG_VERSION_MAJOR");

/// Cache size of the bind0 `HashSet`.
pub const BIND0_CACHE_SIZE: usize = 64;

/// Default buffer size used in `compat::get_directory_entries` function.
pub const GETDENTS_BUFFER_SIZE: usize = 8192;

// Sandbox Restrictions

macro_rules! ioctl_entry {
    ($name:ident) => {
        (nix::libc::$name as u64, stringify!($name))
    };
}

type KeyValue<'a> = (u64, &'a str);

/// Size of the allowlisted ioctl requests list.
pub const ALLOWLIST_IOCTL_SIZE: usize = 58;

/// List of allowlisted ioctl requests.
pub const ALLOWLIST_IOCTL: [KeyValue; ALLOWLIST_IOCTL_SIZE] = [
    // Generic ioctls
    ioctl_entry!(FIOCLEX),
    ioctl_entry!(FIONCLEX),
    ioctl_entry!(FIONREAD),
    ioctl_entry!(FIONBIO),
    ioctl_entry!(FIOASYNC),
    // Allow some font ioctls
    (0x4B66, "GIO_UNIMAP"),
    // Get and set terminal attributes
    ioctl_entry!(TCGETS),
    ioctl_entry!(TCSETS),
    ioctl_entry!(TCSETSW),
    ioctl_entry!(TCSETSF),
    ioctl_entry!(TCGETA),
    ioctl_entry!(TCSETA),
    ioctl_entry!(TCSETAW),
    ioctl_entry!(TCSETAF),
    // Locking the termios structure
    ioctl_entry!(TIOCGLCKTRMIOS),
    ioctl_entry!(TIOCSLCKTRMIOS),
    // Get and set window size
    ioctl_entry!(TIOCGWINSZ),
    ioctl_entry!(TIOCSWINSZ),
    // Sending a break
    ioctl_entry!(TCSBRK),
    ioctl_entry!(TCSBRKP),
    ioctl_entry!(TIOCSBRK),
    ioctl_entry!(TIOCCBRK),
    // Software flow control
    ioctl_entry!(TCXONC),
    // Buffer count and flushing
    ioctl_entry!(TIOCINQ),
    ioctl_entry!(TIOCOUTQ),
    ioctl_entry!(TCFLSH),
    // Controlling terminal.
    // SAFETY: These two ioctls are known-unsafe.
    // We disallow them.
    //ioctl_entry!(TIOCSTI),
    //ioctl_entry!(TIOCCONS),

    // Pty handling.
    ioctl_entry!(TIOCGPTN),
    ioctl_entry!(TIOCSPTLCK),
    ioctl_entry!(TIOCGDEV),
    ioctl_entry!(TCGETX),
    ioctl_entry!(TCSETX),
    ioctl_entry!(TCSETXF),
    ioctl_entry!(TCSETXW),
    ioctl_entry!(TIOCSIG),
    ioctl_entry!(TIOCVHANGUP),
    ioctl_entry!(TIOCGPKT),
    ioctl_entry!(TIOCGPTLCK),
    ioctl_entry!(TIOCGEXCL),
    ioctl_entry!(TIOCGPTPEER),
    // Process group and session ID
    ioctl_entry!(TIOCGPGRP),
    ioctl_entry!(TIOCSPGRP),
    ioctl_entry!(TIOCGSID),
    // Exclusive mode
    ioctl_entry!(TIOCEXCL),
    ioctl_entry!(TIOCGEXCL),
    ioctl_entry!(TIOCNXCL),
    // Get/Set line discipline
    ioctl_entry!(TIOCGETD),
    ioctl_entry!(TIOCSETD),
    // Pseudoterminal ioctls
    ioctl_entry!(TIOCPKT),
    ioctl_entry!(TIOCGPKT),
    ioctl_entry!(TIOCSPTLCK),
    ioctl_entry!(TIOCGPTLCK),
    ioctl_entry!(TIOCGPTPEER),
    // SAFETY: Skio modem control
    // ioctls: TIOCMGET, TIOCMSET
    // et al.

    // Marking a line as local
    ioctl_entry!(TIOCGSOFTCAR),
    ioctl_entry!(TIOCSSOFTCAR),
    // Allow seccomp ioctls
    (0x40082102, "SECCOMP_IOCTL_NOTIF_ID_VALID"),
    (0xc0502100, "SECCOMP_IOCTL_NOTIF_RECV"),
    (0xc0182101, "SECCOMP_IOCTL_NOTIF_SEND"),
    (0x40182103, "SECCOMP_IOCTL_NOTIF_ADDFD"),
];

/// Size of the allowlisted prctl requests list.
pub const ALLOWLIST_PRCTL_SIZE: usize = 51;

/// List of allowlisted ioctl requests.
pub const ALLOWLIST_PRCTL: [KeyValue; ALLOWLIST_PRCTL_SIZE] = [
    (1, "PR_SET_PDEATHSIG"),
    (2, "PR_GET_PDEATHSIG"),
    (3, "PR_GET_DUMPABLE"),
    (4, "PR_SET_DUMPABLE"),
    (5, "PR_GET_ALIGN"),
    (6, "PR_GET_SETALIGN"),
    (7, "PR_GET_KEEPCAPS"),
    (8, "PR_SET_KEEPCAPS"),
    (9, "PR_GET_FPEMU"),
    (10, "PR_SET_FPEMU"),
    (11, "PR_GET_FPEXC"),
    (12, "PR_SET_FPEXC"),
    (13, "PR_GET_TIMING"),
    (14, "PR_SET_TIMING"),
    (15, "PR_SET_NAME"),
    (16, "PR_GET_NAME"),
    (19, "PR_GET_ENDIAN"),
    (20, "PR_SET_ENDIAN"),
    (21, "PR_GET_SECCOMP"),
    (22, "PR_SET_SECCOMP"),
    (23, "PR_CAPBSET_READ"),
    (24, "PR_CAPBSET_DROP"),
    (25, "PR_GET_TSC"),
    (26, "PR_SET_TSC"),
    (27, "PR_GET_SECUREBITS"),
    (28, "PR_SET_SECUREBITS"),
    (29, "PR_SET_TIMERSLACK"),
    (30, "PR_GET_TIMERSLACK"),
    // SAFETY: no perf! PR_TASK_PERF_EVENTS_DISABLE
    // SAFETY: no perf! PR_TASK_PERF_EVENTS_ENABLE
    (33, "PR_MCE_KILL"),
    (34, "PR_MCE_KILL_GET"),
    // SAFETY: no self modifying executables! PR_SET_MM
    // SAFETY: no ptrace! PR_SET_PTRACER
    (36, "PR_SET_CHILD_SUBREAPER"),
    (37, "PR_GET_CHILD_SUBREAPER"),
    (38, "PR_SET_NO_NEW_PRIVS"),
    (39, "PR_GET_NO_NEW_PRIVS"),
    (40, "PR_GET_TID_ADDRESS"),
    (41, "PR_SET_THP_DISABLE"),
    (42, "PR_GET_THP_DISABLE"),
    // SAFETY: unimplemented! PR_MPX_ENABLE_MANAGEMENT
    // SAFETY: unimplemented! PR_MPX_DISABLE_MANAGEMENT
    (45, "PR_SET_FP_MODE"),
    (46, "PR_GET_FP_MODE"),
    (47, "PR_CAP_AMBIENT"),
    (50, "PR_SVE_SET_VL"),
    (51, "PR_SVE_GET_VL"),
    // SAFETY: do not access speculation misfeature! PR_GET_SPECULATION_CTRL
    // SAFETY: do not change speculation misfeature! PR_SET_SPECULATION_CTRL
    (54, "PR_PAC_RESET_KEYS"),
    (55, "PR_SET_TAGGED_ADDR_CTRL"),
    (56, "PR_GET_TAGGED_ADDR_CTRL"),
    (57, "PR_SET_IO_FLUSHER"),
    (58, "PR_GET_IO_FLUSHER"),
    (59, "PR_SET_SYSCALL_USER_DISPATCH"),
    (60, "PR_PAC_SET_ENABLED_KEYS"),
    (61, "PR_PAC_GET_ENABLED_KEYS"),
    (62, "PR_SCHED_CORE"),
];

// Sandboxing profiles

/// No ipv4 sandboxing profile.
pub const PROFILE_NOIPV4: [&str; PROFILE_NOIPV4_SIZE] =
    ["sandbox/net:on", "allowlist/net/connect+::/0@0-65535"];
/// Size of the no ipv4 sandboxing profile.
pub const PROFILE_NOIPV4_SIZE: usize = 2;

/// No ipv6 sandboxing profile.
pub const PROFILE_NOIPV6: [&str; PROFILE_NOIPV6_SIZE] =
    ["sandbox/net:on", "allowlist/net/connect+0.0.0.0/0@0-65535"];
/// Size of the no ipv6 sandboxing profile.
pub const PROFILE_NOIPV6_SIZE: usize = 2;

/// Size of the paludis sandboxing profile.
pub const PROFILE_PALUDIS_SIZE: usize = 205;
/// Paludis sandboxing profile, activated with:
/// `syd --profile paludis`.
pub const PROFILE_PALUDIS: [&str; PROFILE_PALUDIS_SIZE] = [
    "sandbox/read:on",
    "sandbox/stat:on",
    "sandbox/write:on",
    "sandbox/exec:off",
    "sandbox/net:on",
    "trace/allow_successful_bind:true",
    "trace/allow_unsupported_socket_families:true",
    "trace/allow_unsafe_chmod:true",
    "trace/allow_unsafe_getrandom:true",
    "allowlist/read+/***",
    "denylist/read+/dev/*/***",
    "denylist/read+/etc/crypttab",
    "denylist/read+/etc/gshadow*",
    "denylist/read+/etc/shadow*",
    "denylist/read+/proc/*/***",
    "denylist/read+/sys/*/***",
    "denylist/read+/var/log/***",
    "allowlist/read+/dev/console",
    "allowlist/read+/dev/fd/**",
    "allowlist/read+/dev/full",
    "allowlist/read+/dev/null",
    "allowlist/read+/dev/ptmx",
    "allowlist/read+/dev/pts/**",
    "allowlist/read+/dev/pty*",
    "allowlist/read+/dev/shm/***",
    "allowlist/read+/dev/stderr",
    "allowlist/read+/dev/stdout",
    "allowlist/read+/dev/tts",
    "allowlist/read+/dev/tty*",
    "allowlist/read+/dev/urandom",
    "allowlist/read+/dev/zero",
    "allowlist/read+/proc/[0-9]*",
    "denylist/read+/proc/[0-9]*/*",
    "allowlist/read+/proc/[0-9]*/[ug]id_map",
    "allowlist/read+/proc/[0-9]*/cgroup",
    "allowlist/read+/proc/[0-9]*/clear_refs",
    "allowlist/read+/proc/[0-9]*/cmdline",
    "allowlist/read+/proc/[0-9]*/comm",
    "allowlist/read+/proc/[0-9]*/coredump_filter",
    "allowlist/read+/proc/[0-9]*/coredump_filter",
    "allowlist/read+/proc/[0-9]*/cpuset",
    "allowlist/read+/proc/[0-9]*/cwd",
    "allowlist/read+/proc/[0-9]*/environ",
    "allowlist/read+/proc/[0-9]*/exe",
    "allowlist/read+/proc/[0-9]*/fd/***",
    "allowlist/read+/proc/[0-9]*/fdinfo/***",
    "allowlist/read+/proc/[0-9]*/io",
    "allowlist/read+/proc/[0-9]*/limits",
    "allowlist/read+/proc/[0-9]*/loginuid",
    "allowlist/read+/proc/[0-9]*/map_files/***",
    "allowlist/read+/proc/[0-9]*/maps",
    "allowlist/read+/proc/[0-9]*/mounts/***",
    "allowlist/read+/proc/[0-9]*/mountinfo",
    "allowlist/read+/proc/[0-9]*/net/***",
    "allowlist/read+/proc/[0-9]*/ns/***",
    "allowlist/read+/proc/[0-9]*/oom_*",
    "allowlist/read+/proc/[0-9]*/personality",
    "allowlist/read+/proc/[0-9]*/root",
    "allowlist/read+/proc/[0-9]*/sched{,stat}",
    "allowlist/read+/proc/[0-9]*/seccomp",
    "allowlist/read+/proc/[0-9]*/sessionid",
    "allowlist/read+/proc/[0-9]*/setgroups",
    "allowlist/read+/proc/[0-9]*/stat{,m,us}",
    "allowlist/read+/proc/[0-9]*/task",
    "allowlist/read+/proc/[0-9]*/task/*",
    "allowlist/read+/proc/[0-9]*/time*",
    "allowlist/read+/proc/cgroups",
    "allowlist/read+/proc/cmdline",
    "allowlist/read+/proc/consoles",
    "allowlist/read+/proc/devices",
    "allowlist/read+/proc/diskstats",
    "allowlist/read+/proc/execdomains",
    "allowlist/read+/proc/fb",
    "allowlist/read+/proc/filesystems",
    "allowlist/read+/proc/interrupts",
    "allowlist/read+/proc/loadavg",
    "allowlist/read+/proc/mounts/***",
    "allowlist/read+/proc/net/***",
    "allowlist/read+/proc/stat",
    "allowlist/read+/proc/sys/fs/inotify/***",
    "allowlist/read+/proc/sys/kernel/***",
    "allowlist/read+/proc/uptime",
    "allowlist/read+/proc/version",
    "allowlist/read+/proc/{,thread-}self",
    "allowlist/read+/proc/{cpu,mem}info",
    "allowlist/read+/sys/class/i2c-adapter",
    "allowlist/read+/sys/class/power_supply/***",
    "allowlist/read+/sys/fs",
    "allowlist/read+/sys/fs/cgroup/***",
    "allowlist/read+/sys/kernel/mm/transparent_hugepage/***",
    "allowlist/read+/sys/kernel/security/apparmor/***",
    "allowlist/stat+/***",
    "denylist/stat+/dev/*/***",
    "denylist/stat+/etc/crypttab",
    "denylist/stat+/etc/gshadow*",
    "denylist/stat+/etc/shadow*",
    "denylist/stat+/proc/*/***",
    "denylist/stat+/sys/*/***",
    "denylist/stat+/var/log/***",
    "allowlist/stat+/dev/console",
    "allowlist/stat+/dev/fd/**",
    "allowlist/stat+/dev/full",
    "allowlist/stat+/dev/null",
    "allowlist/stat+/dev/ptmx",
    "allowlist/stat+/dev/pts/**",
    "allowlist/stat+/dev/pty*",
    "allowlist/stat+/dev/shm/***",
    "allowlist/stat+/dev/stderr",
    "allowlist/stat+/dev/stdout",
    "allowlist/stat+/dev/tts",
    "allowlist/stat+/dev/tty*",
    "allowlist/stat+/dev/{,u}random",
    "allowlist/stat+/dev/zero",
    "allowlist/stat+/proc/[0-9]*",
    "denylist/stat+/proc/[0-9]*/*",
    "allowlist/stat+/proc/[0-9]*/[ug]id_map",
    "allowlist/stat+/proc/[0-9]*/cgroup",
    "allowlist/stat+/proc/[0-9]*/clear_refs",
    "allowlist/stat+/proc/[0-9]*/cmdline",
    "allowlist/stat+/proc/[0-9]*/comm",
    "allowlist/stat+/proc/[0-9]*/coredump_filter",
    "allowlist/stat+/proc/[0-9]*/coredump_filter",
    "allowlist/stat+/proc/[0-9]*/cpuset",
    "allowlist/stat+/proc/[0-9]*/cwd",
    "allowlist/stat+/proc/[0-9]*/environ",
    "allowlist/stat+/proc/[0-9]*/exe",
    "allowlist/stat+/proc/[0-9]*/fd/***",
    "allowlist/stat+/proc/[0-9]*/fdinfo/***",
    "allowlist/stat+/proc/[0-9]*/io",
    "allowlist/stat+/proc/[0-9]*/limits",
    "allowlist/stat+/proc/[0-9]*/loginuid",
    "allowlist/stat+/proc/[0-9]*/map_files/***",
    "allowlist/stat+/proc/[0-9]*/maps",
    "allowlist/stat+/proc/[0-9]*/mounts/***",
    "allowlist/stat+/proc/[0-9]*/mountinfo",
    "allowlist/stat+/proc/[0-9]*/net/***",
    "allowlist/stat+/proc/[0-9]*/ns/***",
    "allowlist/stat+/proc/[0-9]*/oom_*",
    "allowlist/stat+/proc/[0-9]*/personality",
    "allowlist/stat+/proc/[0-9]*/root",
    "allowlist/stat+/proc/[0-9]*/sched{,stat}",
    "allowlist/stat+/proc/[0-9]*/seccomp",
    "allowlist/stat+/proc/[0-9]*/sessionid",
    "allowlist/stat+/proc/[0-9]*/setgroups",
    "allowlist/stat+/proc/[0-9]*/stat{,m,us}",
    "allowlist/stat+/proc/[0-9]*/task",
    "allowlist/stat+/proc/[0-9]*/task/*",
    "allowlist/stat+/proc/[0-9]*/time*",
    "allowlist/stat+/proc/cgroups",
    "allowlist/stat+/proc/cmdline",
    "allowlist/stat+/proc/consoles",
    "allowlist/stat+/proc/devices",
    "allowlist/stat+/proc/diskstats",
    "allowlist/stat+/proc/execdomains",
    "allowlist/stat+/proc/fb",
    "allowlist/stat+/proc/filesystems",
    "allowlist/stat+/proc/interrupts",
    "allowlist/stat+/proc/loadavg",
    "allowlist/stat+/proc/mounts/***",
    "allowlist/stat+/proc/net/***",
    "allowlist/stat+/proc/stat",
    "allowlist/stat+/proc/sys/fs/inotify/***",
    "allowlist/stat+/proc/sys/kernel/***",
    "allowlist/stat+/proc/uptime",
    "allowlist/stat+/proc/version",
    "allowlist/stat+/proc/{,thread-}self",
    "allowlist/stat+/proc/{cpu,mem}info",
    "allowlist/stat+/sys/class/i2c-adapter",
    "allowlist/stat+/sys/class/power_supply/***",
    "allowlist/stat+/sys/fs",
    "allowlist/stat+/sys/fs/cgroup/***",
    "allowlist/stat+/sys/kernel/mm/transparent_hugepage/***",
    "allowlist/stat+/sys/kernel/security/apparmor/***",
    "allowlist/write+/dev/console",
    "allowlist/write+/dev/fd/**",
    "allowlist/write+/dev/full",
    "allowlist/write+/dev/null",
    "allowlist/write+/dev/ptmx",
    "allowlist/write+/dev/pts/**",
    "allowlist/write+/dev/pty*",
    "allowlist/write+/dev/shm/**",
    "allowlist/write+/dev/stderr",
    "allowlist/write+/dev/stdout",
    "allowlist/write+/dev/tts",
    "allowlist/write+/dev/tty*",
    "allowlist/write+/dev/zero",
    "allowlist/write+/proc/[0-9]*/[ug]id_map",
    "allowlist/write+/proc/[0-9]*/attr/***",
    "allowlist/write+/proc/[0-9]*/fd/***",
    "allowlist/write+/proc/[0-9]*/oom_score_adj",
    "allowlist/write+/proc/[0-9]*/setgroups",
    "allowlist/write+/proc/[0-9]*/task/***",
    "allowlist/write+/selinux/context/**",
    "allowlist/write+/tmp/***",
    "allowlist/write+/var/cache/**",
    "allowlist/write+/var/tmp/***",
    "allowlist/net/bind+LOOPBACK@0",
    "allowlist/net/bind+LOOPBACK@1024-65535",
    "allowlist/net/bind+LOOPBACK6@0",
    "allowlist/net/bind+LOOPBACK6@1024-65535",
    "allowlist/net/connect+/run/nscd/socket",
    "allowlist/net/connect+/var/lib/sss/pipes/nss",
    "allowlist/net/connect+/var/run/nscd/socket",
    // allow getaddrinfo() with AI_ADDRCONFIG on musl systems.
    "allowlist/net/connect+LOOPBACK@65535",
    "allowlist/net/connect+LOOPBACK6@65535",
];

/// Size of the paludis sandboxing profile.
pub const PROFILE_USER_SIZE: usize = 253;
/// User sandboxing profile, activated with:
/// `syd --profile user`.
pub const PROFILE_USER: [&str; PROFILE_USER_SIZE] = [
    "sandbox/lock:on",
    "sandbox/read:on",
    "sandbox/stat:on",
    "sandbox/write:on",
    "sandbox/exec:on",
    "sandbox/net:on",
    "trace/allow_successful_bind:true",
    "trace/allow_unsupported_socket_families:true",
    "allowlist/lock/read+/bin",
    "allowlist/lock/read+/dev",
    "allowlist/lock/read+/etc",
    "allowlist/lock/read+/opt",
    "allowlist/lock/read+/run",
    "allowlist/lock/read+/snap",
    "allowlist/lock/read+/sys",
    "allowlist/lock/read+/usr",
    "allowlist/lock/read+/var",
    "allowlist/lock/write+/dev/console",
    "allowlist/lock/write+/dev/dri",
    "allowlist/lock/write+/dev/fd",
    "allowlist/lock/write+/dev/full",
    "allowlist/lock/write+/dev/null",
    "allowlist/lock/write+/dev/ptmx",
    "allowlist/lock/write+/dev/pts",
    "allowlist/lock/write+/dev/shm",
    "allowlist/lock/write+/dev/snd",
    "allowlist/lock/write+/dev/stderr",
    "allowlist/lock/write+/dev/stdout",
    "allowlist/lock/write+/dev/tts",
    "allowlist/lock/write+/dev/tty",
    "allowlist/lock/write+/dev/zero",
    "allowlist/lock/write+/proc",
    "allowlist/lock/write+/run/user",
    "allowlist/lock/write+/selinux/context",
    "allowlist/lock/write+/tmp",
    "allowlist/lock/write+/var/cache",
    "allowlist/lock/write+/var/tmp",
    "allowlist/read+/***",
    "denylist/read+/dev/*/***",
    "denylist/read+/etc/crypttab",
    "denylist/read+/etc/gshadow*",
    "denylist/read+/etc/shadow*",
    "denylist/read+/proc/*/***",
    "denylist/read+/sys/*/***",
    "denylist/read+/var/log/***",
    "allowlist/read+/dev/console",
    "allowlist/read+/dev/dri/***",
    "allowlist/read+/dev/fd/**",
    "allowlist/read+/dev/full",
    "allowlist/read+/dev/input/***",
    "allowlist/read+/dev/null",
    "allowlist/read+/dev/ptmx",
    "allowlist/read+/dev/pts/**",
    "allowlist/read+/dev/pty*",
    "allowlist/read+/dev/shm/***",
    "allowlist/read+/dev/snd/***",
    "allowlist/read+/dev/stderr",
    "allowlist/read+/dev/stdout",
    "allowlist/read+/dev/tts",
    "allowlist/read+/dev/tty*",
    "allowlist/read+/dev/zero",
    "allowlist/read+/dev/urandom",
    "allowlist/read+/proc/[0-9]*",
    "denylist/read+/proc/[0-9]*/*",
    "allowlist/read+/proc/[0-9]*/[ug]id_map",
    "allowlist/read+/proc/[0-9]*/cgroup",
    "allowlist/read+/proc/[0-9]*/clear_refs",
    "allowlist/read+/proc/[0-9]*/cmdline",
    "allowlist/read+/proc/[0-9]*/comm",
    "allowlist/read+/proc/[0-9]*/coredump_filter",
    "allowlist/read+/proc/[0-9]*/coredump_filter",
    "allowlist/read+/proc/[0-9]*/cpuset",
    "allowlist/read+/proc/[0-9]*/cwd",
    "allowlist/read+/proc/[0-9]*/environ",
    "allowlist/read+/proc/[0-9]*/exe",
    "allowlist/read+/proc/[0-9]*/fd/***",
    "allowlist/read+/proc/[0-9]*/fdinfo/***",
    "allowlist/read+/proc/[0-9]*/io",
    "allowlist/read+/proc/[0-9]*/limits",
    "allowlist/read+/proc/[0-9]*/loginuid",
    "allowlist/read+/proc/[0-9]*/map_files/***",
    "allowlist/read+/proc/[0-9]*/maps",
    "allowlist/read+/proc/[0-9]*/mounts/***",
    "allowlist/read+/proc/[0-9]*/mountinfo",
    "allowlist/read+/proc/[0-9]*/net/***",
    "allowlist/read+/proc/[0-9]*/ns/***",
    "allowlist/read+/proc/[0-9]*/oom_*",
    "allowlist/read+/proc/[0-9]*/personality",
    "allowlist/read+/proc/[0-9]*/root",
    "allowlist/read+/proc/[0-9]*/sched{,stat}",
    "allowlist/read+/proc/[0-9]*/seccomp",
    "allowlist/read+/proc/[0-9]*/sessionid",
    "allowlist/read+/proc/[0-9]*/setgroups",
    "allowlist/read+/proc/[0-9]*/stat{,m,us}",
    "allowlist/read+/proc/[0-9]*/task",
    "allowlist/read+/proc/[0-9]*/task/*",
    "allowlist/read+/proc/[0-9]*/time*",
    "allowlist/read+/proc/cgroups",
    "allowlist/read+/proc/cmdline",
    "allowlist/read+/proc/consoles",
    "allowlist/read+/proc/devices",
    "allowlist/read+/proc/diskstats",
    "allowlist/read+/proc/execdomains",
    "allowlist/read+/proc/fb",
    "allowlist/read+/proc/filesystems",
    "allowlist/read+/proc/interrupts",
    "allowlist/read+/proc/loadavg",
    "allowlist/read+/proc/mounts/***",
    "allowlist/read+/proc/net/***",
    "allowlist/read+/proc/stat",
    "allowlist/read+/proc/sys/fs/inotify/***",
    "allowlist/read+/proc/sys/kernel/***",
    "allowlist/read+/proc/uptime",
    "allowlist/read+/proc/version",
    "allowlist/read+/proc/{,thread-}self",
    "allowlist/read+/proc/{cpu,mem}info",
    "allowlist/read+/sys/{bus,class,dev,devices,fs}/***",
    "allowlist/read+/sys/class/***",
    "allowlist/read+/sys/dev/***",
    "allowlist/read+/sys/devices/***",
    "allowlist/read+/sys/fs/***",
    "allowlist/read+/sys/kernel/mm/transparent_hugepage/***",
    "allowlist/read+/sys/kernel/security/apparmor/***",
    "allowlist/stat+/***",
    "denylist/stat+/dev/*/***",
    "denylist/stat+/proc/*/***",
    "denylist/stat+/sys/*/***",
    "denylist/stat+/var/log/***",
    "allowlist/stat+/dev/console",
    "allowlist/stat+/dev/dri/***",
    "allowlist/stat+/dev/fd/**",
    "allowlist/stat+/dev/full",
    "allowlist/stat+/dev/input/***",
    "allowlist/stat+/dev/null",
    "allowlist/stat+/dev/ptmx",
    "allowlist/stat+/dev/pts/**",
    "allowlist/stat+/dev/pty*",
    "allowlist/stat+/dev/shm/***",
    "allowlist/stat+/dev/snd/***",
    "allowlist/stat+/dev/stderr",
    "allowlist/stat+/dev/stdout",
    "allowlist/stat+/dev/tts",
    "allowlist/stat+/dev/tty*",
    "allowlist/stat+/dev/zero",
    "allowlist/stat+/dev/{,u}random",
    "allowlist/stat+/proc/[0-9]*",
    "denylist/stat+/proc/[0-9]*/*",
    "allowlist/stat+/proc/[0-9]*/[ug]id_map",
    "allowlist/stat+/proc/[0-9]*/cgroup",
    "allowlist/stat+/proc/[0-9]*/clear_refs",
    "allowlist/stat+/proc/[0-9]*/cmdline",
    "allowlist/stat+/proc/[0-9]*/comm",
    "allowlist/stat+/proc/[0-9]*/coredump_filter",
    "allowlist/stat+/proc/[0-9]*/coredump_filter",
    "allowlist/stat+/proc/[0-9]*/cpuset",
    "allowlist/stat+/proc/[0-9]*/cwd",
    "allowlist/stat+/proc/[0-9]*/environ",
    "allowlist/stat+/proc/[0-9]*/exe",
    "allowlist/stat+/proc/[0-9]*/fd/***",
    "allowlist/stat+/proc/[0-9]*/fdinfo/***",
    "allowlist/stat+/proc/[0-9]*/io",
    "allowlist/stat+/proc/[0-9]*/limits",
    "allowlist/stat+/proc/[0-9]*/loginuid",
    "allowlist/stat+/proc/[0-9]*/map_files/***",
    "allowlist/stat+/proc/[0-9]*/maps",
    "allowlist/stat+/proc/[0-9]*/mounts/***",
    "allowlist/stat+/proc/[0-9]*/mountinfo",
    "allowlist/stat+/proc/[0-9]*/net/***",
    "allowlist/stat+/proc/[0-9]*/ns/***",
    "allowlist/stat+/proc/[0-9]*/oom_*",
    "allowlist/stat+/proc/[0-9]*/personality",
    "allowlist/stat+/proc/[0-9]*/root",
    "allowlist/stat+/proc/[0-9]*/sched{,stat}",
    "allowlist/stat+/proc/[0-9]*/seccomp",
    "allowlist/stat+/proc/[0-9]*/sessionid",
    "allowlist/stat+/proc/[0-9]*/setgroups",
    "allowlist/stat+/proc/[0-9]*/stat{,m,us}",
    "allowlist/stat+/proc/[0-9]*/task",
    "allowlist/stat+/proc/[0-9]*/task/*",
    "allowlist/stat+/proc/[0-9]*/time*",
    "allowlist/stat+/proc/cgroups",
    "allowlist/stat+/proc/cmdline",
    "allowlist/stat+/proc/consoles",
    "allowlist/stat+/proc/devices",
    "allowlist/stat+/proc/diskstats",
    "allowlist/stat+/proc/execdomains",
    "allowlist/stat+/proc/fb",
    "allowlist/stat+/proc/filesystems",
    "allowlist/stat+/proc/interrupts",
    "allowlist/stat+/proc/loadavg",
    "allowlist/stat+/proc/mounts/***",
    "allowlist/stat+/proc/net/***",
    "allowlist/stat+/proc/stat",
    "allowlist/stat+/proc/sys/fs/inotify/***",
    "allowlist/stat+/proc/sys/kernel/***",
    "allowlist/stat+/proc/uptime",
    "allowlist/stat+/proc/version",
    "allowlist/stat+/proc/{,thread-}self",
    "allowlist/stat+/proc/{cpu,mem}info",
    "allowlist/stat+/sys/{bus,class,dev,devices,fs}/***",
    "allowlist/stat+/sys/kernel/mm/transparent_hugepage/***",
    "allowlist/stat+/sys/kernel/security/apparmor/***",
    "allowlist/exec+/bin/**",
    "allowlist/exec+/lib*/**",
    "allowlist/exec+/sbin/**",
    "allowlist/exec+/snap/**",
    "allowlist/exec+/usr/**/bin/**",
    "allowlist/exec+/usr/**/lib*/**",
    "allowlist/exec+/usr/**/sbin/**",
    "allowlist/exec+/usr/games/**",
    "allowlist/write+/dev/console",
    "allowlist/write+/dev/dri/**",
    "allowlist/write+/dev/fd/**",
    "allowlist/write+/dev/full",
    "allowlist/write+/dev/hidraw[0-9]*",
    "allowlist/write+/dev/null",
    "allowlist/write+/dev/ptmx",
    "allowlist/write+/dev/pts/**",
    "allowlist/write+/dev/pty*",
    "allowlist/write+/dev/shm/**",
    "allowlist/write+/dev/snd/**",
    "allowlist/write+/dev/stderr",
    "allowlist/write+/dev/stdout",
    "allowlist/write+/dev/tts",
    "allowlist/write+/dev/tty*",
    "allowlist/write+/dev/zero",
    "allowlist/write+/proc/[0-9]*/[ug]id_map",
    "allowlist/write+/proc/[0-9]*/attr/***",
    "allowlist/write+/proc/[0-9]*/fd/***",
    "allowlist/write+/proc/[0-9]*/oom_score_adj",
    "allowlist/write+/proc/[0-9]*/setgroups",
    "allowlist/write+/proc/[0-9]*/task/***",
    "allowlist/write+/run/**",
    "allowlist/write+/selinux/context/**",
    "allowlist/write+/tmp/***",
    "allowlist/write+/var/cache/**",
    "allowlist/write+/var/tmp/***",
    "allowlist/net/bind+LOOPBACK@0",
    "allowlist/net/bind+LOOPBACK@1024-65535",
    "allowlist/net/bind+LOOPBACK6@0",
    "allowlist/net/bind+LOOPBACK6@1024-65535",
    "allowlist/net/connect+/run/nscd/socket",
    "allowlist/net/connect+/var/lib/sss/pipes/nss",
    "allowlist/net/connect+/var/run/nscd/socket",
    // allow getaddrinfo() with AI_ADDRCONFIG on musl systems.
    "allowlist/net/connect+LOOPBACK@65535",
    "allowlist/net/connect+LOOPBACK6@65535",
    // Allow UNIX socket access to run & /tmp (think X, tmux, screen et a.)
    "allowlist/net/bind+/run/**",
    "allowlist/net/bind+/tmp/**",
    "allowlist/net/connect+/run/**",
    "allowlist/net/connect+/tmp/**",
    // Allow access to syslog
    "allowlist/net/connect+/dev/log",
    // Allow access to systemd journal
    "allowlist/read+/var/log/journal/***",
    // Allow access to PulseAudio
    "allowlist/net/connect+/var/run/pulse/native",
];

/// The size of getid syscalls array
pub const GET_ID_SYSCALLS_SIZE: usize = 8;
/// The list of system calls which are of the getid family.
pub const GET_ID_SYSCALLS: [&str; GET_ID_SYSCALLS_SIZE] = [
    "getuid",
    "getuid32",
    "getgid",
    "getgid32",
    "geteuid",
    "geteuid32",
    "getegid",
    "getegid32",
];

/// The size of the setid syscalls array
pub const SET_ID_SYSCALLS_SIZE: usize = 16;
/// The list of system calls which are of the setid family.
pub const SET_ID_SYSCALLS: [&str; SET_ID_SYSCALLS_SIZE] = [
    "setfsgid",
    "setfsgid32",
    "setfsuid",
    "setfsuid32",
    "setgid",
    "setgid32",
    "setregid",
    "setregid32",
    "setresgid",
    "setresgid32",
    "setresuid",
    "setresuid32",
    "setreuid",
    "setreuid32",
    "setuid",
    "setuid32",
];

/// The size of hook syscalls array
pub const HOOK_SYSCALLS_SIZE: usize = 69;
/// The list of system calls which SydB☮x has a `UNotify` callback for
pub const HOOK_SYSCALLS: [&str; HOOK_SYSCALLS_SIZE] = [
    "bind",
    "connect",
    "recvfrom",
    "sendto",
    "execve",
    "execveat",
    "chdir",
    "fchdir",
    "getdents",
    "getdents64",
    "stat",
    "stat64",
    "fstat",
    "lstat",
    "statx",
    "newfstatat",
    "access",
    "faccessat",
    "faccessat2",
    "chmod",
    "fchmod",
    "fchmodat",
    "chown",
    "fchown",
    "fchown32",
    "lchown",
    "fchownat",
    "creat",
    "link",
    "symlink",
    "unlink",
    "linkat",
    "symlinkat",
    "unlinkat",
    "mkdir",
    "rmdir",
    "mkdirat",
    "mknod",
    "mknodat",
    "mount",
    "umount",
    "umount2",
    "open",
    "openat",
    "openat2",
    "rename",
    "renameat",
    "renameat2",
    "utime",
    "utimes",
    "futimesat",
    "utimensat",
    "truncate",
    "truncate64",
    "getxattr",
    "fgetxattr",
    "lgetxattr",
    "setxattr",
    "fsetxattr",
    "lsetxattr",
    "listxattr",
    "flistxattr",
    "llistxattr",
    "removexattr",
    "fremovexattr",
    "lremovexattr",
    "getrandom",
    "ioctl",
    "prctl",
];

/// The size of safe syscalls array
pub const SAFE_SYSCALLS_SIZE: usize = 242;
/// The list of system calls which are allowlisted without any filtering.
pub const SAFE_SYSCALLS: [&str; SAFE_SYSCALLS_SIZE] = [
    "_llseek",
    "_newselect",
    //"_sysctl",
    "accept",
    "accept4",
    //"acct",
    //add_key
    //adjtimex
    //afs_syscall
    "alarm",
    "arch_prctl",
    //"bpf",
    "breakpoint", // arm
    "brk",
    "cacheflush", // arm
    "capget",
    "capset",
    "chdir",
    //"chroot",
    //clock_adjtime
    "clock_getres",
    "clock_gettime",
    "clock_nanosleep",
    "clock_settime",
    "clone",
    "clone3",
    "close",
    "close_range",
    "copy_file_range",
    //create_module
    //delete_module
    "dup",
    "dup2",
    "dup3",
    "epoll_create",
    "epoll_create1",
    "epoll_ctl",
    "epoll_ctl_old",
    "epoll_pwait",
    "epoll_pwait2",
    "epoll_wait",
    "epoll_wait_old",
    "eventfd",
    "eventfd2",
    "exit",
    "exit_group",
    "fadvise64",
    "fallocate",
    "fanotify_init",
    "fanotify_mark",
    "fcntl",
    "fcntl64",
    "fdatasync",
    //finit_module
    "flock",
    "fork",
    //fsconfig
    //fsmount
    //fsopen
    //fspick
    "fstatfs",
    "fstatfs64",
    "fsync",
    "ftruncate",
    "ftruncate64",
    "futex",
    //get_kernel_syms
    "get_mempolicy",
    "get_robust_list",
    "get_thread_area",
    "getcpu",
    "getcwd",
    "getdents",
    "getdents64",
    "getegid",
    "getegid32",
    "geteuid",
    "geteuid32",
    "getgid",
    "getgid32",
    "getgroups",
    "getgroups32",
    "getitimer",
    "getpeername",
    "getpgid",
    "getpgrp",
    "getpid",
    "getpmsg",
    "getppid",
    "getpriority",
    //"getrandom", restrictions applied, see setup_seccomp.
    "getresgid",
    "getresuid",
    "getrlimit",
    "getrusage",
    "getsid",
    "getuid",
    "getsockname",
    "getsockopt",
    "gettid",
    "gettimeofday",
    //init_module
    "inotify_add_watch",
    "inotify_init",
    "inotify_init1",
    "inotify_rm_watch",
    /*
     * SAFETY: io-uring can be used to bypass sandbox restrictions.
    "io_cancel",
    "io_destroy",
    "io_getevents",
    "io_pgetevents",
    "io_setup",
    "io_submit",
    "io_uring_enter",
    "io_uring_register",
    "io_uring_setup",
    */
    //"ioctl", restrictions applied, see setup_seccomp.
    //"ioperm",
    //"iopl",
    "ioprio_get",
    "ioprio_set",
    "kcmp",
    //kexec_file_load,
    //kexec_load,
    //keyctl,
    "kill",
    "landlock_add_rule",
    "landlock_create_ruleset",
    "landlock_restrict_self",
    "listen",
    "lookup_dcookie",
    "lseek",
    "madvise",
    //"mbind",
    "membarrier",
    "memfd_create",
    "memfd_secret",
    //"migrate_pages",
    "mincore",
    "mlock",
    "mlock2",
    "mlockall",
    "mmap",
    "modify_ldt",
    //mount_setattr
    //"move_pages",
    "mprotect",
    "mq_getsetattr",
    "mq_notify",
    "mq_open",
    "mq_timedreceive",
    "mq_timedsend",
    "mq_unlink",
    "mremap",
    "msgctl",
    "msgget",
    "msgrcv",
    "msgsnd",
    "msync",
    "munlock",
    "munlockall",
    "munmap",
    //TODO:name_to_handle_at
    "nanosleep",
    //nfsservctl,
    "open_by_handle_at",
    //open_tree
    "pause",
    //"perf_event_open",
    "personality",
    "pidfd_getfd",
    "pidfd_open",
    "pidfd_send_signal",
    "pipe",
    "pipe2",
    //pivot_root,
    "pkey_alloc",
    "pkey_free",
    "pkey_mprotect",
    "poll",
    "ppoll",
    // "prctl", restrictions applied, see setup_seccomp.
    "pread64",
    "preadv",
    "preadv2",
    "prlimit64",
    "process_madvise",
    "process_mrelease",
    //process_vm_readv
    //process_vm_writev
    "pselect6",
    //ptrace
    //"putpmsg",
    "pwrite64",
    "pwritev",
    "pwritev2",
    //query_module,
    //quotactl
    //quotactl_fd
    "read",
    "readahead",
    "readlink",
    "readlinkat",
    "readv",
    //reboot
    "recvmmsg",
    "recvmsg",
    "remap_file_pages",
    //request_key
    "restart_syscall",
    "rseq",
    "rt_sigaction",
    "rt_sigpending",
    "rt_sigprocmask",
    "rt_sigqueueinfo",
    "rt_sigreturn",
    "rt_sigsuspend",
    "rt_sigtimedwait",
    "rt_tgsigqueueinfo",
    "s390_pci_mmio_read",
    "s390_pci_mmio_write",
    "s390_runtime_instr",
    "sched_get_priority_max",
    "sched_get_priority_min",
    "sched_getaffinity",
    "sched_getattr",
    "sched_getparam",
    "sched_getscheduler",
    "sched_setaffinity",
    "sched_setattr",
    "sched_setscheduler",
    "sched_yield",
    "seccomp",
    //security
    "select",
    "semctl",
    "semget",
    "semop",
    "semtimedop",
    "sendfile",
    "sendfile64",
    "sendmmsg",
    "sendmsg",
    //"set_mempolicy",
    "set_robust_list",
    "set_thread_area",
    "set_tid_address",
    "setdomainname",
    //"setfsgid",
    //"setfsgid32",
    //"setfsuid",
    //"setfsuid32",
    //"setgid",
    //"setgid32",
    //"setgroups",
    //"setgroups32",
    "sethostname",
    "setitimer",
    "setns",
    "setpgid",
    "setpriority",
    //"setregid",
    //"setregid32",
    //"setresgid",
    //"setresgid32",
    //"setresuid",
    //"setresuid32",
    //"setreuid",
    //"setreuid32",
    "setrlimit",
    "setsid",
    "setsockopt",
    //"settimeofday"
    //"setuid",
    //"setuid32",
    "set_tls", // arm
    "shmat",
    "shmctl",
    "shmdt",
    "shmget",
    "shutdown",
    "sigaltstack",
    "signalfd",
    "signalfd4",
    "socket",
    "socketpair",
    "splice",
    "statfs",
    "statfs64",
    //swapoff
    //swapon
    "sync",
    "sync_file_range",
    "syncfs",
    "sysinfo",
    //"syslog",
    "tee",
    "tgkill",
    "time",
    "timer_create",
    "timer_delete",
    "timer_getoverrun",
    "timer_gettime",
    "timer_settime",
    "timerfd_create",
    "timerfd_gettime",
    "timerfd_settime",
    "times",
    "tkill",
    //tuxcall
    "umask",
    "uname",
    "unshare",
    //"uselib",
    //"userfaultfd",
    "ustat",
    "vfork",
    //"vhangup",
    "vmsplice",
    //vserver
    "wait4",
    "waitid",
    "write",
    "writev",
];
