//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/proc.rs: /proc utilities
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    fs::{read_link, File},
    io::{BufRead, BufReader, Read},
};

use nix::{errno::Errno, unistd::Pid};

/// Retrieves the thread group ID (TGID) for the specified process ID (PID).
///
/// This function reads the `/proc/<pid>/status` file and extracts the
/// TGID from the "Tgid:" line.
///
/// # Arguments
///
/// * `pid` - The process ID for which to retrieve the thread group ID.
///
/// # Returns
///
/// * `Ok(Pid)` - The thread group ID as a `Pid` type if found.
/// * `Err(Errno)` - An error, typically:
///     * `Errno::ENOENT` if the "Tgid:" line is not found in the status file.
///     * `Errno::EINVAL` if the TGID is found but cannot be parsed as an integer.
///
/// # Examples
///
/// ```rust
/// use nix::unistd::Pid;
/// use syd::proc::proc_tgid;
///
/// let tgid = proc_tgid(Pid::this());
/// assert!(tgid.is_ok());
/// ```
pub fn proc_tgid(pid: Pid) -> Result<Pid, Errno> {
    let file =
        BufReader::new(File::open(format!("/proc/{pid}/status")).map_err(|_| Errno::last())?);

    for line in file.lines() {
        match line {
            Ok(line) => {
                if line.starts_with("Tgid:") {
                    let tgid_str = line.trim_start_matches("Tgid:").trim();
                    let tgid: i32 = tgid_str.parse().map_err(|_| Errno::EINVAL)?;
                    return Ok(Pid::from_raw(tgid));
                }
            }
            Err(error) => {
                return Err(error
                    .raw_os_error()
                    .map(Errno::from_i32)
                    .unwrap_or(Errno::ESRCH));
            }
        };
    }

    Err(Errno::ESRCH)
}

/// Retrieves the command line of the specified process ID (PID)
/// concatenated as a single string.
///
/// This function reads the `/proc/<pid>/cmdline` file and concatenates
/// the arguments using spaces. The function takes care of replacing null
/// bytes (`'\0'`) with spaces to format the command line as a readable string.
///
/// # Arguments
///
/// * `pid` - The process ID for which to retrieve the command line.
///
/// # Returns
///
/// * `Ok(String)` - The command line of the process as a single string.
/// * `Err(Errno)` - An error, depending on the issue encountered while reading
///                  or processing the cmdline file.
///
/// # Examples
///
/// ```rust
/// use nix::unistd::Pid;
/// use syd::proc::proc_cmdline;
///
/// let cmdline = proc_cmdline(Pid::this());
/// assert!(cmdline.is_ok());
/// ```
pub fn proc_cmdline(pid: Pid) -> Result<String, Errno> {
    const LIMIT: usize = 256;
    let path = format!("/proc/{pid}/cmdline");
    let file = File::open(path).map_err(|_| Errno::last())?;

    let mut data = Vec::with_capacity(LIMIT + 1 /* sentinel */);
    let reader = BufReader::new(file);

    // Read up to LIMIT+1 characters or until EOF
    for byte in reader.bytes().take(LIMIT + 1 /* sentinel */) {
        let byte = byte.map_err(|_| Errno::last())?;
        if byte == 0 {
            // Replace null bytes with spaces
            data.push(b' ');
        } else {
            data.push(byte);
        }
    }

    let cmdline = String::from_utf8_lossy(&data);
    let cmdline = if cmdline.len() > LIMIT {
        let cutoff = cmdline[..LIMIT].rfind(' ').unwrap_or(LIMIT);
        format!("{}…", &cmdline[..cutoff].trim_end_matches(' '))
    } else {
        cmdline.trim_end_matches(' ').to_string()
    };

    Ok(cmdline)
}

/// Retrieves the current working directory (CWD) of the specified process ID (PID).
///
/// This function reads the symbolic link `/proc/<pid>/cwd` to determine the CWD.
///
/// # Arguments
///
/// * `pid` - The process ID for which to retrieve the current working directory.
///
/// # Returns
///
/// * `Ok(String)` - The current working directory of the process.
/// * `Err(Errno)` - An error, depending on the issue encountered while reading
///                  or resolving the `cwd` symlink.
///
/// # Examples
///
/// ```rust
/// use nix::unistd::Pid;
/// use syd::proc::proc_cwd;
///
/// let cwd = proc_cwd(Pid::this());
/// assert!(cwd.is_ok());
/// ```
pub fn proc_cwd(pid: Pid) -> Result<String, Errno> {
    let path = format!("/proc/{pid}/cwd");
    read_link(path)
        .map(|p| p.to_string_lossy().to_string())
        .map_err(|_| Errno::last())
}
