use std::{ffi::CString, mem, ptr};

use nix::{
    self,
    errno::Errno,
    libc::{
        self, c_ulong, setdomainname, signal, sigset_t, FD_CLOEXEC, F_GETFD, F_SETFD, MNT_DETACH,
        SIG_DFL, SIG_SETMASK,
    },
    sched::CloneFlags,
    unistd::{close, sethostname},
};

use crate::unshare::{caps, error::ErrorCode as Err, run::ChildInfo};

// And at this point we've reached a special time in the life of the
// child. The child must now be considered hamstrung and unable to
// do anything other than syscalls really.
//
// ESPECIALLY YOU CAN NOT DO MEMORY (DE)ALLOCATIONS
//
// See better explanation at:
// https://github.com/rust-lang/rust/blob/c1e865c/src/libstd/sys/unix/process.rs#L202
//

// In particular ChildInfo is passed by refernce here to avoid
// deallocating (parts of) it.
#[allow(clippy::cognitive_complexity)]
pub unsafe fn child_after_clone(child: &ChildInfo) -> ! {
    if let Some(&sig) = child.cfg.death_sig.as_ref() {
        if libc::prctl(ffi::PR_SET_PDEATHSIG, sig as c_ulong, 0, 0, 0) != 0 {
            fail(Err::ParentDeathSignal);
        }
    }

    // Set {host,domain}name for UTS namespace.
    if child.cfg.namespaces.contains(CloneFlags::CLONE_NEWUTS) {
        if let Some(name) = child.host_name {
            let _ = sethostname(name);
        }
        if let Some(name) = child.domain_name {
            #[allow(clippy::disallowed_methods)]
            let c_name = CString::new(name.as_str()).unwrap();
            let len = c_name.as_bytes_with_nul().len().saturating_sub(1);
            let _ = setdomainname(c_name.as_ptr(), len);
        }
    }

    if let Some(piv) = child.pivot.as_ref() {
        if ffi::pivot_root(piv.new_root.as_ptr(), piv.put_old.as_ptr()) != 0 {
            fail(Err::ChangeRoot);
        }
        if libc::chdir(piv.workdir.as_ptr()) != 0 {
            fail(Err::ChangeRoot);
        }
        if piv.unmount_old_root && libc::umount2(piv.old_inside.as_ptr(), MNT_DETACH) != 0 {
            fail(Err::ChangeRoot);
        }
    }

    if let Some(chroot) = child.chroot.as_ref() {
        if libc::chroot(chroot.root.as_ptr()) != 0 {
            fail(Err::ChangeRoot);
        }
        if libc::chdir(chroot.workdir.as_ptr()) != 0 {
            fail(Err::ChangeRoot);
        }
    }

    if let Some(&gid) = child.cfg.gid.as_ref() {
        if libc::setgid(gid) != 0 {
            fail(Err::SetUser);
        }
    }

    if let Some(&uid) = child.cfg.uid.as_ref() {
        if libc::setuid(uid) != 0 {
            fail(Err::SetUser);
        }
    }

    //if let Some(ref drop_caps) = child.cfg.drop_caps {
    for cap in CAPS_DROP {
        if libc::prctl(libc::PR_CAPBSET_DROP, cap, 0, 0, 0) != 0 {
            match Errno::last() {
                Errno::EINVAL | Errno::EPERM => {
                    // We either don't have CAP_SETPCAP capability,
                    // or capabilities are not enabled in the kernel.
                    // Let's move on.
                }
                _ => {
                    // Genuine error condition, let's fail.
                    fail(Err::DropCap);
                }
            };
        }
    }

    if let Some(dir) = child.cfg.work_dir.as_ref() {
        if libc::chdir(dir.as_ptr()) != 0 {
            fail(Err::Chdir);
        }
    }

    for &(dest_fd, src_fd) in child.fds {
        if src_fd == dest_fd {
            let flags = libc::fcntl(src_fd, F_GETFD);
            if flags < 0 || libc::fcntl(src_fd, F_SETFD, flags & !FD_CLOEXEC) < 0 {
                fail(Err::StdioError);
            }
        } else if libc::dup2(src_fd, dest_fd) < 0 {
            fail(Err::StdioError);
        }
    }

    for &(start, end) in child.close_fds {
        if start < end {
            for fd in start..end {
                if !child.fds.iter().any(|&(cfd, _)| cfd == fd) {
                    // Close may fail with ebadf, and it's okay
                    libc::close(fd);
                }
            }
        }
    }

    if child.cfg.restore_sigmask {
        let mut sigmask: sigset_t = mem::zeroed();
        libc::sigemptyset(&mut sigmask);
        libc::pthread_sigmask(SIG_SETMASK, &sigmask, ptr::null_mut());
        for sig in 1..32 {
            signal(sig, SIG_DFL);
        }
    }

    if let Some(callback) = child.pre_exec {
        if let Err(e) = callback() {
            fail_errno(Err::PreExec, e.raw_os_error().unwrap_or(10873289));
        }
    }

    if let Some(seccomp_filter) = child.seccomp_filter {
        if seccomp_filter.load().is_err() {
            fail(Err::Seccomp);
        } else if let Some(seccomp_sender) = child.seccomp_sender {
            let fd = match seccomp_filter.get_notify_fd() {
                Ok(fd) => fd,
                Err(_) => {
                    fail(Err::Seccomp);
                }
            };
            let seccomp_sender = crate::hook::Sender {
                fd: *seccomp_sender,
            };
            if seccomp_sender.sendfd(fd).is_err() {
                fail(Err::SeccompNotify);
            }
            let _ = close(fd);
            let _ = close(seccomp_sender.fd);
        }
    }

    libc::execvp(child.filename, child.args.as_ptr());
    fail(Err::Exec);
}

unsafe fn fail(code: Err) -> ! {
    fail_errno(code, nix::errno::errno())
}
unsafe fn fail_errno(code: Err, errno: i32) -> ! {
    let errnox = Errno::from_i32(errno);
    eprintln!("exec error: {code:?} errno:{errnox}");
    libc::_exit(errno);
}

/// We don't use functions from nix here because they may allocate memory
/// which we can't to this this module.
mod ffi {
    use nix::libc::{c_char, c_int};

    pub const PR_SET_PDEATHSIG: c_int = 1;

    extern "C" {
        pub fn pivot_root(new_root: *const c_char, put_old: *const c_char) -> c_int;
    }
}

// The list of Linux capabilities(7) to drop on startup.
// TODO: Make this configurable.
const CAPS_DROP: [caps::Capability; 36] = [
    caps::Capability::CAP_AUDIT_READ,
    caps::Capability::CAP_AUDIT_WRITE,
    caps::Capability::CAP_AUDIT_CONTROL,
    caps::Capability::CAP_BLOCK_SUSPEND,
    caps::Capability::CAP_BPF,
    caps::Capability::CAP_DAC_OVERRIDE,
    caps::Capability::CAP_DAC_READ_SEARCH,
    caps::Capability::CAP_FOWNER,
    caps::Capability::CAP_FSETID,
    caps::Capability::CAP_SETGID,
    caps::Capability::CAP_SETUID,
    caps::Capability::CAP_LINUX_IMMUTABLE,
    caps::Capability::CAP_IPC_LOCK,
    caps::Capability::CAP_SYS_CHROOT,
    caps::Capability::CAP_SYS_PTRACE,
    caps::Capability::CAP_SYS_PACCT,
    caps::Capability::CAP_MAC_ADMIN,
    caps::Capability::CAP_MAC_OVERRIDE,
    caps::Capability::CAP_MKNOD,
    caps::Capability::CAP_LEASE,
    caps::Capability::CAP_NET_ADMIN,
    caps::Capability::CAP_NET_RAW,
    caps::Capability::CAP_PERFMON,
    caps::Capability::CAP_SYSLOG,
    caps::Capability::CAP_SYS_ADMIN,
    caps::Capability::CAP_SYS_BOOT,
    caps::Capability::CAP_SYS_CHROOT,
    caps::Capability::CAP_SYS_MODULE,
    caps::Capability::CAP_SYS_NICE,
    caps::Capability::CAP_SYS_PACCT,
    caps::Capability::CAP_SYS_PTRACE,
    caps::Capability::CAP_SYS_RAWIO,
    caps::Capability::CAP_SYS_RESOURCE,
    caps::Capability::CAP_SYS_TIME,
    caps::Capability::CAP_WAKE_ALARM,
    caps::Capability::CAP_SETPCAP,
];
