//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/hook.rs: Secure computing hooks
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
// Based in part upon greenhook which is under public domain.
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    collections::{HashMap, HashSet},
    ffi::{CStr, CString, OsStr, OsString},
    fs::{File, OpenOptions},
    io::{self, IoSlice, IoSliceMut, Read, Result as IOResult, Seek, SeekFrom, Write},
    mem::{self, size_of, MaybeUninit},
    net::{IpAddr, Ipv4Addr},
    os::{
        fd::{AsRawFd, FromRawFd, OwnedFd, RawFd},
        unix::{
            ffi::{OsStrExt, OsStringExt},
            process::ExitStatusExt,
        },
    },
    path::{Path, PathBuf},
    ptr,
    str::FromStr,
    sync::Arc,
    thread::{Builder, JoinHandle},
};

use anyhow::{bail, Context};
use globset::{Glob, GlobMatcher};
use libseccomp::{
    error::SeccompErrno, ScmpAction, ScmpArch, ScmpArgCompare, ScmpCompareOp, ScmpFilterContext,
    ScmpNotifReq, ScmpNotifResp, ScmpNotifRespFlags, ScmpSyscall, ScmpVersion,
};
use log::{log_enabled, Level};
use nix::{
    cmsg_space,
    errno::Errno,
    fcntl::{open, openat, renameat, AtFlags, OFlag},
    libc::{
        self, c_uint, c_void, cmsghdr, ioctl, msghdr, CMSG_DATA, CMSG_FIRSTHDR, CMSG_LEN,
        CMSG_SPACE,
    },
    mount::{mount, umount, umount2, MntFlags, MsFlags},
    sys::{
        epoll::{
            epoll_create1, epoll_ctl, epoll_wait, EpollCreateFlags, EpollEvent, EpollFlags, EpollOp,
        },
        signal::{kill, SIGHUP, SIGKILL, SIGTSTP, SIGTTIN, SIGTTOU},
        socket::{
            bind, connect, getsockname, recvfrom, recvmsg, sendto, socketpair, AddressFamily,
            ControlMessageOwned, MsgFlags, SockFlag, SockType, SockaddrLike, SockaddrStorage,
        },
        stat::{
            fchmod, fchmodat, fstat, fstatat, lstat, mkdirat, mknodat, umask, FchmodatFlags, Mode,
            SFlag,
        },
        uio::{process_vm_readv, process_vm_writev, RemoteIoVec},
        utsname::uname,
    },
    unistd::{
        close, faccessat, fchown, fchownat, ftruncate, linkat, lseek, mkstemp, symlinkat, truncate,
        unlink, unlinkat, write, AccessFlags, FchownatFlags, Gid, LinkatFlags, Pid, Uid,
        UnlinkatFlags, Whence,
    },
    NixPath,
};
use nonempty::{nonempty, NonEmpty};
use once_cell::sync::Lazy;
use parking_lot::{Mutex, MutexGuard, RwLock};
use serde::{ser::SerializeMap, Serialize};
use threadpool::ThreadPool;

use crate::{
    compat::getdents,
    config::*,
    debug, error,
    fs::normalize_path,
    info,
    log::JsonLinesLogger,
    oflag_fmt,
    proc::proc_tgid,
    proc_umask, replace_self_with_pid,
    sandbox::{Action, Capability, Sandbox, SandboxGuard},
    trace, warn,
};

/// Match the special file paths under /proc
/// 1. /proc/pid/fd/pipe:[100]
/// 2. /proc/pid/fd/socket:[100]
/// 3. /proc/pid/fd/anon_inode:seccomp notify
/// 4. /proc/pid/fd/anon_inode:[eventpoll]
/// 5. /proc/pid/fd/foo (deleted)
/// 6. /proc/pid/task/pid/fd/{1..5}
static GLOB_PROC: Lazy<GlobMatcher> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    Glob::new(r"{* (deleted),/proc/[0-9]*/fd/[!:]*:[!:]*,/proc/[0-9]*/task/[0-9]*/fd/[!:]*:[!:]*}")
        .expect("Invalid proc glob, please file a bug!")
        .compile_matcher()
});

/*
 * Seccomp constants
 */

#[cfg(target_env = "musl")]
type SeccompInt = i32;
#[cfg(not(target_env = "musl"))]
type SeccompInt = u64;

const SECCOMP_IOCTL_NOTIF_ADDFD: SeccompInt = 0x40182103;
/*
const SECCOMP_IOCTL_NOTIF_ID_VALID: SeccompInt = 0x40082102;
const SECCOMP_IOCTL_NOTIF_RECV: SeccompInt = 0xc0502100;
const SECCOMP_IOCTL_NOTIF_SEND: SeccompInt = 0xc0182101;
*/

/*
 * Utilities
 */
/// Converts a reference to `T` into a slice of bytes.
///
/// # Safety
///
/// This function is unsafe because it may produce a slice that does not
/// respect the original alignment requirements of `T`. The caller must ensure
/// that the resultant slice of bytes is only used in ways that are consistent
/// with the memory layout of `T`.
///
/// It's also crucial to ensure that the lifetime of the returned slice does not
/// exceed the lifetime of the input reference, `p`.
unsafe fn any_as_u8_slice<T: Sized>(p: &T) -> &[u8] {
    ::core::slice::from_raw_parts((p as *const T) as *const u8, ::core::mem::size_of::<T>())
}

/// Converts a mutable reference to `T` into a mutable slice of bytes.
///
/// # Safety
///
/// This function is unsafe for the same reasons as `any_as_u8_slice`. It can
/// produce a slice that does not respect the original alignment requirements
/// of `T`. The caller must ensure the resultant slice of bytes is only used in
/// ways that are consistent with the memory layout of `T`.
///
/// Also, the caller should ensure that no other references to `p` exist when
/// this function is called, as this could lead to aliasing mutable references,
/// which is undefined behavior in Rust.
unsafe fn any_as_u8_mut_slice<T: Sized>(p: &mut T) -> &mut [u8] {
    ::core::slice::from_raw_parts_mut((p as *mut T) as *mut u8, ::core::mem::size_of::<T>())
}

fn seccomp2errno(errno: SeccompErrno) -> Errno {
    match errno {
        SeccompErrno::EACCES => Errno::EACCES,
        SeccompErrno::ECANCELED => Errno::ECANCELED,
        SeccompErrno::EDOM => Errno::EDOM,
        SeccompErrno::EEXIST => Errno::EEXIST,
        SeccompErrno::EFAULT => Errno::EFAULT,
        SeccompErrno::EINVAL => Errno::EINVAL,
        SeccompErrno::ENOENT => Errno::ENOENT,
        SeccompErrno::ENOMEM => Errno::ENOMEM,
        SeccompErrno::EOPNOTSUPP => Errno::EOPNOTSUPP,
        SeccompErrno::ERANGE => Errno::ERANGE,
        SeccompErrno::ESRCH => Errno::ESRCH,
        _ => Errno::ECANCELED,
    }
}

fn seccomp_add_arch(ctx: &mut ScmpFilterContext, arch: ScmpArch) -> IOResult<bool> {
    ctx.add_arch(arch).map_err(|e| {
        if let Some(errno) = e.errno() {
            std::io::Error::from_raw_os_error(seccomp2errno(errno) as i32)
        } else {
            std::io::Error::new(std::io::ErrorKind::Other, "Unknown seccomp error")
        }
    })
}

/*
 * Macros
 */
macro_rules! validate_request_or_fail {
    ($request:expr) => {
        if !$request.is_valid() {
            return Ok($request.fail_syscall(nix::libc::EACCES));
        }
    };
}

macro_rules! syscall_handler {
    ($request:expr, $body:expr) => {{
        let inner = |request: &UNotifyEventRequest| -> Result<libseccomp::ScmpNotifResp, Errno> {
            let (req, proc) = request.prepare()?;
            validate_request_or_fail!(request);
            $body(req, &proc)
        };

        match inner($request) {
            Ok(result) => result,
            Err(error) => $request.fail_syscall(error as i32),
        }
    }};
}

macro_rules! remote_path_n {
    ($remote_process:expr, $request:expr, $n:expr, $event_request:expr) => {
        $remote_process.remote_path($request.data.args[$n] as usize, &$event_request)
    };
}

/// Seccomp sandbox profile export modes.
#[derive(Debug, Clone, Copy, Eq, PartialEq)]
pub enum ExportMode {
    /// Berkeley Packet Filter (binary, machine readable)
    BerkeleyPacketFilter,
    /// Pseudo Filter Code (text, human readable)
    PseudoFiltercode,
}

impl FromStr for ExportMode {
    type Err = anyhow::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "bpf" => Ok(Self::BerkeleyPacketFilter),
            "pfc" => Ok(Self::PseudoFiltercode),
            _ => bail!("Invalid export mode, expected one of bpf, or pfc!"),
        }
    }
}

/// `SyscallPathArgument` represents a system call path argument,
/// coupled with a directory file descriptor as necessary.
pub(crate) struct SyscallPathArgument {
    /// DirFd index in syscall args, if applicable.
    pub dirfd: Option<usize>,
    /// Path index in syscall args, if applicable.
    pub path: Option<usize>,
    /// Whether if it's ok for the path to be a NULL pointer.
    pub null: bool,
    /// Whether if it's ok for the path to be empty.
    pub empty: bool,
    /// Whether symbolic links should be resolved.
    pub resolve: bool,
    /// Whether dot as final component must return the given `Errno`.
    pub dotlast: Option<Errno>,
}

impl Serialize for SyscallPathArgument {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let mut map = serializer.serialize_map(Some(2))?;
        map.serialize_entry("dirfd", &self.dirfd)?;
        map.serialize_entry("path", &self.path)?;
        map.serialize_entry("null", &self.null)?;
        map.serialize_entry("empty", &self.empty)?;
        map.serialize_entry("resolve", &self.resolve)?;
        map.serialize_entry("dotlast", &self.dotlast.map(|e| e.to_string()))?;
        map.end()
    }
}

/// `SocketPair` is used to copy fd from child to parent
/// with sendmsg/recvmsg and `SCM_RIGHTS`
#[derive(Debug)]
pub(crate) struct SocketPair {
    /// child fd
    pub sender: RawFd,
    /// parent fd
    receiver: RawFd,
}

/// This struct represents the `Sender` end of a `SocketPair`.
pub(crate) struct Sender {
    /// The inner raw file descriptor
    pub fd: RawFd,
}

/// This struct represents the `Sender` end of a `SocketPair`.
pub(crate) struct Receiver {
    /// The inner raw file descriptor
    pub fd: RawFd,
}

impl SocketPair {
    pub(crate) fn init() -> Result<Self, io::Error> {
        let pairs = socketpair(
            AddressFamily::Unix,
            SockType::Stream,
            None,
            SockFlag::SOCK_CLOEXEC,
        )?;
        Ok(SocketPair {
            sender: pairs.0.as_raw_fd(),
            receiver: pairs.1.as_raw_fd(),
        })
    }

    pub(crate) fn channel(&self) -> (Sender, Receiver) {
        (Sender { fd: self.sender }, Receiver { fd: self.receiver })
    }
}

impl Drop for SocketPair {
    fn drop(&mut self) {
        let _ = close(self.sender);
        let _ = close(self.receiver);
    }
}

impl Sender {
    // nix::sys::socket::sendmsg allocates when cmsgs is not empty
    // which is not a good idea inside pre_exec
    // ref: nix's sendmsg implementation (MIT license)
    // (https://github.com/nix-rust/nix/blob/c6f9e2332efcf62c751d7a0174bb791e732b90a8/src/sys/socket/mod.rs#L1474)
    pub(crate) fn sendfd(&self, fd: RawFd) -> IOResult<()> {
        #[allow(clippy::cast_possible_truncation)]
        const FD_SIZE: c_uint = size_of::<RawFd>() as c_uint;
        const CAPACITY: u32 = unsafe { CMSG_SPACE(FD_SIZE) };
        const ALIGNMENT: usize = mem::align_of::<cmsghdr>();
        let buf = [0u8; ALIGNMENT + CAPACITY as usize];
        let (_prefix, aligned_buf, _suffix) = unsafe { buf.align_to::<cmsghdr>() };
        #[allow(clippy::as_ptr_cast_mut)]
        let cmsg_ptr = aligned_buf.as_ptr() as *mut c_void;
        let mut _binding = [0; 1];
        let mut _iov_buffer = [IoSliceMut::new(&mut _binding); 1];

        let mhdr = unsafe {
            // Musl's msghdr has private fields, so this is the only way to
            // initialize it.
            let mut mhdr = mem::MaybeUninit::<msghdr>::zeroed();
            let p = mhdr.as_mut_ptr();
            (*p).msg_name = ptr::null::<()>() as *mut _;
            (*p).msg_namelen = 0;
            // transmute iov into a mutable pointer.  sendmsg doesn't really mutate
            // the buffer, but the standard says that it takes a mutable pointer
            (*p).msg_iov = _iov_buffer.as_ref().as_ptr() as *mut _;
            (*p).msg_iovlen = 1;
            (*p).msg_control = cmsg_ptr;
            #[cfg(not(target_env = "musl"))]
            {
                (*p).msg_controllen = CAPACITY as usize;
            }
            #[cfg(target_env = "musl")]
            {
                (*p).msg_controllen = CAPACITY as u32;
            }
            (*p).msg_flags = 0;
            mhdr.assume_init()
        };

        let pmhdr: *mut cmsghdr = unsafe { CMSG_FIRSTHDR(&mhdr) };

        unsafe {
            (*pmhdr).cmsg_level = libc::SOL_SOCKET;
            (*pmhdr).cmsg_type = libc::SCM_RIGHTS;
            #[cfg(not(target_env = "musl"))]
            {
                (*pmhdr).cmsg_len = CMSG_LEN(FD_SIZE) as usize;
            }
            #[cfg(target_env = "musl")]
            {
                (*pmhdr).cmsg_len = CMSG_LEN(FD_SIZE) as u32;
            }
            ptr::copy_nonoverlapping(
                &[fd] as *const _ as *const u8,
                CMSG_DATA(pmhdr),
                FD_SIZE as usize,
            )
        }
        let ret = unsafe { libc::sendmsg(self.fd, &mhdr, 0) };

        if ret < 0 {
            Err(io::Error::last_os_error())
        } else {
            Ok(())
        }
    }
}

impl Receiver {
    pub(crate) fn recvfd(&self) -> Result<RawFd, Errno> {
        let mut cmsg_buffer = cmsg_space!(RawFd);
        let mut _binding = [0; 1];
        let mut _iov_buffer = [IoSliceMut::new(&mut _binding); 1];
        let res = recvmsg::<()>(
            self.fd,
            &mut _iov_buffer,
            Some(&mut cmsg_buffer),
            MsgFlags::empty(),
        )
        .map_err(|error| Errno::from_i32(error as i32))?;
        for cmsg in res.cmsgs() {
            if let ControlMessageOwned::ScmRights(fds) = cmsg {
                return Ok(fds[0]);
            }
        }
        Err(Errno::EINVAL)
    }
}

/// `UNotifyEventRequest` is the type of parameter that user's function
/// would get.
#[derive(Debug)]
pub struct UNotifyEventRequest {
    request: libseccomp::ScmpNotifReq,
    notify_fd: RawFd,
    sandbox: Arc<RwLock<Sandbox>>,
    umask_lock: Arc<Mutex<()>>,
}

type RequestWithProcess<'a> = (&'a ScmpNotifReq, RemoteProcess);

impl UNotifyEventRequest {
    fn new(
        request: libseccomp::ScmpNotifReq,
        notify_fd: RawFd,
        sandbox: Arc<RwLock<Sandbox>>,
        umask_lock: Arc<Mutex<()>>,
    ) -> Self {
        UNotifyEventRequest {
            request,
            notify_fd,
            sandbox,
            umask_lock,
        }
    }

    /// Given a `UNotifyEventRequest` return the seccomp request and a corresponding `RemoteProcess` instance.
    pub fn prepare(&self) -> Result<RequestWithProcess, Errno> {
        let req = self.get_request();
        #[allow(clippy::cast_possible_wrap)]
        let pid = Pid::from_raw(req.pid as i32);
        let proc = RemoteProcess::new(pid)?;
        Ok((req, proc))
    }

    /// Returns the unotify request (`libseccomp::ScmpNotifReq`) of
    /// this event.
    pub fn get_request(&self) -> &libseccomp::ScmpNotifReq {
        &self.request
    }

    /// Returns the internal `Sandbox` object locking it as necessary,
    /// and wrapped in a `SandboxGuard`.
    pub fn get_sandbox(&self, write: bool) -> SandboxGuard {
        if write {
            SandboxGuard::Write(self.sandbox.write())
        } else {
            SandboxGuard::Read(self.sandbox.read())
        }
    }

    /// Locks the umask lock and returns a `MutexGuard`,
    /// which when dropped releases the lock.
    pub fn umask_lock(&self) -> MutexGuard<()> {
        self.umask_lock.lock()
    }

    /// Let the kernel continue the syscall.
    ///
    /// # Safety
    /// CAUTION! This method is unsafe because it may suffer TOCTOU attack.
    /// Please read `seccomp_unotify(2)` "NOTES/Design goals; use of `SECCOMP_USER_NOTIF_FLAG_CONTINUE`"
    /// before using this method.
    pub unsafe fn continue_syscall(&self) -> libseccomp::ScmpNotifResp {
        libseccomp::ScmpNotifResp::new(self.request.id, 0, 0, ScmpNotifRespFlags::CONTINUE.bits())
    }

    /// Returns error to supervised process.
    /// `err` parameter should be a number larger than 0.
    pub fn fail_syscall(&self, err: i32) -> libseccomp::ScmpNotifResp {
        debug_assert!(err > 0);
        #[allow(clippy::arithmetic_side_effects)]
        libseccomp::ScmpNotifResp::new(self.request.id, 0, -err, 0)
    }

    /// Returns value to supervised process.
    pub fn return_syscall(&self, val: i64) -> libseccomp::ScmpNotifResp {
        libseccomp::ScmpNotifResp::new(self.request.id, val, 0, 0)
    }

    /// Check if this event is still valid.
    /// In some cases this is necessary, please check `seccomp_unotify(2)` for more information.
    pub fn is_valid(&self) -> bool {
        libseccomp::notify_id_valid(self.notify_fd, self.request.id).is_ok()
    }

    /// Add a file descriptor to the supervised process.
    /// This could help avoid TOCTOU attack in some cases.
    pub fn add_fd(&self, src_fd: RawFd, close_on_exec: bool) -> Result<RawFd, Errno> {
        let newfd_flags = if close_on_exec {
            nix::libc::O_CLOEXEC as u32
        } else {
            0
        };
        let addfd: libseccomp_sys::seccomp_notif_addfd = libseccomp_sys::seccomp_notif_addfd {
            id: self.request.id,
            flags: 0,
            srcfd: src_fd as u32,
            newfd: 0,
            newfd_flags,
        };

        let new_fd = unsafe {
            ioctl(
                self.notify_fd,
                SECCOMP_IOCTL_NOTIF_ADDFD,
                std::ptr::addr_of!(addfd),
            )
        };
        if new_fd < 0 {
            Err(Errno::last())
        } else {
            Ok(new_fd as RawFd)
        }
    }
}

/// By using `RemoteProcess`, you can get some information about the supervised process.
#[derive(Debug)]
pub struct RemoteProcess {
    pid: Pid,
    fd: RawFd,
}

impl RemoteProcess {
    /// Create a `RemoteProcess` object from a `Pid`.
    ///
    /// # Examples
    ///
    /// ```ignore
    /// let remote = RemoteProcess::new(Pid::from_raw(req.request.pid as i32)).unwrap();
    /// ```
    pub fn new(pid: Pid) -> Result<Self, Errno> {
        // get TGID of given pid (TID)
        let tgid = proc_tgid(pid)?;
        let fd = unsafe { libc::syscall(libc::SYS_pidfd_open, tgid, 0) };
        if fd < 0 {
            return Err(Errno::last());
        }

        Ok(RemoteProcess {
            pid: tgid,
            fd: fd as RawFd,
        })
    }

    /// Get file descriptor from remote process with `pidfd_getfd()`.
    /// This function requires Linux 5.6+.
    pub fn get_fd(&self, remote_fd: RawFd) -> Result<RawFd, Errno> {
        let local_fd = unsafe { libc::syscall(libc::SYS_pidfd_getfd, self.fd, remote_fd, 0) };
        if local_fd < 0 {
            Err(Errno::last())
        } else {
            Ok(local_fd as RawFd)
        }
    }

    /// Get process ID of the process
    pub fn get_pid(&self) -> Pid {
        self.pid
    }

    /// Read data from remote process's memory with `process_vm_readv()`.
    pub fn read_mem(
        &self,
        local_buffer: &mut [u8],
        remote_addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<usize, Errno> {
        static FORCE_PROC: Lazy<bool> =
            Lazy::new(|| std::env::var(ENV_NO_CROSS_MEMORY_ATTACH).is_ok());
        if *FORCE_PROC {
            return self.read_mem_proc(local_buffer, remote_addr, request);
        }

        if remote_addr == 0 {
            return Err(Errno::EFAULT);
        } else if !request.is_valid() {
            return Err(Errno::ESRCH);
        }
        let len = local_buffer.len();
        match process_vm_readv(
            self.pid,
            &mut [IoSliceMut::new(local_buffer)],
            &[RemoteIoVec {
                len,
                base: remote_addr,
            }],
        ) {
            Ok(n) => {
                if !request.is_valid() {
                    Err(Errno::ESRCH)
                } else {
                    Ok(n)
                }
            }
            Err(Errno::ENOSYS | Errno::EPERM) => {
                self.read_mem_proc(local_buffer, remote_addr, request)
            }
            Err(e) => Err(e),
        }
    }

    /// Write data to remote process's memory with `process_vm_writev()`.
    pub fn write_mem(
        &self,
        local_buffer: &[u8],
        remote_addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<usize, Errno> {
        static FORCE_PROC: Lazy<bool> =
            Lazy::new(|| std::env::var(ENV_NO_CROSS_MEMORY_ATTACH).is_ok());
        if *FORCE_PROC {
            return self.write_mem_proc(local_buffer, remote_addr, request);
        }

        if remote_addr == 0 {
            return Err(Errno::EFAULT);
        } else if !request.is_valid() {
            return Err(Errno::ESRCH);
        }
        let len = local_buffer.len();
        match process_vm_writev(
            self.pid,
            &[IoSlice::new(local_buffer)],
            &[RemoteIoVec {
                len,
                base: remote_addr,
            }],
        ) {
            Err(Errno::ENOSYS | Errno::EPERM) => {
                self.write_mem_proc(local_buffer, remote_addr, request)
            }
            Ok(n) => Ok(n),
            Err(e) => Err(e),
        }
    }

    /// Fallback method to read data from `/proc/$pid/mem` when `process_vm_readv()` is unavailable.
    pub fn read_mem_proc(
        &self,
        local_buffer: &mut [u8],
        remote_addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<usize, Errno> {
        if remote_addr == 0 {
            return Err(Errno::EFAULT);
        }
        let path = format!("/proc/{}/mem", self.pid);
        let mut file = File::open(path).map_err(|_| Errno::last())?;
        if !request.is_valid() {
            return Err(Errno::ESRCH);
        }
        file.seek(SeekFrom::Start(remote_addr as u64))
            .map_err(|_| Errno::last())?;
        let nread = file.read(local_buffer).map_err(|_| Errno::last())?;
        if request.is_valid() {
            Ok(nread)
        } else {
            Err(Errno::ESRCH)
        }
    }

    /// Fallback method to write data to `/proc/$pid/mem` when `process_vm_writev()` is unavailable.
    fn write_mem_proc(
        &self,
        local_buffer: &[u8],
        remote_addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<usize, Errno> {
        if remote_addr == 0 {
            return Err(Errno::EFAULT);
        }
        let path = format!("/proc/{}/mem", self.pid);
        let mut file = OpenOptions::new()
            .write(true)
            .open(path)
            .map_err(|_| Errno::last())?;
        if !request.is_valid() {
            return Err(Errno::ESRCH);
        }
        file.seek(SeekFrom::Start(remote_addr as u64))
            .map_err(|_| Errno::last())?;
        file.write(local_buffer).map_err(|_| Errno::last())
    }

    /// Read path from the given system call argument with the given request,
    /// returning together the optional raw system call argument.
    #[allow(clippy::type_complexity)]
    fn read_path_opt(
        &self,
        request: &UNotifyEventRequest,
        arg: &SyscallPathArgument,
    ) -> Result<(PathBuf, Option<PathBuf>), Errno> {
        let req = request.get_request();
        let path = match arg.path {
            Some(idx) => {
                if req.data.args[idx] == 0 {
                    if arg.null {
                        // NULL is ok, use fd
                        None
                    } else {
                        return Err(Errno::EFAULT);
                    }
                } else {
                    Some(remote_path_n!(self, req, idx, request)?)
                }
            }
            None => None,
        };
        let (path, path_raw) = if let Some(ref path) = path {
            if !arg.empty && path.is_empty() {
                return Err(Errno::ENOENT);
            }
            let (ends_with_dot, is_dot) = match path.to_bytes() {
                b"." | b".." => (true, true),
                _ => (crate::fs::path_ends_with_dot(path), false),
            };
            if ends_with_dot {
                if let Some(errno) = arg.dotlast {
                    // rmdir(.) = EINVAL, mkdir(.) = EEXIST.
                    return Err(if is_dot && errno == Errno::ENOENT {
                        Errno::EEXIST
                    } else {
                        errno
                    });
                }
            }
            let path = OsStr::from_bytes(path.to_bytes());
            let path_raw = Some(PathBuf::from(path));
            let path = normalize_path(path);
            #[allow(clippy::cast_possible_wrap)]
            let path = replace_self_with_pid(path, Pid::from_raw(req.pid as i32));
            if path.is_absolute() {
                (path, path_raw)
            } else if path.is_empty() && !ends_with_dot {
                // SAFETY: The ends_with_dot check above
                // ensures we return ENOTDIR when e.g. path is
                // a dot and the file descriptor argument is a
                // regular file. This happens because in this
                // case, joining the directory with an empty
                // path on the next branch essentially adds a
                // trailing slash to the path, making the
                // system call emulator fail with ENOTDIR if
                // the argument is not a directory. This way,
                // we avoid stat'ing the path here to
                // determine whether it's a directory or not.
                (self.read_directory(request, arg)?, path_raw)
            } else {
                (self.read_directory(request, arg)?.join(&path), path_raw)
            }
        } else {
            (self.read_directory(request, arg)?, None)
        };

        // SAFETY: For system calls that don't resolve
        // symbolic links we resolve symbolic links
        // in all parts of the path except last to
        // avoid TOCTOU between check and handle.
        // This is handled transparently by the
        // syd::fs::canonicalize function.
        crate::fs::canonicalize(path, arg.resolve, crate::fs::MissingHandling::Missing)
            .map(|p| (p, path_raw))
    }

    /// Read path from the given system call argument with the given request.
    fn read_path(
        &self,
        request: &UNotifyEventRequest,
        arg: &SyscallPathArgument,
    ) -> Result<PathBuf, Errno> {
        let req = request.get_request();
        let path = match arg.path {
            Some(idx) => {
                if req.data.args[idx] == 0 {
                    if arg.null {
                        // NULL is ok, use fd
                        None
                    } else {
                        return Err(Errno::EFAULT);
                    }
                } else {
                    Some(remote_path_n!(self, req, idx, request)?)
                }
            }
            None => None,
        };
        let path = if let Some(ref path) = path {
            if !arg.empty && path.is_empty() {
                return Err(Errno::ENOENT);
            }
            let (ends_with_dot, is_dot) = match path.to_bytes() {
                b"." | b".." => (true, true),
                _ => (crate::fs::path_ends_with_dot(path), false),
            };
            if ends_with_dot {
                if let Some(errno) = arg.dotlast {
                    // rmdir(.) = EINVAL, mkdir(.) = EEXIST.
                    return Err(if is_dot && errno == Errno::ENOENT {
                        Errno::EEXIST
                    } else {
                        errno
                    });
                }
            }
            let path = OsStr::from_bytes(path.to_bytes());
            let path = normalize_path(path);
            #[allow(clippy::cast_possible_wrap)]
            let path = replace_self_with_pid(path, Pid::from_raw(req.pid as i32));
            if path.is_absolute() {
                path
            } else if path.is_empty() && !ends_with_dot {
                // SAFETY: The ends_with_dot check above
                // ensures we return ENOTDIR when e.g. path is
                // a dot and the file descriptor argument is a
                // regular file. This happens because in this
                // case, joining the directory with an empty
                // path on the next branch essentially adds a
                // trailing slash to the path, making the
                // system call emulator fail with ENOTDIR if
                // the argument is not a directory. This way,
                // we avoid stat'ing the path here to
                // determine whether it's a directory or not.
                self.read_directory(request, arg)?
            } else if path == PathBuf::from(".") {
                // SAFETY: Add a trailing slash to the directory
                // to assert it must be a directory.
                self.read_directory(request, arg)?.join("")
            } else {
                self.read_directory(request, arg)?.join(&path)
            }
        } else {
            self.read_directory(request, arg)?
        };

        // SAFETY: For system calls that don't resolve
        // symbolic links we resolve symbolic links
        // in all parts of the path except last to
        // avoid TOCTOU between check and handle.
        // This is handled transparently by the
        // syd::fs::canonicalize function.
        crate::fs::canonicalize(path, arg.resolve, crate::fs::MissingHandling::Missing)
    }

    /// Read directory from the given system call argument with the given request.
    fn read_directory(
        &self,
        request: &UNotifyEventRequest,
        arg: &SyscallPathArgument,
    ) -> Result<PathBuf, Errno> {
        let req = request.get_request();
        let mut is_dirfd: bool = false;
        let dir = match arg.dirfd {
            Some(idx) => {
                let fd = Self::remote_fd(req, idx);
                if let Some(fd) = fd {
                    is_dirfd = true;
                    if fd < 0 {
                        return Err(Errno::EBADF);
                    }
                }
                let dir = Self::remote_dirfd(req.pid, fd);
                if is_dirfd
                    && faccessat(None, &dir, AccessFlags::F_OK, AtFlags::AT_SYMLINK_NOFOLLOW)
                        == Err(Errno::ENOENT)
                {
                    return Err(Errno::EBADF);
                }
                dir
            }
            None => Self::remote_dirfd(req.pid, None),
        };
        match crate::fs::canonicalize(dir, true, crate::fs::MissingHandling::Missing) {
            Ok(dir) => {
                if !request.is_valid() {
                    Err(Errno::ESRCH)
                } else {
                    Ok(dir)
                }
            }
            Err(error) => {
                warn!("ctx": "read_path", "arg": arg, "dir": arg.dirfd.unwrap_or(-nix::libc::AT_FDCWD as usize), "err": error.to_string(), "pid": req.pid);
                Err(if is_dirfd { Errno::EBADF } else { error })
            }
        }
    }

    /// Convert the file descriptor argument into a `RawFd`.
    /// If the argument refers to the current working directory,
    /// this function returns `None`.
    pub fn remote_fd(req: &ScmpNotifReq, idx: usize) -> Option<RawFd> {
        #[allow(clippy::cast_possible_truncation)]
        let fd = req.data.args[idx] as i32;
        if fd == nix::libc::AT_FDCWD {
            None
        } else {
            Some(fd as RawFd)
        }
    }

    /// Returns the file descriptor or current working directory path for the given `Pid`.
    pub fn remote_dirfd(pid: u32, fd: Option<RawFd>) -> PathBuf {
        if let Some(fd) = fd {
            PathBuf::from(format!("/proc/{pid}/fd/{fd}"))
        } else {
            PathBuf::from(format!("/proc/{pid}/cwd"))
        }
    }

    /// Read the `nix::libc::open_how` struct from process memory
    /// at the given address and size.
    pub fn remote_ohow(
        &self,
        addr: usize,
        size: usize,
        request: &UNotifyEventRequest,
    ) -> Result<nix::libc::open_how, Errno> {
        if size != std::mem::size_of::<nix::libc::open_how>() {
            return Err(Errno::EINVAL);
        }

        let mut buf = [0u8; std::mem::size_of::<nix::libc::open_how>()];
        self.read_mem(&mut buf, addr, request)?;

        // SAFETY: The following unsafe block assumes that:
        // 1. The memory layout of open_how in our Rust environment matches that of the target process.
        // 2. The proc.read_mem call has populated buf with valid data of the appropriate size (ensured by the size check above).
        // 3. The buffer is appropriately aligned for reading an open_how struct. If the remote process's representation of open_how
        //    was correctly aligned, our local buffer should be too, since it's an array on the stack.
        Ok(unsafe { std::ptr::read_unaligned(buf.as_ptr() as *const _) })
    }

    /// Read the `nix::libc::utimbuf` struct from process memory at the given address.
    /// Convert it to a `nix::libc::timespec[2]` for easy interoperability.
    fn remote_utimbuf(
        &self,
        addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<Option<[nix::libc::timespec; 2]>, Errno> {
        if addr == 0 {
            return Ok(None);
        }

        let mut buf = [0u8; std::mem::size_of::<nix::libc::utimbuf>()];
        self.read_mem(&mut buf, addr, request)?;

        // SAFETY: The following unsafe block assumes that:
        // 1. The memory layout of utimbuf in our Rust environment matches that of the target process.
        // 2. The proc.read_mem call has populated buf with valid data of the appropriate size (ensured by the size check above).
        // 3. The buffer is appropriately aligned for reading a utimbuf struct. If the remote process's representation of utimbuf
        //    was correctly aligned, our local buffer should be too, since it's an array on the stack.
        let t: nix::libc::utimbuf = unsafe { std::ptr::read_unaligned(buf.as_ptr() as *const _) };
        Ok(Some([
            nix::libc::timespec {
                tv_sec: t.actime,
                tv_nsec: 0,
            },
            nix::libc::timespec {
                tv_sec: t.modtime,
                tv_nsec: 0,
            },
        ]))
    }

    /// Read the `nix::libc::timeval[2]` struct from process memory at the given address.
    /// Convert it to a `nix::libc::timespec[2]` for easy interoperability.
    fn remote_timeval(
        &self,
        addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<Option<[nix::libc::timespec; 2]>, Errno> {
        if addr == 0 {
            return Ok(None);
        }

        let mut buf = [0u8; std::mem::size_of::<nix::libc::timeval>() * 2];
        self.read_mem(&mut buf, addr, request)?;

        // SAFETY: The following unsafe block assumes that:
        // 1. The memory layout of timeval in our Rust environment matches that of the target process.
        // 2. The proc.read_mem call has populated buf with valid data of the appropriate size (ensured by the size check above).
        // 3. The buffer is appropriately aligned for reading a timeval struct. If the remote process's representation of timeval
        //    was correctly aligned, our local buffer should be too, since it's an array on the stack.
        let t: [nix::libc::timeval; 2] =
            unsafe { std::ptr::read_unaligned(buf.as_ptr() as *const _) };
        Ok(Some([
            nix::libc::timespec {
                tv_sec: t[0].tv_sec,
                tv_nsec: (t[0].tv_usec as i64).saturating_mul(1_000), /* ms->ns */
            },
            nix::libc::timespec {
                tv_sec: t[1].tv_sec,
                tv_nsec: (t[1].tv_usec as i64).saturating_mul(1_000), /* ms->ns */
            },
        ]))
    }

    /// Read the `nix::libc::timespec[2]` struct from process memory at the given address.
    fn remote_timespec(
        &self,
        addr: usize,
        request: &UNotifyEventRequest,
    ) -> Result<Option<[nix::libc::timespec; 2]>, Errno> {
        if addr == 0 {
            return Ok(None);
        }

        let mut buf = [0u8; std::mem::size_of::<nix::libc::timespec>() * 2];
        self.read_mem(&mut buf, addr, request)?;

        // SAFETY: The following unsafe block assumes that:
        // 1. The memory layout of timespec in our Rust environment matches that of the target process.
        // 2. The proc.read_mem call has populated buf with valid data of the appropriate size (ensured by the size check above).
        // 3. The buffer is appropriately aligned for reading a timespec struct. If the remote process's representation of timespec
        //    was correctly aligned, our local buffer should be too, since it's an array on the stack.
        Ok(Some(unsafe {
            std::ptr::read_unaligned(buf.as_ptr() as *const _)
        }))
    }

    /// Read the path from memory of the process with the given `Pid` with the given address.
    fn remote_path(&self, addr: usize, request: &UNotifyEventRequest) -> Result<Box<CStr>, Errno> {
        let mut buf = [0u8; nix::libc::PATH_MAX as usize];
        self.read_mem(&mut buf, addr, request)?;

        // SAFETY: If buffer has no null byte,
        // return ENAMETOOLONG as the path is too
        // long for us to handle.
        Ok(Box::from(
            CStr::from_bytes_until_nul(&buf)
                .map_err(|_| Errno::ENAMETOOLONG)?
                .to_owned(),
        ))
    }
}

impl Drop for RemoteProcess {
    fn drop(&mut self) {
        let _ = close(self.fd);
    }
}

type ChildHandle = JoinHandle<Result<(), Errno>>;
type Child = (Pid, ChildHandle, ThreadPool);
type UserHookFunc = Box<dyn Fn(&UNotifyEventRequest) -> libseccomp::ScmpNotifResp + Send + Sync>;

/// The main component of greenhook.
pub struct Supervisor {
    export: Option<ExportMode>,
    handlers: HashMap<ScmpSyscall, Arc<UserHookFunc>>,
    sysallow: HashSet<ScmpSyscall>,
    thread_pool: ThreadPool,
    sandbox: Arc<RwLock<Sandbox>>,
    umask_lock: Arc<Mutex<()>>,
}

impl Supervisor {
    /// Create a new `Supervisor` object. You can specify the number of threads in the thread pool.
    /// This function will also check your kernel version and show warning or return error if necessary.
    ///
    /// # Examples
    /// ```
    /// use syd::{hook::Supervisor, sandbox::Sandbox};
    /// let supervisor = Supervisor::new(Sandbox::default(), num_cpus::get(), None).unwrap();
    /// ```
    pub fn new(
        sandbox: Sandbox,
        thread_num: usize,
        export_mode: Option<ExportMode>,
    ) -> Result<Self, io::Error> {
        if thread_num == 0 {
            return Err(io::Error::new(
                io::ErrorKind::InvalidInput,
                "thread_num should be greater than 0",
            ));
        }
        // detect kernel version and show warning
        let version = uname().map_err(|e| io::Error::from_raw_os_error(e as i32))?;
        let version = version.release();

        macro_rules! parse_error {
            () => {
                io::Error::new(io::ErrorKind::Other, "unknown version")
            };
        }

        let (major, minor) = {
            let mut iter = version.to_str().ok_or_else(|| parse_error!())?.split('.');
            let major = iter
                .next()
                .ok_or_else(|| parse_error!())?
                .parse::<u32>()
                .map_err(|_| parse_error!())?;
            let minor = iter
                .next()
                .ok_or_else(|| parse_error!())?
                .parse::<u32>()
                .map_err(|_| parse_error!())?;
            (major, minor)
        };
        if major < 5 {
            error!("ctx": "check_kernel", "err": "Your kernel version is too old.");
            return Err(io::Error::new(io::ErrorKind::Other, "kernel too old"));
        } else if major == 5 && minor < 5 {
            error!("ctx": "check_kernel", "err": "Your kernel version is too old (Does not support SECCOMP_USER_NOTIF_FLAG_CONTINUE, etc.).");
            return Err(io::Error::new(io::ErrorKind::Other, "kernel too old"));
        } else if major == 5 && minor < 6 {
            error!("ctx": "check_kernel", "err": "Your kernel version is too old (Does not support pidfd_getfd() and SECCOMP_IOCTL_NOTIF_ADDFD).");
            return Err(io::Error::new(io::ErrorKind::Other, "kernel too old"));
        } else if major == 5 && minor < 9 {
            error!("ctx": "check_kernel", "err": "Your kernel version is too old (Does not support SECCOMP_IOCTL_NOTIF_ADDFD).");
            return Err(io::Error::new(io::ErrorKind::Other, "kernel too old"));
        }
        Ok(Supervisor {
            export: export_mode,
            handlers: HashMap::new(),
            sysallow: HashSet::new(),
            thread_pool: ThreadPool::with_name(
                module_path!().to_string().replace('o', "☮"),
                thread_num,
            ),
            sandbox: Arc::new(RwLock::new(sandbox)),
            umask_lock: Arc::new(Mutex::new(())),
        })
    }

    /// Initialize the environment for the sandbox.
    /// Call this before `init`.
    pub fn init_env() -> anyhow::Result<()> {
        // SAFETY: The nix::sys::signal::signal function is unsafe because it affects the global state
        // of the program by changing how a signal (SIGPIPE in this case) is handled. It's safe to call
        // here because changing the SIGPIPE signal to its default behavior will not interfere with any
        // other part of this program that could be relying on a custom SIGPIPE signal handler.
        unsafe {
            nix::sys::signal::signal(
                nix::sys::signal::Signal::SIGPIPE,
                nix::sys::signal::SigHandler::SigDfl,
            )
        }
        .context("Failed to set SIGPIPE signal handler to default.")?;

        // Initialize logging.
        JsonLinesLogger::init().context("Failed to initialize logging.")?;

        // Set the no_new_privs attribute.
        // SAFETY: In libc we trust.
        if unsafe { nix::libc::prctl(nix::libc::PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0) } == -1 {
            let errno = Errno::last();
            bail!("Error setting no_new_privs attribute: {errno}");
        }

        Ok(())
    }

    /// Initilizes the supervisor by adding the system call handlers.
    pub fn init(&mut self) -> &mut Self {
        // network sandboxing
        self.insert_handler(ScmpSyscall::new("bind"), sys_bind);
        self.insert_handler(ScmpSyscall::new("connect"), sys_connect);
        self.insert_handler(ScmpSyscall::new("recvfrom"), sys_recvfrom);
        self.insert_handler(ScmpSyscall::new("sendto"), sys_sendto);

        // exec sandboxing
        self.insert_handler(ScmpSyscall::new("execve"), sys_execve);
        self.insert_handler(ScmpSyscall::new("execveat"), sys_execveat);

        // stat sandboxing
        self.insert_handler(ScmpSyscall::new("chdir"), sys_chdir);
        self.insert_handler(ScmpSyscall::new("fchdir"), sys_fchdir);
        self.insert_handler(ScmpSyscall::new("getdents"), sys_getdents);
        self.insert_handler(ScmpSyscall::new("getdents64"), sys_getdents);
        self.insert_handler(ScmpSyscall::new("stat"), sys_stat);
        self.insert_handler(ScmpSyscall::new("stat64"), sys_stat);
        self.insert_handler(ScmpSyscall::new("fstat"), sys_fstat);
        self.insert_handler(ScmpSyscall::new("fstat64"), sys_fstat);
        self.insert_handler(ScmpSyscall::new("lstat"), sys_lstat);
        self.insert_handler(ScmpSyscall::new("lstat64"), sys_lstat);
        self.insert_handler(ScmpSyscall::new("statx"), sys_statx);
        self.insert_handler(ScmpSyscall::new("fstatat64"), sys_newfstatat);
        self.insert_handler(ScmpSyscall::new("newfstatat"), sys_newfstatat);

        // read/write sandboxing
        self.insert_handler(ScmpSyscall::new("access"), sys_access);
        self.insert_handler(ScmpSyscall::new("faccessat"), sys_faccessat);
        self.insert_handler(ScmpSyscall::new("faccessat2"), sys_faccessat2);
        self.insert_handler(ScmpSyscall::new("chmod"), sys_chmod);
        self.insert_handler(ScmpSyscall::new("fchmod"), sys_fchmod);
        self.insert_handler(ScmpSyscall::new("fchmodat"), sys_fchmodat);
        self.insert_handler(ScmpSyscall::new("chown"), sys_chown);
        self.insert_handler(ScmpSyscall::new("fchown"), sys_fchown);
        self.insert_handler(ScmpSyscall::new("fchown32"), sys_fchown);
        self.insert_handler(ScmpSyscall::new("lchown"), sys_lchown);
        self.insert_handler(ScmpSyscall::new("lchown32"), sys_lchown);
        self.insert_handler(ScmpSyscall::new("fchownat"), sys_fchownat);
        self.insert_handler(ScmpSyscall::new("creat"), sys_creat);
        self.insert_handler(ScmpSyscall::new("link"), sys_link);
        self.insert_handler(ScmpSyscall::new("symlink"), sys_symlink);
        self.insert_handler(ScmpSyscall::new("unlink"), sys_unlink);
        self.insert_handler(ScmpSyscall::new("linkat"), sys_linkat);
        self.insert_handler(ScmpSyscall::new("symlinkat"), sys_symlinkat);
        self.insert_handler(ScmpSyscall::new("unlinkat"), sys_unlinkat);
        self.insert_handler(ScmpSyscall::new("mkdir"), sys_mkdir);
        self.insert_handler(ScmpSyscall::new("rmdir"), sys_rmdir);
        self.insert_handler(ScmpSyscall::new("mkdirat"), sys_mkdirat);
        self.insert_handler(ScmpSyscall::new("mknod"), sys_mknod);
        self.insert_handler(ScmpSyscall::new("mknodat"), sys_mknodat);
        self.insert_handler(ScmpSyscall::new("mount"), sys_mount);
        self.insert_handler(ScmpSyscall::new("umount"), sys_umount);
        self.insert_handler(ScmpSyscall::new("umount2"), sys_umount2);
        self.insert_handler(ScmpSyscall::new("open"), sys_open);
        self.insert_handler(ScmpSyscall::new("openat"), sys_openat);
        self.insert_handler(ScmpSyscall::new("openat2"), sys_openat2);
        self.insert_handler(ScmpSyscall::new("rename"), sys_rename);
        self.insert_handler(ScmpSyscall::new("renameat"), sys_renameat);
        self.insert_handler(ScmpSyscall::new("renameat2"), sys_renameat2);
        self.insert_handler(ScmpSyscall::new("utime"), sys_utime);
        self.insert_handler(ScmpSyscall::new("utimes"), sys_utimes);
        self.insert_handler(ScmpSyscall::new("futimesat"), sys_futimesat);
        self.insert_handler(ScmpSyscall::new("utimensat"), sys_utimensat);
        self.insert_handler(ScmpSyscall::new("truncate"), sys_truncate);
        self.insert_handler(ScmpSyscall::new("truncate64"), sys_truncate);
        self.insert_handler(ScmpSyscall::new("ftruncate"), sys_ftruncate);
        self.insert_handler(ScmpSyscall::new("getxattr"), sys_getxattr);
        self.insert_handler(ScmpSyscall::new("fgetxattr"), sys_fgetxattr);
        self.insert_handler(ScmpSyscall::new("lgetxattr"), sys_lgetxattr);
        self.insert_handler(ScmpSyscall::new("setxattr"), sys_setxattr);
        self.insert_handler(ScmpSyscall::new("fsetxattr"), sys_fsetxattr);
        self.insert_handler(ScmpSyscall::new("lsetxattr"), sys_lsetxattr);
        self.insert_handler(ScmpSyscall::new("listxattr"), sys_listxattr);
        self.insert_handler(ScmpSyscall::new("flistxattr"), sys_flistxattr);
        self.insert_handler(ScmpSyscall::new("llistxattr"), sys_llistxattr);
        self.insert_handler(ScmpSyscall::new("removexattr"), sys_removexattr);
        self.insert_handler(ScmpSyscall::new("fremovexattr"), sys_fremovexattr);
        self.insert_handler(ScmpSyscall::new("lremovexattr"), sys_lremovexattr);

        // Allowlist safe system calls.
        for sysname in SAFE_SYSCALLS.iter().take(SAFE_SYSCALLS_SIZE) {
            let syscall = ScmpSyscall::new(sysname);
            if i32::from(syscall) == libseccomp_sys::__NR_SCMP_ERROR {
                error!("ctx": "allowlist_safe_syscall", "err": "scmp_error", "sys": sysname);
                continue;
            }
            self.allow_syscall(syscall);
        }

        self
    }

    /// Insert this system call to the list of allowed system calls.
    /// No filtering is done one these system calls and they're allowed at the kernel level.
    pub fn allow_syscall(&mut self, syscall: ScmpSyscall) {
        self.sysallow.insert(syscall);
    }

    ///
    /// # Examples
    ///
    /// ```no_run
    /// use libseccomp::ScmpSyscall;
    /// use syd::{
    ///     hook::{Supervisor, UNotifyEventRequest},
    ///     sandbox::Sandbox,
    /// };
    ///
    /// fn close_handler(req: &UNotifyEventRequest) -> libseccomp::ScmpNotifResp {
    ///     println!("close");
    ///     unsafe { req.continue_syscall() }
    /// }
    ///
    /// let mut supervisor = Supervisor::new(Sandbox::default(), num_cpus::get(), None).unwrap();
    /// supervisor.insert_handler(ScmpSyscall::new("open"), |req| {
    ///     println!("open: {}", req.get_request().data.args[0]);
    ///     unsafe { req.continue_syscall() }
    /// });
    /// supervisor.insert_handler(ScmpSyscall::new("close"), close_handler);
    /// ```
    pub fn insert_handler(
        &mut self,
        syscall: ScmpSyscall,
        handler: impl Fn(&UNotifyEventRequest) -> libseccomp::ScmpNotifResp + Send + Sync + 'static,
    ) {
        self.handlers.insert(syscall, Arc::new(Box::new(handler)));
    }

    /// Run a command with seccomp filter.
    /// This method will fork a child process, do some preparations and run the command in it.
    /// It returns a `Pid`, a `JoinHandle` of supervising thread, and a `ThreadPool` handle of syscall user functions.
    /// It's recommended to use `Supervisor::wait()` to wait for the child process.
    pub fn spawn(self, command: &mut crate::unshare::Command) -> anyhow::Result<Child> {
        let seccomp_filter = self
            .setup_seccomp()
            .context("Failed to set up seccomp filters.")?;
        command
            .seccomp_filter(seccomp_filter)
            .context("Failed to setup socket pair.")?;

        // Spawn child under sandbox.
        let sandbox = command
            .spawn()
            .context("Failed to spawn sandbox process.")?;
        let pid = sandbox.id();
        #[allow(clippy::disallowed_methods)]
        let fd = sandbox.seccomp_fd.unwrap();

        // Ignore some signals to ensure uniterrupted tracing.
        let _ = crate::ignore_signal(SIGTSTP);
        let _ = crate::ignore_signal(SIGTTIN);
        let _ = crate::ignore_signal(SIGTTOU);
        let _ = crate::ignore_signal(SIGHUP);

        #[allow(clippy::cast_possible_wrap)]
        let pid = Pid::from_raw(pid as i32);
        let mut sandbox = self.sandbox.write();
        sandbox.set_child_pid(pid);
        drop(sandbox);

        // See the TODO item in setup_seccomp_parent function.
        // Apply seccomp strengtening for the Sydb☮x process itself.
        // Since we have already spawned the child and received the
        // seccomp notification file descriptor at this point, we can
        // restrict a lot of potential attack space.
        //#[allow(clippy::disallowed_methods)]
        //self.setup_seccomp_parent().unwrap().load().unwrap();

        self.supervise(pid, fd)
    }

    // Set up seccomp for the sandbox process.
    #[allow(clippy::cognitive_complexity)]
    fn setup_seccomp(&self) -> IOResult<ScmpFilterContext> {
        let mut ctx =
            ScmpFilterContext::new_filter(ScmpAction::Errno(nix::libc::EACCES)).map_err(|e| {
                io::Error::new(
                    io::ErrorKind::Other,
                    format!("failed to create seccomp filter: {}", e),
                )
            })?;
        // Enforce the NO_NEW_PRIVS functionality before
        // loading the seccomp filter into the kernel.
        let _ = ctx.set_ctl_nnp(true);
        // We don't want ECANCELED, we want actual errnos.
        let _ = ctx.set_api_sysrawrc(true);
        // We deny with EACCES for bad system call, and kill process for bad arch.
        let _ = ctx.set_act_badarch(ScmpAction::KillProcess);
        // We log all filter actions other than Allow to kernel log if the log level is <=Debug.
        if log_enabled!(Level::Debug) {
            let _ = ctx.set_ctl_log(true);
        }

        // Add architectures based on the current architecture
        if cfg!(target_arch = "x86_64") {
            seccomp_add_arch(&mut ctx, ScmpArch::X8664)?;
            seccomp_add_arch(&mut ctx, ScmpArch::X86)?;
            seccomp_add_arch(&mut ctx, ScmpArch::X32)?;
        } else if cfg!(target_arch = "x86") {
            seccomp_add_arch(&mut ctx, ScmpArch::X86)?;
        } else if cfg!(target_arch = "arm") {
            seccomp_add_arch(&mut ctx, ScmpArch::Arm)?;
        } else if cfg!(target_arch = "aarch64") {
            seccomp_add_arch(&mut ctx, ScmpArch::Aarch64)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Arm)?;
        } else if cfg!(target_arch = "mips") {
            seccomp_add_arch(&mut ctx, ScmpArch::Mips)?;
        } else if cfg!(target_arch = "mips64") {
            seccomp_add_arch(&mut ctx, ScmpArch::Mips64)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Mips64N32)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Mips)?;
        } else if cfg!(target_arch = "mips64n32") {
            seccomp_add_arch(&mut ctx, ScmpArch::Mips64N32)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Mips)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Mips64)?;
        } else if cfg!(target_arch = "mipsel") {
            seccomp_add_arch(&mut ctx, ScmpArch::Mipsel64)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Mipsel64N32)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Mipsel)?;
        } else if cfg!(target_arch = "mipsel64") {
            seccomp_add_arch(&mut ctx, ScmpArch::Mipsel64)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Mipsel)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Mipsel64N32)?;
        } else if cfg!(target_arch = "mipsel64n32") {
            seccomp_add_arch(&mut ctx, ScmpArch::Mipsel64N32)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Mipsel)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Mipsel64)?;
        } else if cfg!(target_arch = "powerpc") {
            seccomp_add_arch(&mut ctx, ScmpArch::Ppc)?;
        } else if cfg!(target_arch = "powerpc64") {
            seccomp_add_arch(&mut ctx, ScmpArch::Ppc64)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Ppc64Le)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Ppc)?;
        } else if cfg!(target_arch = "parisc") {
            seccomp_add_arch(&mut ctx, ScmpArch::Parisc)?;
        } else if cfg!(target_arch = "parisc64") {
            seccomp_add_arch(&mut ctx, ScmpArch::Parisc64)?;
            seccomp_add_arch(&mut ctx, ScmpArch::Parisc)?;
        } else if cfg!(target_arch = "riscv64") {
            seccomp_add_arch(&mut ctx, ScmpArch::Riscv64)?;
        } else if cfg!(target_arch = "s390x") {
            seccomp_add_arch(&mut ctx, ScmpArch::S390X)?;
            seccomp_add_arch(&mut ctx, ScmpArch::S390)?;
        }

        // Acquire the read lock to sandbox configuration.
        let sandbox = self.sandbox.read();
        let fake_root = sandbox.get_root();
        drop(sandbox);

        // Fakeroot
        let id_action = if fake_root {
            ScmpAction::Errno(0)
        } else {
            ScmpAction::Allow
        };
        for sysname in GET_ID_SYSCALLS {
            let syscall = ScmpSyscall::new(sysname);
            ctx.add_rule(id_action, syscall).map_err(|e| {
                io::Error::new(
                    io::ErrorKind::Other,
                    format!(
                        "failed to add rule for system call {} ( {} ): {}",
                        syscall, sysname, e
                    ),
                )
            })?;
        }

        // Setid restrictions
        for sysname in SET_ID_SYSCALLS {
            let syscall = ScmpSyscall::new(sysname);
            ctx.add_rule(ScmpAction::Errno(0), syscall).map_err(|e| {
                io::Error::new(
                    io::ErrorKind::Other,
                    format!(
                        "failed to add rule for system call {} ( {} ): {}",
                        syscall, sysname, e
                    ),
                )
            })?;
        }

        // Add notify rules for system calls with handlers.
        let syscall_notif: Vec<_> = self.handlers.keys().copied().collect();
        for syscall in &syscall_notif {
            // This is a verified handler, the name must exist, calling unwrap is OK.
            #[allow(clippy::disallowed_methods)]
            let name = syscall.get_name().unwrap();
            let _ = ctx.set_syscall_priority(*syscall, crate::syscall_priority(&name));
            ctx.add_rule(ScmpAction::Notify, *syscall).map_err(|e| {
                io::Error::new(
                    io::ErrorKind::Other,
                    format!(
                        "failed to add rule for system call {} ( {} ): {}",
                        *syscall,
                        ScmpSyscall::get_name(*syscall).unwrap_or("?".to_string()),
                        e
                    ),
                )
            })?;
        }

        // Add allow rules for system calls in the default allow list.
        let syscall_allow: Vec<_> = self.sysallow.iter().copied().collect();
        for syscall in &syscall_allow {
            ctx.add_rule(ScmpAction::Allow, *syscall).map_err(|e| {
                io::Error::new(
                    io::ErrorKind::Other,
                    format!(
                        "failed to add rule for system call {} ( {} ): {}",
                        *syscall,
                        ScmpSyscall::get_name(*syscall).unwrap_or("?".to_string()),
                        e
                    ),
                )
            })?;
        }

        // Apply sandbox restrictions as necessary.
        let sandbox = self.sandbox.read();
        let restrict_ioctl = !sandbox.allow_unsafe_ioctl();
        let restrict_prctl = !sandbox.allow_unsafe_prctl();
        let restrict_getrandom = !sandbox.allow_unsafe_getrandom();
        drop(sandbox);

        // Restriction 1: Allowlist known-safe ioctls.
        #[allow(clippy::disallowed_methods)]
        let sys_ioctl = ScmpSyscall::from_name("ioctl").unwrap();
        if restrict_ioctl {
            for (opt, _) in &ALLOWLIST_IOCTL {
                let cmp = ScmpArgCompare::new(1, ScmpCompareOp::Equal, *opt);
                ctx.add_rule_conditional(ScmpAction::Allow, sys_ioctl, &[cmp])
                    .map_err(|error| {
                        io::Error::new(io::ErrorKind::Other, format!("SeccompError: {error}"))
                    })?;
            }
        } else {
            ctx.add_rule(ScmpAction::Allow, sys_ioctl)
                .map_err(|error| {
                    io::Error::new(io::ErrorKind::Other, format!("SeccompError: {error}"))
                })?;
        }

        // Restriction 2: Allowlist known-safe prctls.
        // Things like PR_SET_MM, PR_SET_PTRACER, and PR_SET_SPECULATION_CTRL are left out.
        #[allow(clippy::disallowed_methods)]
        let sys_prctl = ScmpSyscall::from_name("prctl").unwrap();
        if restrict_prctl {
            for (opt, _) in &ALLOWLIST_PRCTL {
                let cmp = ScmpArgCompare::new(0, ScmpCompareOp::Equal, *opt);
                ctx.add_rule_conditional(ScmpAction::Allow, sys_prctl, &[cmp])
                    .map_err(|error| {
                        io::Error::new(io::ErrorKind::Other, format!("SeccompError: {error}"))
                    })?;
            }
        } else {
            ctx.add_rule(ScmpAction::Allow, sys_prctl)
                .map_err(|error| {
                    io::Error::new(io::ErrorKind::Other, format!("SeccompError: {error}"))
                })?;
        }

        // Restriction 3: Disallow getrandom accessing /dev/random.
        // SAFETY: We must ensure the sandbox process can't
        // drain the system entropy. Accessing /dev/urandom is fine.
        #[allow(clippy::disallowed_methods)]
        let sys_getrandom = ScmpSyscall::from_name("getrandom").unwrap();
        if restrict_getrandom {
            ctx.add_rule_conditional(
                ScmpAction::Allow,
                sys_getrandom,
                &[ScmpArgCompare::new(
                    2,
                    ScmpCompareOp::MaskedEqual(u64::from(nix::libc::GRND_RANDOM)),
                    0,
                )],
            )
            .map_err(|error| {
                io::Error::new(io::ErrorKind::Other, format!("SeccompError: {error}"))
            })?;
        } else {
            ctx.add_rule(ScmpAction::Allow, sys_getrandom)
                .map_err(|error| {
                    io::Error::new(io::ErrorKind::Other, format!("SeccompError: {error}"))
                })?;
        }

        // Export seccomp rules if requested.
        if let Some(mode) = self.export {
            self.seccomp_export(&ctx, mode)?;
        }

        Ok(ctx)
    }

    /*
     * Currently unused because of cmd/exec.
     * TODO: Remove cmd/exec, and implement this plus Landlock restrictions.
    /// Set up seccomp for the Sydb☮x process.
    /// This is important to restrict potential attack space in case
    /// Sydb☮x process is compromised anyhow.
    fn setup_seccomp_parent(&self) -> IOResult<ScmpFilterContext> {
        let mut ctx = ScmpFilterContext::new_filter(ScmpAction::KillThread).map_err(|e| {
            io::Error::new(
                io::ErrorKind::Other,
                format!("failed to create seccomp filter for Sydb☮x: {}", e),
            )
        })?;
        // We don't want ECANCELED, we want actual errnos.
        let _ = ctx.set_api_sysrawrc(true);
        // We kill for bad system cal and bad arch.
        let _ = ctx.set_act_badarch(ScmpAction::KillThread);
        // We log all filter actions other than Allow to kernel log.
        let _ = ctx.set_ctl_log(true);

        // Add allow rules for system calls in the Sydb☮x allow list.
        let syscall_allow: Vec<_> = vec![
            "brk",
            "clone",
            "clone3",
            "close",
            "execve", // required for cmd/exec
            "exit",
            "exit_group",
            "fcntl",
            "fork", // required for cmd/exec
            "futex",
            "getdents",   // required for stat sandboxing
            "getdents64", // ditto
            "getrandom",
            "getsockname", // required for allowlist/successful_bind
            "madvise",
            "mmap",
            "mprotect",
            "munmap",
            "open",
            "openat",
            "openat2",
            "poll",
            "read",
            "rseq",
            "rt_sigaction",
            "rt_sigprocmask",
            "sched_getaffinity",
            "seccomp",
            "set_robust_list",
            "set_tid_address",
            "sigaltstack",
            "uname",
            "wait4",
        ];
        for sysname in &syscall_allow {
            let syscall = ScmpSyscall::new(sysname);
            if let Err(error) = ctx.add_rule(ScmpAction::Allow, syscall) {
                eprintln!("Error adding Sydb☮x rule: {error}, continuing...");
            }
        }

        // Harden prctl.
        let syscall = ScmpSyscall::new("prctl");
        if let Err(error) = ctx.add_rule_conditional(
            ScmpAction::Allow,
            syscall,
            &[scmp_cmp!($arg0 == nix::libc::PR_SET_NAME as u64)],
        ) {
            eprintln!("Error hardening prctl for Sydb☮x: {error}, continuing...");
        }
        // Harden ioctl.
        let syscall = ScmpSyscall::new("ioctl");
        for ioctl_req in [
            SECCOMP_IOCTL_NOTIF_ID_VALID,
            SECCOMP_IOCTL_NOTIF_RECV,
            SECCOMP_IOCTL_NOTIF_SEND,
            SECCOMP_IOCTL_NOTIF_ADDFD,
        ] {
            if let Err(error) = ctx.add_rule_conditional(
                ScmpAction::Allow,
                syscall,
                &[scmp_cmp!($arg1 == ioctl_req as u64)],
            ) {
                eprintln!(
                    "Error hardening ioctl request {ioctl_req} for Sydb☮x: {error}, continuing..."
                );
            }
        }

        // Export seccomp rules if requested.
        // Only in PFC mode because with BPF mode the export of this and the child will be mangled.
        if let Some(ExportMode::PseudoFiltercode) = self.export {
            println!("# Sydb☮x parent rules");
            ctx.export_pfc(&mut io::stdout()).map_err(|error| {
                io::Error::new(
                    io::ErrorKind::Other,
                    format!("failed to export rules in pfc format: {error}"),
                )
            })?;
        }

        Ok(ctx)
    }
    */

    /// Export a seccomp context as bpf or pfc.
    fn seccomp_export(&self, ctx: &ScmpFilterContext, mode: ExportMode) -> IOResult<()> {
        match mode {
            ExportMode::BerkeleyPacketFilter => ctx.export_bpf(&mut io::stdout()),
            ExportMode::PseudoFiltercode => {
                #[allow(clippy::disallowed_methods)]
                let fd = match mkstemp("syd-tmp-XXXXXX") {
                    Ok((fd, path)) => {
                        unlink(path.as_path()).expect("Failed to delete temporary file.");
                        fd
                    }
                    Err(error) => {
                        return Err(error.into());
                    }
                };
                // SAFETY: In libc, we trust.
                let mut file = unsafe { File::from_raw_fd(fd) };

                if let Err(error) = ctx.export_pfc(&mut file) {
                    return Err(io::Error::new(io::ErrorKind::Other, format!("{error}")));
                }

                file.seek(SeekFrom::Start(0))?;
                let mut buffer = Vec::new();
                file.read_to_end(&mut buffer)?;

                let output = String::from_utf8_lossy(&buffer);
                let output = output.replace("0x7fc00000", "NOTIFY");

                #[allow(clippy::disallowed_methods)]
                let libver = ScmpVersion::current()
                    .expect("Failed to determine libseccomp library version.");
                println!(
                    "# Sydb☮x v{} seccomp rules generated by libsecc☮mp v{}.{}.{}",
                    env!("CARGO_PKG_VERSION"),
                    libver.major,
                    libver.minor,
                    libver.micro
                );
                println!("# API Version: {API_VERSION}");

                #[allow(clippy::disallowed_methods)]
                let mut syscall_allow: Vec<_> = self
                    .sysallow
                    .iter()
                    .copied()
                    .map(|sys| sys.get_name().unwrap())
                    .collect();
                #[allow(clippy::disallowed_methods)]
                let mut syscall_notif: Vec<_> = self
                    .handlers
                    .keys()
                    .copied()
                    .map(|sys| sys.get_name().unwrap())
                    .collect();
                syscall_allow.sort();
                syscall_notif.sort();

                println!("# System calls with Action=ALLOW: {}", syscall_allow.len());
                println!("# System calls with Action=NOTIF: {}", syscall_notif.len());

                let uidcall = GET_ID_SYSCALLS.to_vec().join(", ");
                let sandbox = self.sandbox.read();
                println!(
                    "# Fake Root: {} ( {uidcall} )",
                    if sandbox.get_root() { "yes" } else { "no" }
                );
                println!(
                    "{}",
                    sandbox
                        .to_string()
                        .lines()
                        .map(|line| format!("# {}", line))
                        .collect::<Vec<_>>()
                        .join("\n")
                );
                drop(sandbox);

                println!("# Action=NOTIF: {}", syscall_notif.len());
                for name in &syscall_notif {
                    println!("#    - {name}");
                }
                println!("# Action=ALLOW: {}", syscall_allow.len());
                for name in &syscall_allow {
                    println!("#    - {name}");
                }
                print!("{output}");
                Ok(())
            }
        }
        .map_err(|error| {
            io::Error::new(
                io::ErrorKind::Other,
                format!("failed to export rules in {mode:?} format: {error}"),
            )
        })
    }

    /// Logic for the supervise child thread.
    fn supervise(self, pid: Pid, fd: RawFd) -> anyhow::Result<Child> {
        let pool_handle = self.thread_pool.clone();
        let thread_handle = Builder::new().name("syd::p☮ll".into()).spawn(move || {
            // 0. Wrap fd in an OwnedFd to ensure it gets closed properly.
            // SAFETY: We're passing a valid fd to OwnedFd.
            let _guard_fd = unsafe { OwnedFd::from_raw_fd(fd) };

            // 1. Create an epoll instance
            let epfd = epoll_create1(EpollCreateFlags::EPOLL_CLOEXEC)?;

            // 2. Wrap epoll fd in an OwnedFd to ensure it gets closed properly.
            // SAFETY: We're passing a valid fd to OwnedFd.
            let _guard_epfd = unsafe { OwnedFd::from_raw_fd(epfd) };

            // 3. Add the file descriptor to the epoll instance
            let mut event = EpollEvent::new(EpollFlags::EPOLLIN, fd as u64);
            epoll_ctl(epfd, EpollOp::EpollCtlAdd, fd, Some(&mut event))?;

            // 4. Wait for events
            loop {
                let mut events = [EpollEvent::empty(); 1];
                let nfds = match epoll_wait(epfd, &mut events, -1) {
                    Ok(nfds) => nfds,
                    Err(Errno::EINTR) => {
                        // An epoll_wait call can - in absence of any
                        // user defined signal handlers - fail with
                        // EINTR on SIGTRAP (eg, attaching strace to a
                        // running process) or SIGCONT (continuing a
                        // stopped process):
                        // https://lwn.net/Articles/851813/
                        continue;
                    },
                    Err(error) => return Err(error),
                };

                if nfds == 0 {
                    error!("ctx": "epoll", "error": "no file descriptors returned from epoll_wait");
                    return Err(Errno::EINVAL);
                } else if events[0].events().contains(EpollFlags::EPOLLHUP) {
                    debug!("ctx": "epoll", "error": "unexpected epoll hung up event", "events": format!("{:?}", events[0].events()));
                    return Err(Errno::EINVAL);
                } else if !events[0].events().contains(EpollFlags::EPOLLIN) {
                    error!("ctx": "epoll", "error": "unexpected epoll event", "events": format!("{:?}", events[0].events()));
                    return Err(Errno::EINVAL);
                }

                match ScmpNotifReq::receive(fd) {
                    Ok(req) => {
                        if !self.syscall(fd, req) {
                            // Handling system call failed, break.
                            return Ok(())
                        }
                    }
                    Err(error) => {
                        // SAFETY: We need to continue with epoll after
                        // any error as otherwise the sandbox process will
                        // fail all syscalls with ENOSYS or will just deadlock.
                        match Errno::last() {
                            Errno::ENOENT => {
                                // ENOENT: The task was interrupted by a fatal
                                // signal between the time we get the poll event
                                // and when we attempted to receive the notification.
                            },
                            errno => {
                                error!("ctx": "seccomp_notif_recv", "error": error.to_string(), "errno": errno.to_string());
                            }
                        }
                    }
                }
            }
        }).context("Failed to spawn the poller thread.")?;

        Ok((pid, thread_handle, pool_handle))
    }

    fn syscall(&self, fd: RawFd, req: ScmpNotifReq) -> bool {
        let event_req = UNotifyEventRequest::new(
            req,
            fd,
            Arc::clone(&self.sandbox),
            Arc::clone(&self.umask_lock),
        );
        let syscall_id = event_req.get_request().data.syscall;

        match self.handlers.get(&syscall_id) {
            Some(handler) => {
                let handler_in_thread = Arc::clone(handler);
                self.thread_pool.execute(move || {
                    let response = handler_in_thread(&event_req);

                    if event_req.is_valid() {
                        if let Err(error) = response.respond(fd) {
                            let errno = Errno::last();
                            error!("ctx": "spawn_response", "err": format!("failed to send response: {error}, {errno}"));
                        }
                    } else {
                        info!("ctx": "spawn_response", "err": "failed to send response: invalid request");
                    }
                });
                true
            }
            None => {
                error!("ctx": "syscall_handle", "err": format!("got unknown syscall to handle: {syscall_id}"));
                match event_req.fail_syscall(libc::ENOSYS).respond(fd) {
                    Ok(_) => true,
                    Err(error) => {
                        let errno = Errno::last();
                        error!("ctx": "syscall_handle", "err": format!("failed to send response: {error}, {errno}"));
                        false
                    }
                }
            }
        }
    }

    /// Wait for the child process to exit and cleanup the supervisor thread and thread pool.
    /// It returns `WaitStatus` of the child process.
    ///
    /// # Examples
    ///
    /// ```ignore
    /// let status = Supervisor::wait(pid, thread_handle, pool).unwrap();
    /// ```
    pub fn wait(
        pid: Pid,
        thread_handle: ChildHandle,
        pool_handle: ThreadPool,
    ) -> Result<std::process::ExitStatus, Errno> {
        let mut status = 0;
        loop {
            match unsafe { nix::libc::waitpid(pid.into(), &mut status, 0) } {
                -1 if nix::errno::errno() == nix::libc::EINTR => continue,
                -1 => return Err(Errno::last()),
                _ => break,
            }
        }
        let _ = thread_handle.join().map_err(|_| Errno::EPIPE)?;
        pool_handle.join();
        Ok(std::process::ExitStatus::from_raw(status))
    }
}

/// Processes the address family of a `SockaddrStorage` object and performs logging or other
/// required operations specific to the syscall being handled.
///
/// This helper function isolates the logic involved in dealing with different address families
/// and reduces code duplication across different syscall handler functions.
///
/// # Parameters
///
/// - `addr`: Reference to a `SockaddrStorage`, representing the socket address involved in the syscall.
/// - `syscall_name`: A string slice holding the name of the syscall being handled, used for logging purposes.
///
/// # Safety
///
/// The function contains unsafe blocks due to potential TOCTOU (Time-of-Check Time-of-Use)
/// vulnerabilities. Each unsafe block within this function has been annotated with a detailed
/// safety comment to ensure that unsafe operations are used correctly and securely.
///
/// # Errors
///
/// The function returns an `io::Error` in cases where:
/// - The conversion from `SockaddrStorage` to a specific address family representation fails.
/// - Any other unexpected error condition occurs during the processing of the address family.
///
/// # Returns
///
/// Returns an `Result<(), Errno>`:
/// - `Ok(())` if the processing is successful.
/// - `Err(Errno)` containing a description of the error, if any error occurs during processing.
fn sandbox_addr(
    proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    req: &ScmpNotifReq,
    addr: &SockaddrStorage,
    syscall_name: &str,
    caps: Capability,
) -> Result<(), Errno> {
    match addr.family() {
        Some(AddressFamily::Unix) => {
            sandbox_addr_unix(proc, request, req, addr, syscall_name, caps)
        }
        Some(AddressFamily::Inet) => {
            sandbox_addr_inet(proc, request, req, addr, syscall_name, caps)
        }
        Some(AddressFamily::Inet6) => {
            sandbox_addr_inet6(proc, request, req, addr, syscall_name, caps)
        }
        Some(_) | None => sandbox_addr_notsup(proc, request),
    }
}

/// Process a `AddressFamily::Unix` socket address.
fn sandbox_addr_unix(
    proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    req: &ScmpNotifReq,
    addr: &SockaddrStorage,
    syscall_name: &str,
    caps: Capability,
) -> Result<(), Errno> {
    let addr = addr.as_unix_addr().ok_or(Errno::EINVAL)?;
    let (path, abs) = match (addr.path(), addr.as_abstract()) {
        (Some(path), _) => {
            let path = path.as_os_str().as_bytes();
            let null = path.iter().position(|&b| b == 0).unwrap_or(path.len());
            let path = PathBuf::from(OsStr::from_bytes(&path[..null]));
            trace!("ctx": "syscall", "sys": syscall_name, "addr": format!("{}", path.display()), "cap": caps);
            (path, false)
        }
        (_, Some(path)) => {
            let null = path.iter().position(|&b| b == 0).unwrap_or(path.len());
            let path = PathBuf::from(OsStr::from_bytes(&path[..null]));
            trace!("ctx": "syscall", "sys": syscall_name, "addr": format!("{}", path.display()), "cap": caps);
            (path, true)
        }
        _ => {
            // unnamed unix socket
            return Ok(());
        }
    };

    let path = if path.is_relative() {
        let arg = SyscallPathArgument {
            dirfd: None,
            path: None,
            null: false,
            empty: false,
            resolve: true,
            dotlast: None,
        };
        proc.read_directory(request, &arg)?.join(&path)
    } else {
        path
    };

    // Check for access.
    let sandbox = request.get_sandbox(false);
    let action = sandbox.check_unix(caps, &path);
    drop(sandbox);

    if action == Action::Deny {
        // Report access violation.
        warn!("ctx": "access", "cap": caps, "unix": format!("{}", path.display()), "abs": abs, "pid": req.pid, "sys": syscall_name);
    }
    if action == Action::Allow {
        Ok(())
    } else {
        // Deny or Filter.
        Err(Errno::EACCES)
    }
}

/// Process an `AddressFamily::Inet` socket address.
fn sandbox_addr_inet(
    _proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    req: &ScmpNotifReq,
    addr: &SockaddrStorage,
    syscall_name: &str,
    caps: Capability,
) -> Result<(), Errno> {
    let addr = addr.as_sockaddr_in().ok_or(Errno::EINVAL)?;
    let port = addr.port();
    let addr = IpAddr::V4(Ipv4Addr::from(addr.ip()));
    trace!("ctx": "syscall", "sys": syscall_name, "addr": format!("{addr}@{port}"), "cap": caps);

    // Check for access.
    let sandbox = request.get_sandbox(false);
    let action = sandbox.check_ip(caps, addr, port);
    drop(sandbox);

    if action == Action::Deny {
        // Report access violation.
        warn!("ctx": "access", "cap": caps, "addr": format!("{addr}@{port}"), "pid": req.pid, "sys": syscall_name);
    }
    if action == Action::Allow {
        Ok(())
    } else {
        // Deny or Filter.
        Err(Errno::EACCES)
    }
}

/// Process an `AddressFamily::Inet6` socket address.
fn sandbox_addr_inet6(
    _proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    req: &ScmpNotifReq,
    addr: &SockaddrStorage,
    syscall_name: &str,
    caps: Capability,
) -> Result<(), Errno> {
    let addr = addr.as_sockaddr_in6().ok_or(Errno::EINVAL)?;
    let port = addr.port();
    let addr = IpAddr::V6(addr.ip());
    trace!("ctx": "syscall", "sys": syscall_name, "addr": format!("{addr}@{port}"), "cap": caps);

    // Check for access.
    let sandbox = request.get_sandbox(false);
    let action = sandbox.check_ip(caps, addr, port);
    drop(sandbox);

    if action == Action::Deny {
        // Report access violation.
        warn!("ctx": "access", "cap": caps, "addr": format!("{addr}@{port}"), "pid": req.pid, "sys": syscall_name);
    }
    if action == Action::Allow {
        Ok(())
    } else {
        // Deny or Filter.
        Err(Errno::EACCES)
    }
}

/// Process a socket address of an unsupported socket family.
fn sandbox_addr_notsup(_proc: &RemoteProcess, request: &UNotifyEventRequest) -> Result<(), Errno> {
    let sandbox = request.get_sandbox(false);
    let ok = sandbox.allow_unsupported_socket_families();
    drop(sandbox);

    if ok {
        Ok(())
    } else {
        Err(Errno::EAFNOSUPPORT)
    }
}

/// Process the first path argument.
fn sandbox_path_1(
    paths: &NonEmpty<PathBuf>,
    caps: Capability,
    syscall_name: &str,
    request: &UNotifyEventRequest,
    _: &RemoteProcess,
) -> Result<(), Errno> {
    let req = request.get_request();
    let path = paths.first(); // paths are NonEmpty, so first always exists.
    trace!("ctx": "syscall", "sys": syscall_name, "cap": caps, "path": format!("{}", path.display()));

    if caps.is_empty() {
        return Err(Errno::EINVAL);
    }

    // Special case for system calls that must create the path argument.
    // Note: dangling symbolic links are considered existing.
    if (syscall_name.starts_with("mkdir") || syscall_name.starts_with("mknod"))
        && lstat(Path::new(path)).is_ok()
    {
        debug!("ctx": "sandbox_path_1", "err": "EEXIST", "path": format!("{}", path.display()));
        return Err(Errno::EEXIST);
    }

    let sandbox = request.get_sandbox(false);
    let mut action = Action::Allow;

    // Sandboxing
    if caps.contains(Capability::CAP_READ) {
        action = sandbox.check_path(Capability::CAP_READ, path);
    }
    if action == Action::Allow && caps.contains(Capability::CAP_STAT) {
        action = sandbox.check_path(Capability::CAP_STAT, path);
    }
    if action == Action::Allow && caps.contains(Capability::CAP_WRITE) {
        action = sandbox.check_path(Capability::CAP_WRITE, path);
    }
    if action == Action::Allow && caps.contains(Capability::CAP_EXEC) {
        action = sandbox.check_path(Capability::CAP_EXEC, path);
    }

    // exec/kill
    if caps.contains(Capability::CAP_EXEC) && sandbox.check_exec(path) == Action::Kill {
        warn!("ctx": "exec/kill", "cap": caps, "path": format!("{}", path.display()), "pid": req.pid, "sys": syscall_name);
        #[allow(clippy::cast_possible_wrap)]
        let _ = kill(Pid::from_raw(req.pid as i32), Some(SIGKILL));
        action = Action::Kill;
    }

    match action {
        Action::Allow => Ok(()),
        _ if caps.contains(Capability::CAP_STAT) => {
            // We do not report violations for Stat capability because we are essentially hiding paths.
            Err(Errno::ENOENT)
        }
        Action::Filter | Action::Kill => Err(Errno::EACCES),
        Action::Deny => {
            // Report access violation.
            warn!("ctx": "access", "cap": caps, "path": format!("{}", path.display()), "pid": req.pid, "sys": syscall_name);
            Err(Errno::EACCES)
        }
    }
}

/// Process both the first and the second path argument.
fn sandbox_path_2(
    paths: &NonEmpty<PathBuf>,
    caps: Capability,
    syscall_name: &str,
    request: &UNotifyEventRequest,
    proc: &RemoteProcess,
) -> Result<(), Errno> {
    let (source, target) = (paths.first(), paths.last()); // paths are NonEmpty, so first/last always exists.
    trace!("ctx": "syscall", "sys": syscall_name, "cap": caps, "source": format!("{}", source.display()), "target": format!("{}", target.display()));

    // SAFETY: For symbolic linking we require only read access to the target.
    // This is safe because on a subsequent write access we will be resolving
    // symbolic links before deciding on whether to grant access.
    let (caps_1, caps_2) = if matches!(syscall_name, "symlink" | "symlinkat") {
        (Capability::CAP_READ, Capability::CAP_WRITE)
    } else {
        (caps, caps)
    };

    sandbox_path_1(
        &nonempty![source.to_path_buf()],
        caps_1,
        syscall_name,
        request,
        proc,
    )?;
    sandbox_path_1(
        &nonempty![target.to_path_buf()],
        caps_2,
        syscall_name,
        request,
        proc,
    )?;
    Ok(())
}

/*
 * System call handlers
 */
fn sys_bind(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: If addr is NULL we must return EFAULT
    // here without further processing.
    let req = request.get_request();
    if req.data.args[1] == 0 {
        return request.fail_syscall(nix::libc::EFAULT);
    }
    syscall_network_handler(request, 1, 2, false, "bind", sandbox_addr)
}

fn sys_connect(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: If addr is NULL we must return EFAULT
    // here without further processing.
    let req = request.get_request();
    if req.data.args[1] == 0 {
        return request.fail_syscall(nix::libc::EFAULT);
    }
    syscall_network_handler(request, 1, 2, false, "connect", sandbox_addr)
}

fn sys_recvfrom(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: If receieve buffer is NULL we must
    // return EFAULT here without further
    // processing.
    let req = request.get_request();
    if req.data.args[1] == 0 {
        return request.fail_syscall(nix::libc::EFAULT);
    }
    syscall_network_handler(request, 4, 5, true, "recvfrom", sandbox_addr)
}

fn sys_sendto(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: If send buffer is NULL we must
    // return EFAULT here without further
    // processing.
    let req = request.get_request();
    if req.data.args[1] == 0 {
        return request.fail_syscall(nix::libc::EFAULT);
    }
    syscall_network_handler(request, 4, 5, false, "sendto", sandbox_addr)
}

fn sys_execve(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(request, "execve", argv, sandbox_path_1, |_, _, _, _| {
        // SAFETY: This is vulnerable to TOCTOU,
        // unfortunately we cannot emulate exec,
        // so we have to live with it...
        Ok(unsafe { request.continue_syscall() })
    })
}

fn sys_execveat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    let empty = req.data.args[4] & nix::libc::AT_EMPTY_PATH as u64 != 0;
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        empty,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(request, "execveat", argv, sandbox_path_1, |_, _, _, _| {
        // SAFETY: This is vulnerable to TOCTOU,
        // unfortunately we cannot emulate exec,
        // so we have to live with it...
        Ok(unsafe { request.continue_syscall() })
    })
}

fn sys_chdir(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(request, "chdir", argv, sandbox_path_1, |_, _, _, _| {
        // SAFETY: This is vulnerable to TOCTOU,
        // unfortunately we cannot emulate chdir,
        // so we have to live with it...
        Ok(unsafe { request.continue_syscall() })
    })
}

fn sys_fchdir(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(request, "chdir", argv, sandbox_path_1, |_, _, _, _| {
        // SAFETY: This is vulnerable to TOCTOU,
        // unfortunately we cannot emulate fchdir,
        // so we have to live with it...
        Ok(unsafe { request.continue_syscall() })
    })
}

fn sys_getdents(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_handler!(request, |req: &ScmpNotifReq, proc: &RemoteProcess| {
        if req.data.args[1] == 0 {
            // SAFETY: If the second argument which must hold a pointer to a
            // linux_dirent structure is NULL, we must return EFAULT
            // without further processing here.
            return Err(Errno::EFAULT);
        } else if req.data.args[2] == 0 {
            // SAFETY:Result buffer is too small
            return Err(Errno::EINVAL);
        }

        // If sandboxing for List capability is off, return immediately.
        let sandbox = request.get_sandbox(false);
        let enabled = sandbox.enabled(Capability::CAP_STAT);
        #[allow(clippy::cast_possible_wrap)]
        let is_lock = sandbox.locked_for_pid(req.pid as nix::libc::pid_t);
        drop(sandbox); // release the read lock.
        if !enabled {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU.
            // Since stat sandboxing is disabled here, this is safe.
            return unsafe { Ok(request.continue_syscall()) };
        }

        let fd = proc.get_fd(req.data.args[0] as RawFd)?;
        let dir = match crate::fs::canonicalize(
            format!("/proc/self/fd/{fd}"),
            true,
            crate::fs::MissingHandling::Missing,
        ) {
            Ok(dir) => dir,
            Err(error) => {
                error!("ctx": "canonicalize", "fd": fd, "err": format!("{error}"));
                let _ = close(fd);
                return Err(Errno::ENOENT);
            }
        };

        // SAFETY: The count argument to the getdents call
        // must not be fully trusted, it can be overly large,
        // and allocating a Vector of that capacity may overflow.
        // This bug was discovered by trinity in this build:
        // https://builds.sr.ht/~alip/job/1077263
        #[allow(clippy::cast_possible_truncation)]
        let count = req.data.args[2] as usize;
        let count = count.min(1000000); // Cap count at 1mio
        let mut entries = Vec::with_capacity(count.saturating_add(1) /* /dev/syd */);
        loop {
            match getdents(fd, count) {
                Err(error) => {
                    let _ = close(fd);
                    return Err(error);
                }
                Ok(None) => {
                    let _ = close(fd);
                    // getdents returned None, there are no more entries.
                    return Ok(request.return_syscall(0));
                }
                Ok(Some(e)) => {
                    for entry in e.iter() {
                        let name = entry.to_path_buf();
                        let path = dir.join(&name);
                        match sandbox_path_1(
                            &nonempty![path.clone()],
                            Capability::CAP_STAT,
                            "getdents",
                            request,
                            proc,
                        ) {
                            Ok(_) /*allow*/ => {
                                trace!("ctx": "stat", "path": format!("{}", path.display()), "act": "allow");
                                entries.push(entry.clone());
                            },
                            /*
                            Err(_) if name == Path::new(".") /*deny the whole dir */ => {
                                debug!("ctx": "stat", "path": format!("{}", path.display()), "act": "deny");
                                let _ = close(fd);
                                return Ok(request.fail_syscall(nix::libc::ENOENT))
                            }
                            */
                            _ => {
                                /* this entry is denied, skip it. */
                                debug!("ctx": "stat", "path": format!("{}", path.display()), "act": "hide");
                            },
                        };
                    }
                    if !entries.is_empty() {
                        break; // exit the loop once we have allowed entries
                    }
                }
            };
        }

        let list_syd = !is_lock
            && crate::DEV_STAT.map_or(false, |(dev, ino)| {
                fstat(fd).map_or(false, |metadata| {
                    metadata.st_dev == dev && metadata.st_ino == ino
                })
            });

        let _ = close(fd);

        if list_syd {
            crate::compat::DirEntry::append_dev_syd(&mut entries);
        }
        crate::compat::DirEntry::fix_offsets(&mut entries);

        let mut buffer = Vec::new();
        for entry in &entries {
            let bytes = &entry.dirent;

            // Ensure we don't append more bytes than the buffer can hold.
            if buffer.len().saturating_add(bytes.len()) > count {
                break;
            }

            buffer.extend_from_slice(bytes);
        }

        #[allow(clippy::cast_possible_truncation)]
        proc.write_mem(&buffer, req.data.args[1] as usize, request)?;

        #[allow(clippy::cast_possible_wrap)]
        Ok(request.return_syscall(buffer.len() as i64))
    })
}

fn sys_access(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "access",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let mode = AccessFlags::from_bits_truncate(req.data.args[1] as nix::libc::c_int);
            faccessat(Some(fd), &name, mode, AtFlags::empty()).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_faccessat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "faccessat",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let mode = AccessFlags::from_bits_truncate(req.data.args[2] as nix::libc::c_int);
            faccessat(Some(fd), &name, mode, AtFlags::empty()).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_faccessat2(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    let resolve = req.data.args[3] & nix::libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        empty: false,
        dotlast: None,
        resolve,
    });
    syscall_path_handler(
        request,
        "faccessat2",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let mode = AccessFlags::from_bits_truncate(req.data.args[2] as nix::libc::c_int);
            #[allow(clippy::cast_possible_truncation)]
            let flags = AtFlags::from_bits_truncate(req.data.args[3] as nix::libc::c_int);
            faccessat(Some(fd), &name, mode, flags).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_chmod(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: Disallow setting setuid/setgid bits as necessary.
    let sandbox = request.get_sandbox(false);
    let restrict_chmod = !sandbox.allow_unsafe_chmod();
    drop(sandbox);
    if restrict_chmod {
        let req = request.get_request();
        if req.data.args[1] & u64::from(nix::libc::S_ISUID | nix::libc::S_ISGID) != 0 {
            return request.fail_syscall(nix::libc::EACCES);
        }
    }
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "chmod",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let mode = Mode::from_bits_truncate(req.data.args[1] as u32);
            fchmodat(Some(fd), &name, mode, FchmodatFlags::FollowSymlink)
                .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_fchmod(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: Disallow setting setuid/setgid bits as necessary.
    let sandbox = request.get_sandbox(false);
    let restrict_chmod = !sandbox.allow_unsafe_chmod();
    drop(sandbox);
    if restrict_chmod {
        let req = request.get_request();
        if req.data.args[1] & u64::from(nix::libc::S_ISUID | nix::libc::S_ISGID) != 0 {
            return request.fail_syscall(nix::libc::EACCES);
        }
    }
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "fchmod",
        argv,
        sandbox_path_1,
        |_, _, request, proc| {
            let req = request.get_request();
            #[allow(clippy::cast_possible_truncation)]
            let mode = Mode::from_bits_truncate(req.data.args[1] as u32);
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            // SAFETY: Valid FD.
            let fd = unsafe { OwnedFd::from_raw_fd(fd) };
            fchmod(fd.as_raw_fd(), mode).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_fchmodat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();

    // SAFETY: Disallow setting setuid/setgid bits as necessary.
    let sandbox = request.get_sandbox(false);
    let restrict_chmod = !sandbox.allow_unsafe_chmod();
    drop(sandbox);
    if restrict_chmod && req.data.args[2] & u64::from(nix::libc::S_ISUID | nix::libc::S_ISGID) != 0
    {
        return request.fail_syscall(nix::libc::EACCES);
    }
    let empty = req.data.args[3] & nix::libc::AT_EMPTY_PATH as u64 != 0;
    let resolve = req.data.args[3] & nix::libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        dotlast: None,
        empty,
        resolve,
    });
    syscall_path_handler(
        request,
        "fchmodat",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let mode = Mode::from_bits_truncate(req.data.args[2] as u32);
            let flags = if resolve {
                FchmodatFlags::FollowSymlink
            } else {
                FchmodatFlags::NoFollowSymlink
            };
            fchmodat(Some(fd), &name, mode, flags).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_chown(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "chown",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_wrap)]
            let owner = if req.data.args[1] as i64 == -1 {
                None
            } else {
                Some(Uid::from_raw(req.data.args[1] as nix::libc::uid_t))
            };
            #[allow(clippy::cast_possible_wrap)]
            let group = if req.data.args[2] as i64 == -1 {
                None
            } else {
                Some(Gid::from_raw(req.data.args[2] as nix::libc::gid_t))
            };
            fchownat(Some(fd), &name, owner, group, FchownatFlags::FollowSymlink)
                .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_lchown(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "chown",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_wrap)]
            let owner = if req.data.args[1] as i64 == -1 {
                None
            } else {
                Some(Uid::from_raw(req.data.args[1] as nix::libc::uid_t))
            };
            #[allow(clippy::cast_possible_wrap)]
            let group = if req.data.args[2] as i64 == -1 {
                None
            } else {
                Some(Gid::from_raw(req.data.args[2] as nix::libc::gid_t))
            };
            fchownat(
                Some(fd),
                &name,
                owner,
                group,
                FchownatFlags::NoFollowSymlink,
            )
            .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_fchown(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "fchown",
        argv,
        sandbox_path_1,
        |_, _, request, proc| {
            let req = request.get_request();
            #[allow(clippy::cast_possible_wrap)]
            let owner = if req.data.args[1] as i64 == -1 {
                None
            } else {
                Some(Uid::from_raw(req.data.args[1] as nix::libc::uid_t))
            };
            #[allow(clippy::cast_possible_wrap)]
            let group = if req.data.args[2] as i64 == -1 {
                None
            } else {
                Some(Gid::from_raw(req.data.args[2] as nix::libc::gid_t))
            };
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            // SAFETY: Valid FD
            let fd = unsafe { OwnedFd::from_raw_fd(fd) };
            fchown(fd.as_raw_fd(), owner, group).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_fchownat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    let empty = req.data.args[4] & nix::libc::AT_EMPTY_PATH as u64 != 0;
    let resolve = req.data.args[4] & nix::libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        dotlast: None,
        empty,
        resolve,
    });
    syscall_path_handler(
        request,
        "fchownat",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_wrap)]
            let owner = if req.data.args[2] as i64 == -1 {
                None
            } else {
                Some(Uid::from_raw(req.data.args[2] as nix::libc::uid_t))
            };
            #[allow(clippy::cast_possible_wrap)]
            let group = if req.data.args[3] as i64 == -1 {
                None
            } else {
                Some(Gid::from_raw(req.data.args[3] as nix::libc::gid_t))
            };
            let flags = if resolve {
                FchownatFlags::FollowSymlink
            } else {
                FchownatFlags::NoFollowSymlink
            };
            fchownat(Some(fd), &name, owner, group, flags).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_creat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_open_handler(request, "creat")
}

fn sys_link(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = nonempty![
        SyscallPathArgument {
            dirfd: None,
            path: Some(0),
            null: false,
            empty: false,
            resolve: false,
            dotlast: None,
        },
        SyscallPathArgument {
            dirfd: None,
            path: Some(1),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::ENOENT),
        }
    ];
    syscall_path_handler(
        request,
        "link",
        argv,
        sandbox_path_2,
        |fds, paths, request, _proc| {
            let fd_old = fds.first().as_raw_fd();
            let fd_new = fds.last().as_raw_fd();
            let name_old = crate::path_filename(paths.first());
            let name_new = crate::path_filename(paths.last());

            // SAFETY: We must respect process' umask here.
            let req = request.get_request();
            #[allow(clippy::cast_possible_wrap)]
            let pid = Pid::from_raw(req.pid as i32);
            let lock = request.umask_lock();
            let prev_umask = umask(proc_umask(Some(pid))?);
            let result = linkat(
                Some(fd_old),
                &name_old,
                Some(fd_new),
                &name_new,
                LinkatFlags::NoSymlinkFollow,
            )
            .map(|_| request.return_syscall(0));
            let _ = umask(prev_umask);
            drop(lock);

            result
        },
    )
}

fn sys_symlink(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_handler!(request, |_req: &ScmpNotifReq, proc: &RemoteProcess| {
        let path_argv = nonempty![
            SyscallPathArgument {
                dirfd: None,
                path: Some(0),
                null: false,
                empty: false,
                resolve: false,
                dotlast: None,
            },
            SyscallPathArgument {
                dirfd: None,
                path: Some(1),
                null: false,
                empty: false,
                resolve: false,
                dotlast: None,
            }
        ];

        // If sandboxing for all the selected capabilities is off, return immediately.
        // Exception: CAP_EXEC is available and we have patterns in exec/kill list.
        let mut check = false;
        let sandbox = request.get_sandbox(false);
        if sandbox.enabled(Capability::CAP_WRITE) {
            check = true;
        }
        drop(sandbox); // release the read lock.
        if !check {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU,
            // however since sandboxing for the respective Capability
            // is off, this is safe here.
            return unsafe { Ok(request.continue_syscall()) };
        }

        let mut paths: Vec<PathBuf> = Vec::new();
        let (path, target) = proc.read_path_opt(request, &path_argv[0])?;
        paths.push(path);
        let path = proc.read_path(request, &path_argv[1])?;
        paths.push(path.clone());

        // At this point, we're absolutely sure the
        // path vectors have at least a single
        // element. Hence, using `unwrap` is fine
        // here.
        #[allow(clippy::disallowed_methods)]
        let non_empty_paths = NonEmpty::from_vec(paths).unwrap();
        sandbox_path_2(
            &non_empty_paths,
            Capability::CAP_WRITE,
            "symlink",
            request,
            proc,
        )?;

        // SAFETY: We have already resolved symbolic
        // links as necessary, from this point on we
        // do not resolve any longer for safety as
        // the filesystem may change which may
        // result in a TOCTOU.
        let mut how = openat2::OpenHow::new(nix::libc::O_PATH, 0);
        how.resolve = openat2::ResolveFlags::NO_SYMLINKS;
        let path = path.parent().unwrap_or(&path);
        let newdirfd = openat2::openat2(None, path, &how)
            .map_err(|e| Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL)))?;
        // SAFETY: Valid FD
        let newdirfd = unsafe { OwnedFd::from_raw_fd(newdirfd) };
        let linkpath = crate::path_filename(non_empty_paths.last());
        symlinkat(
            &target.unwrap_or(PathBuf::from("")),
            Some(newdirfd.as_raw_fd()),
            &linkpath,
        )
        .map(|_| request.return_syscall(0))
    })
}

fn sys_unlink(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: Some(Errno::EINVAL),
    });
    syscall_path_handler(
        request,
        "unlink",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            unlinkat(Some(fd), &name, UnlinkatFlags::NoRemoveDir).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_linkat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    #[allow(clippy::cast_possible_truncation)]
    let flags = req.data.args[4] as nix::libc::c_int;
    let empty = flags & nix::libc::AT_EMPTY_PATH != 0;
    let resolve = flags & nix::libc::AT_SYMLINK_FOLLOW != 0;
    let argv = nonempty![
        SyscallPathArgument {
            dirfd: Some(0),
            path: Some(1),
            null: false,
            dotlast: None,
            empty,
            resolve,
        },
        SyscallPathArgument {
            dirfd: Some(2),
            path: Some(3),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::ENOENT),
        }
    ];
    syscall_path_handler(
        request,
        "linkat",
        argv,
        sandbox_path_2,
        |fds, paths, request, _proc| {
            let fd_old = fds.first().as_raw_fd();
            let fd_new = fds.last().as_raw_fd();
            let name_old = crate::path_filename(paths.first());
            let name_new = crate::path_filename(paths.last());
            let flags = if resolve {
                LinkatFlags::SymlinkFollow
            } else {
                LinkatFlags::NoSymlinkFollow
            };
            // SAFETY: We must respect process' umask here.
            #[allow(clippy::cast_possible_wrap)]
            let pid = Pid::from_raw(req.pid as i32);
            let lock = request.umask_lock();
            let prev_umask = umask(proc_umask(Some(pid))?);
            let result = linkat(Some(fd_old), &name_old, Some(fd_new), &name_new, flags)
                .map(|_| request.return_syscall(0));
            let _ = umask(prev_umask);
            drop(lock);

            result
        },
    )
}

fn sys_symlinkat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_handler!(request, |_req: &ScmpNotifReq, proc: &RemoteProcess| {
        let path_argv = nonempty![
            SyscallPathArgument {
                dirfd: None,
                path: Some(0),
                null: false,
                empty: false,
                resolve: true,
                dotlast: None,
            },
            SyscallPathArgument {
                dirfd: Some(1),
                path: Some(2),
                null: false,
                empty: false,
                resolve: false,
                dotlast: None,
            }
        ];

        // If sandboxing for all the selected capabilities is off, return immediately.
        // Exception: CAP_EXEC is available and we have patterns in exec/kill list.
        let mut check = false;
        let sandbox = request.get_sandbox(false);
        if sandbox.enabled(Capability::CAP_WRITE) {
            check = true;
        }
        drop(sandbox); // release the read lock.
        if !check {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU,
            // however since sandboxing for the respective Capability
            // is off, this is safe here.
            return unsafe { Ok(request.continue_syscall()) };
        }

        let mut paths: Vec<PathBuf> = Vec::new();
        let (path, target) = proc.read_path_opt(request, &path_argv[0])?;
        paths.push(path);
        let path = proc.read_path(request, &path_argv[1])?;
        paths.push(path.clone());

        // At this point, we're absolutely sure the
        // path vectors have at least a single
        // element. Hence, using `unwrap` is fine
        // here.
        #[allow(clippy::disallowed_methods)]
        let non_empty_paths = NonEmpty::from_vec(paths).unwrap();
        sandbox_path_2(
            &non_empty_paths,
            Capability::CAP_WRITE,
            "symlink",
            request,
            proc,
        )?;

        // SAFETY: We have already resolved symbolic
        // links as necessary, from this point on we
        // do not resolve any longer for safety as
        // the filesystem may change which may
        // result in a TOCTOU.
        let mut how = openat2::OpenHow::new(nix::libc::O_PATH, 0);
        how.resolve = openat2::ResolveFlags::NO_SYMLINKS;
        let path = path.parent().unwrap_or(&path);
        let newdirfd = openat2::openat2(None, path, &how)
            .map_err(|e| Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL)))?;
        // SAFETY: Valid FD
        let newdirfd = unsafe { OwnedFd::from_raw_fd(newdirfd) };
        let linkpath = crate::path_filename(non_empty_paths.last());
        symlinkat(
            &target.unwrap_or(PathBuf::from("")),
            Some(newdirfd.as_raw_fd()),
            &linkpath,
        )
        .map(|_| request.return_syscall(0))
    })
}

fn sys_unlinkat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        empty: false,
        resolve: false,
        dotlast: Some(Errno::EINVAL),
    });
    syscall_path_handler(
        request,
        "unlinkat",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let flags = if req.data.args[2] as nix::libc::c_int & nix::libc::AT_REMOVEDIR != 0 {
                UnlinkatFlags::RemoveDir
            } else {
                UnlinkatFlags::NoRemoveDir
            };
            unlinkat(Some(fd), &name, flags).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_mkdir(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: Some(Errno::ENOENT),
    });
    syscall_path_handler(
        request,
        "mkdir",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let mode = Mode::from_bits_truncate(req.data.args[1] as nix::libc::c_int as u32);

            // SAFETY: We must respect process' umask here.
            #[allow(clippy::cast_possible_wrap)]
            let pid = Pid::from_raw(req.pid as i32);
            let lock = request.umask_lock();
            let prev_umask = umask(proc_umask(Some(pid))?);
            let result = mkdirat(fd, &name, mode).map(|_| request.return_syscall(0));
            let _ = umask(prev_umask);
            drop(lock);

            result
        },
    )
}

fn sys_rmdir(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: Some(Errno::EINVAL),
    });
    syscall_path_handler(
        request,
        "rmdir",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            unlinkat(Some(fd), &name, UnlinkatFlags::RemoveDir).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_mkdirat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        empty: false,
        resolve: false,
        dotlast: Some(Errno::ENOENT),
    });
    syscall_path_handler(
        request,
        "mkdirat",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let mode = Mode::from_bits_truncate(req.data.args[2] as nix::libc::c_int as u32);

            // SAFETY: We must respect process' umask here.
            #[allow(clippy::cast_possible_wrap)]
            let pid = Pid::from_raw(req.pid as i32);
            let lock = request.umask_lock();
            let prev_umask = umask(proc_umask(Some(pid))?);
            let result = mkdirat(fd, &name, mode).map(|_| request.return_syscall(0));
            let _ = umask(prev_umask);
            drop(lock);

            result
        },
    )
}

fn sys_mknod(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    #[allow(clippy::cast_possible_truncation)]
    let kind =
        SFlag::from_bits_truncate(req.data.args[1] as nix::libc::mode_t & SFlag::S_IFMT.bits());
    if !matches!(kind, SFlag::S_IFREG | SFlag::S_IFSOCK) {
        // SAFETY: We do not allow:
        // 1. Device special files because this may
        //    circumvent path based access control.
        // 2. Named pipes because SydB☮x's open
        //    emulator cannot handle them safely.
        // Deny with ENOSYS signaling these file types
        // are unsupported.
        return request.fail_syscall(nix::libc::ENOSYS);
    }
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "mknod",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let perm = Mode::from_bits_truncate(req.data.args[1] as u32);
            #[allow(clippy::cast_possible_truncation)]
            let dev = req.data.args[2] as nix::libc::dev_t;

            // SAFETY: We must respect process' umask here.
            #[allow(clippy::cast_possible_wrap)]
            let pid = Pid::from_raw(req.pid as i32);
            let lock = request.umask_lock();
            let prev_umask = umask(proc_umask(Some(pid))?);
            let result = mknodat(fd, &name, kind, perm, dev).map(|_| request.return_syscall(0));
            let _ = umask(prev_umask);
            drop(lock);

            result
        },
    )
}

fn sys_mknodat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    #[allow(clippy::cast_possible_truncation)]
    let kind =
        SFlag::from_bits_truncate(req.data.args[2] as nix::libc::mode_t & SFlag::S_IFMT.bits());
    if !matches!(kind, SFlag::S_IFREG | SFlag::S_IFSOCK) {
        // SAFETY: We do not allow:
        // 1. Device special files because this may
        //    circumvent path based access control.
        // 2. Named pipes because SydB☮x's open
        //    emulator cannot handle them safely.
        // Deny with ENOSYS signaling these file types
        // are unsupported.
        return request.fail_syscall(nix::libc::ENOSYS);
    }
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "mknodat",
        argv,
        sandbox_path_1,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd = fds.first().as_raw_fd();
            let name = crate::path_filename(paths.first());
            #[allow(clippy::cast_possible_truncation)]
            let perm = Mode::from_bits_truncate(req.data.args[1] as u32);
            #[allow(clippy::cast_possible_truncation)]
            let dev = req.data.args[2] as nix::libc::dev_t;

            // SAFETY: We must respect process' umask here.
            #[allow(clippy::cast_possible_wrap)]
            let pid = Pid::from_raw(req.pid as i32);
            let lock = request.umask_lock();
            let prev_umask = umask(proc_umask(Some(pid))?);
            let result = mknodat(fd, &name, kind, perm, dev).map(|_| request.return_syscall(0));
            let _ = umask(prev_umask);
            drop(lock);

            result
        },
    )
}

fn sys_mount(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = nonempty![
        SyscallPathArgument {
            dirfd: None,
            path: Some(0),
            null: true,
            empty: false,
            resolve: true,
            dotlast: None,
        },
        SyscallPathArgument {
            dirfd: None,
            path: Some(1),
            null: false,
            empty: false,
            resolve: true,
            dotlast: None,
        }
    ];
    syscall_path_handler(
        request,
        "mount",
        argv,
        sandbox_path_2,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let source: Option<&PathBuf> = if req.data.args[0] == 0 {
                None
            } else {
                Some(paths.first())
            };
            let target = if req.data.args[1] == 0 {
                return Err(Errno::EFAULT);
            } else {
                paths.last()
            };
            let fstype = if req.data.args[2] == 0 {
                None
            } else {
                let mut fstype = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut fstype, req.data.args[2] as usize, request)?;
                Some(PathBuf::from(OsString::from_vec(fstype)))
            };
            let flags = MsFlags::from_bits_truncate(req.data.args[3] as nix::libc::c_ulong);
            let data = if req.data.args[4] == 0 {
                None
            } else {
                let mut data = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut data, req.data.args[4] as usize, request)?;
                Some(PathBuf::from(OsString::from_vec(data)))
            };
            mount(source, target, fstype.as_ref(), flags, data.as_ref())
                .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_umount(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "umount",
        argv,
        sandbox_path_1,
        |_fds, paths, request, _proc| {
            let path = paths.first();
            umount(path).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_umount2(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    #[allow(clippy::cast_possible_truncation)]
    let flags = MntFlags::from_bits_truncate(req.data.args[1] as nix::libc::c_int);
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: !flags.contains(MntFlags::UMOUNT_NOFOLLOW),
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "umount2",
        argv,
        sandbox_path_1,
        |_fds, paths, request, _proc| {
            let path = paths.first();
            umount2(path, flags).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_open(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_open_handler(request, "open")
}

fn sys_openat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_open_handler(request, "openat")
}

fn sys_openat2(request: &UNotifyEventRequest) -> ScmpNotifResp {
    syscall_open_handler(request, "openat2")
}

fn sys_rename(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = nonempty![
        SyscallPathArgument {
            dirfd: None,
            path: Some(0),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::EINVAL),
        },
        SyscallPathArgument {
            dirfd: None,
            path: Some(1),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::EINVAL),
        }
    ];
    syscall_path_handler(
        request,
        "rename",
        argv,
        sandbox_path_2,
        |fds, paths, request, _proc| {
            let fd_old = fds.first().as_raw_fd();
            let fd_new = fds.last().as_raw_fd();
            let name_old = crate::path_filename(paths.first());
            let name_new = crate::path_filename(paths.last());
            renameat(Some(fd_old), &name_old, Some(fd_new), &name_new)
                .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_renameat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = nonempty![
        SyscallPathArgument {
            dirfd: Some(0),
            path: Some(1),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::EINVAL),
        },
        SyscallPathArgument {
            dirfd: Some(2),
            path: Some(3),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::EINVAL),
        }
    ];
    syscall_path_handler(
        request,
        "renameat",
        argv,
        sandbox_path_2,
        |fds, paths, request, _proc| {
            let fd_old = fds.first().as_raw_fd();
            let fd_new = fds.last().as_raw_fd();
            let name_old = crate::path_filename(paths.first());
            let name_new = crate::path_filename(paths.last());
            renameat(Some(fd_old), &name_old, Some(fd_new), &name_new)
                .map(|_| request.return_syscall(0))
        },
    )
}

fn sys_renameat2(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = nonempty![
        SyscallPathArgument {
            dirfd: Some(0),
            path: Some(1),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::EINVAL),
        },
        SyscallPathArgument {
            dirfd: Some(2),
            path: Some(3),
            null: false,
            empty: false,
            resolve: false,
            dotlast: Some(Errno::EINVAL),
        }
    ];
    syscall_path_handler(
        request,
        "renameat2",
        argv,
        sandbox_path_2,
        |fds, paths, request, _proc| {
            let req = request.get_request();
            let fd_old = fds.first().as_raw_fd();
            let fd_new = fds.last().as_raw_fd();
            let name_old = crate::path_filename(paths.first());
            let name_new = crate::path_filename(paths.last());
            let name_old =
                CString::new(name_old.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?;
            let name_new =
                CString::new(name_new.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?;
            #[allow(clippy::cast_possible_truncation)]
            let flags = req.data.args[4] as u32;
            // musl does not have renameat2 yet.
            if unsafe {
                nix::libc::syscall(
                    nix::libc::SYS_renameat2,
                    fd_old,
                    name_old.as_ptr(),
                    fd_new,
                    name_new.as_ptr(),
                    flags,
                )
            } == 0
            {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_stat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_stat_handler(request, "stat", argv)
}

fn sys_fstat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_stat_handler(request, "fstat", argv)
}

fn sys_lstat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_stat_handler(request, "lstat", argv)
}

fn sys_statx(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    let empty = req.data.args[2] & nix::libc::AT_EMPTY_PATH as u64 != 0;
    let resolve = req.data.args[2] & nix::libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        dotlast: None,
        empty,
        resolve,
    });
    syscall_stat_handler(request, "statx", argv)
}

fn sys_newfstatat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    let empty = req.data.args[3] & nix::libc::AT_EMPTY_PATH as u64 != 0;
    let resolve = req.data.args[3] & nix::libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: false,
        dotlast: None,
        empty,
        resolve,
    });
    syscall_stat_handler(request, "newfstatat", argv)
}

fn sys_utime(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "utime",
        argv,
        sandbox_path_1,
        |fds, paths, request, proc| {
            let req = request.get_request();
            let name = if req.data.args[0] != 0 {
                Some(crate::path_filename(paths.first()))
            } else {
                None
            };
            let name = if let Some(name) = name {
                Some(CString::new(name.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?)
            } else {
                None
            };
            #[allow(clippy::cast_possible_truncation)]
            let time = match proc.remote_utimbuf(req.data.args[1] as usize, request)? {
                Some(time) => time.as_ptr(),
                None => std::ptr::null(),
            };
            let r = if let Some(name) = name {
                let (fd, name) = (fds.first(), name.as_ptr());
                // SAFETY: To be able to pass the arguments
                // path and time as NULL to the system call
                // we need to directly call it here.
                unsafe {
                    nix::libc::syscall(nix::libc::SYS_utimensat, fd.as_raw_fd(), name, time, 0)
                }
            } else {
                let (fd, name) = (proc.get_fd(req.data.args[0] as RawFd)?, 0);
                // SAFETY: Valid FD.
                let fd = unsafe { OwnedFd::from_raw_fd(fd) };
                unsafe {
                    nix::libc::syscall(nix::libc::SYS_utimensat, fd.as_raw_fd(), name, time, 0)
                }
            };
            if r == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_utimes(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "utimes",
        argv,
        sandbox_path_1,
        |fds, paths, request, proc| {
            let req = request.get_request();
            let name = if req.data.args[0] != 0 {
                Some(crate::path_filename(paths.first()))
            } else {
                None
            };
            let name = if let Some(name) = name {
                Some(CString::new(name.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?)
            } else {
                None
            };
            #[allow(clippy::cast_possible_truncation)]
            let time = match proc.remote_timeval(req.data.args[1] as usize, request)? {
                Some(time) => time.as_ptr(),
                None => std::ptr::null(),
            };
            let r = if let Some(name) = name {
                let (fd, name) = (fds.first(), name.as_ptr());
                // SAFETY: To be able to pass the arguments
                // path and time as NULL to the system call
                // we need to directly call it here.
                unsafe {
                    nix::libc::syscall(nix::libc::SYS_utimensat, fd.as_raw_fd(), name, time, 0)
                }
            } else {
                let (fd, name) = (proc.get_fd(req.data.args[0] as RawFd)?, 0);
                // SAFETY: Valid FD.
                let fd = unsafe { OwnedFd::from_raw_fd(fd) };
                unsafe {
                    nix::libc::syscall(nix::libc::SYS_utimensat, fd.as_raw_fd(), name, time, 0)
                }
            };
            if r == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_futimesat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: pathname may be NULL here:
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: true,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "futimesat",
        argv,
        sandbox_path_1,
        |fds, paths, request, proc| {
            let req = request.get_request();
            let name = if req.data.args[1] != 0 {
                Some(crate::path_filename(paths.first()))
            } else {
                None
            };
            let name = if let Some(name) = name {
                Some(CString::new(name.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?)
            } else {
                None
            };
            #[allow(clippy::cast_possible_truncation)]
            let time = match proc.remote_timeval(req.data.args[2] as usize, request)? {
                Some(time) => time.as_ptr(),
                None => std::ptr::null(),
            };
            let r = if let Some(name) = name {
                let (fd, name) = (fds.first(), name.as_ptr());
                // SAFETY: To be able to pass the arguments
                // path and time as NULL to the system call
                // we need to directly call it here.
                unsafe {
                    nix::libc::syscall(nix::libc::SYS_utimensat, fd.as_raw_fd(), name, time, 0)
                }
            } else {
                let (fd, name) = (proc.get_fd(req.data.args[0] as RawFd)?, 0);
                // SAFETY: Valid FD.
                let fd = unsafe { OwnedFd::from_raw_fd(fd) };
                unsafe {
                    nix::libc::syscall(nix::libc::SYS_utimensat, fd.as_raw_fd(), name, time, 0)
                }
            };
            if r == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_utimensat(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let req = request.get_request();
    #[allow(clippy::cast_possible_truncation)]
    let resolve = req.data.args[3] as nix::libc::c_int & nix::libc::AT_SYMLINK_NOFOLLOW == 0;
    // SAFETY: pathname may be NULL here:
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: Some(1),
        null: true,
        empty: false,
        dotlast: None,
        resolve,
    });
    syscall_path_handler(
        request,
        "utimensat",
        argv,
        sandbox_path_1,
        |fds, paths, request, proc| {
            let req = request.get_request();
            let name = if req.data.args[1] != 0 {
                Some(crate::path_filename(paths.first()))
            } else {
                None
            };
            let name = if let Some(name) = name {
                Some(CString::new(name.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?)
            } else {
                None
            };
            #[allow(clippy::cast_possible_truncation)]
            let time = match proc.remote_timespec(req.data.args[2] as usize, request)? {
                Some(time) => time.as_ptr(),
                None => std::ptr::null(),
            };
            let r = if let Some(name) = name {
                let (fd, name) = (fds.first(), name.as_ptr());
                // SAFETY: To be able to pass the arguments
                // path and time as NULL to the system call
                // we need to directly call it here.
                unsafe {
                    nix::libc::syscall(
                        nix::libc::SYS_utimensat,
                        fd.as_raw_fd(),
                        name,
                        time,
                        req.data.args[3],
                    )
                }
            } else {
                let (fd, name) = (proc.get_fd(req.data.args[0] as RawFd)?, 0);
                // SAFETY: Valid FD.
                let fd = unsafe { OwnedFd::from_raw_fd(fd) };
                unsafe {
                    nix::libc::syscall(
                        nix::libc::SYS_utimensat,
                        fd.as_raw_fd(),
                        name,
                        time,
                        req.data.args[3],
                    )
                }
            };
            if r == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_truncate(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "truncate",
        argv,
        sandbox_path_1,
        |_fds, paths, request, _proc| {
            let req = request.get_request();
            #[allow(clippy::cast_possible_wrap)]
            let len = req.data.args[1] as nix::libc::off_t;
            let path = paths.first();
            truncate(path, len).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_ftruncate(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "ftruncate",
        argv,
        sandbox_path_1,
        |_, _, request, proc| {
            let req = request.get_request();
            #[allow(clippy::cast_possible_wrap)]
            let len = req.data.args[1] as nix::libc::off_t;
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            // SAFETY: Valid FD.
            let fd = unsafe { OwnedFd::from_raw_fd(fd) };
            ftruncate(fd.as_raw_fd(), len).map(|_| request.return_syscall(0))
        },
    )
}

fn sys_getxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "getxattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::ENOENT)?;
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: The size argument to the getxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            #[allow(clippy::cast_possible_truncation)]
            let len = req.data.args[3] as usize;
            let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
            let mut buf = if len == 0 { None } else { Some(vec![0u8; len]) };
            let ptr = match buf.as_mut() {
                Some(b) => b.as_mut_ptr(),
                None => std::ptr::null_mut(),
            };
            // SAFETY: In libc we trust.
            let n = unsafe {
                nix::libc::getxattr(
                    path.as_ptr(),
                    name,
                    ptr as *mut _ as *mut nix::libc::c_void,
                    len,
                )
            };
            let n = if n == -1 {
                return Err(Errno::last());
            } else {
                n as usize
            };
            if let Some(buf) = buf {
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(&buf[..n], req.data.args[2] as usize, request)?;
            }
            #[allow(clippy::cast_possible_wrap)]
            Ok(request.return_syscall(n as i64))
        },
    )
}

fn sys_lgetxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "lgetxattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::ENOENT)?;
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: The size argument to the lgetxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            #[allow(clippy::cast_possible_truncation)]
            let len = req.data.args[3] as usize;
            let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
            let mut buf = if len == 0 { None } else { Some(vec![0u8; len]) };
            let ptr = match buf.as_mut() {
                Some(b) => b.as_mut_ptr(),
                None => std::ptr::null_mut(),
            };
            // SAFETY: In libc we trust.
            let n = unsafe {
                nix::libc::lgetxattr(
                    path.as_ptr(),
                    name,
                    ptr as *mut _ as *mut nix::libc::c_void,
                    len,
                )
            };
            let n = if n == -1 {
                return Err(Errno::last());
            } else {
                n as usize
            };
            if let Some(buf) = buf {
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(&buf[..n], req.data.args[2] as usize, request)?;
            }
            #[allow(clippy::cast_possible_wrap)]
            Ok(request.return_syscall(n as i64))
        },
    )
}

fn sys_fgetxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    // SAFETY: value must not be NULL.
    let req = request.get_request();
    if req.data.args[2] == 0 {
        return request.fail_syscall(nix::libc::EFAULT);
    } else if req.data.args[3] == 0 {
        return request.fail_syscall(nix::libc::ERANGE);
    }
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "fgetxattr",
        argv,
        sandbox_path_1,
        |_fds, _paths, request, proc| {
            let req = request.get_request();
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: The size argument to the fgetxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            #[allow(clippy::cast_possible_truncation)]
            let len = req.data.args[3] as usize;
            let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
            let mut buf = if len == 0 { None } else { Some(vec![0u8; len]) };
            let ptr = match buf.as_mut() {
                Some(b) => b.as_mut_ptr(),
                None => std::ptr::null_mut(),
            };
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            // SAFETY: In libc we trust.
            let n = unsafe {
                nix::libc::fgetxattr(fd, name, ptr as *mut _ as *mut nix::libc::c_void, len)
            };
            let e = Errno::last();
            let _ = close(fd);
            let n = if n == -1 {
                return Err(e);
            } else {
                n as usize
            };
            if let Some(buf) = buf {
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(&buf[..n], req.data.args[2] as usize, request)?;
            }
            #[allow(clippy::cast_possible_wrap)]
            Ok(request.return_syscall(n as i64))
        },
    )
}

fn sys_setxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "setxattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?;
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: The size argument to the setxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            let (buf, len) = if req.data.args[3] == 0 {
                (None, 0)
            } else {
                #[allow(clippy::cast_possible_truncation)]
                let len = req.data.args[3] as usize;
                let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
                let mut buf = vec![0u8; len];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[2] as usize, request)?;
                (Some(buf), len)
            };
            let buf = if let Some(mut buf) = buf {
                buf.as_mut_ptr()
            } else {
                std::ptr::null_mut()
            };
            #[allow(clippy::cast_possible_truncation)]
            let flags = req.data.args[4] as nix::libc::c_int;

            // SAFETY: In libc we trust.
            if unsafe {
                nix::libc::setxattr(
                    path.as_ptr(),
                    name,
                    buf as *mut _ as *mut nix::libc::c_void,
                    len,
                    flags,
                )
            } == 0
            {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_lsetxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "lsetxattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?;
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: The size argument to the lsetxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            let (buf, len) = if req.data.args[3] == 0 {
                (None, 0)
            } else {
                #[allow(clippy::cast_possible_truncation)]
                let len = req.data.args[3] as usize;
                let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
                let mut buf = vec![0u8; len];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[2] as usize, request)?;
                (Some(buf), len)
            };
            let buf = if let Some(mut buf) = buf {
                buf.as_mut_ptr()
            } else {
                std::ptr::null_mut()
            };
            #[allow(clippy::cast_possible_truncation)]
            let flags = req.data.args[4] as nix::libc::c_int;
            // SAFETY: In libc we trust.
            if unsafe {
                nix::libc::lsetxattr(
                    path.as_ptr(),
                    name,
                    buf as *mut _ as *mut nix::libc::c_void,
                    len,
                    flags,
                )
            } == 0
            {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_fsetxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "fsetxattr",
        argv,
        sandbox_path_1,
        |_fds, _paths, request, proc| {
            let req = request.get_request();
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: The size argument to the lsetxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            let (buf, len) = if req.data.args[3] == 0 {
                (None, 0)
            } else {
                #[allow(clippy::cast_possible_truncation)]
                let len = req.data.args[3] as usize;
                let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
                let mut buf = vec![0u8; len];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[2] as usize, request)?;
                (Some(buf), len)
            };
            let buf = if let Some(mut buf) = buf {
                buf.as_mut_ptr()
            } else {
                std::ptr::null_mut()
            };
            #[allow(clippy::cast_possible_truncation)]
            let flags = req.data.args[4] as nix::libc::c_int;
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            // SAFETY: Valid FD
            let fd = unsafe { &OwnedFd::from_raw_fd(fd) };
            // SAFETY: In libc we trust.
            if unsafe {
                nix::libc::fsetxattr(
                    fd.as_raw_fd(),
                    name,
                    buf as *mut _ as *mut nix::libc::c_void,
                    len,
                    flags,
                )
            } == 0
            {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_listxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "listxattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::ENOENT)?;
            // SAFETY: The size argument to the listxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            #[allow(clippy::cast_possible_truncation)]
            let len = req.data.args[2] as usize;
            let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
            let mut buf = if len == 0 { None } else { Some(vec![0u8; len]) };
            let ptr = match buf.as_mut() {
                Some(b) => b.as_mut_ptr(),
                None => std::ptr::null_mut(),
            };
            // SAFETY: In libc we trust.
            let n = unsafe {
                nix::libc::listxattr(path.as_ptr(), ptr as *mut _ as *mut nix::libc::c_char, len)
            };
            let n = if n == -1 {
                return Err(Errno::last());
            } else {
                n as usize
            };
            if let Some(buf) = buf {
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(&buf[..n], req.data.args[1] as usize, request)?;
            }
            #[allow(clippy::cast_possible_wrap)]
            Ok(request.return_syscall(n as i64))
        },
    )
}

fn sys_flistxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "flistxattr",
        argv,
        sandbox_path_1,
        |_fds, _paths, request, proc| {
            let req = request.get_request();
            // SAFETY: The size argument to the flistxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            #[allow(clippy::cast_possible_truncation)]
            let len = req.data.args[2] as usize;
            let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
            let mut buf = if len == 0 { None } else { Some(vec![0u8; len]) };
            let ptr = match buf.as_mut() {
                Some(b) => b.as_mut_ptr(),
                None => std::ptr::null_mut(),
            };
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            // SAFETY: Valid FD
            let fd = unsafe { &OwnedFd::from_raw_fd(fd) };
            // SAFETY: In libc we trust.
            let n = unsafe {
                nix::libc::flistxattr(fd.as_raw_fd(), ptr as *mut _ as *mut nix::libc::c_char, len)
            };
            let n = if n == -1 {
                return Err(Errno::last());
            } else {
                n as usize
            };
            if let Some(buf) = buf {
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(&buf[..n], req.data.args[1] as usize, request)?;
            }
            #[allow(clippy::cast_possible_wrap)]
            Ok(request.return_syscall(n as i64))
        },
    )
}

fn sys_llistxattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "llistxattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::ENOENT)?;
            // SAFETY: The size argument to the llistxattr call
            // must not be fully trusted, it can be overly large,
            // and allocating a Vector of that capacity may overflow.
            #[allow(clippy::cast_possible_truncation)]
            let len = req.data.args[2] as usize;
            let len = len.min(nix::libc::PATH_MAX as usize); // Cap count at PATH_MAX
            let mut buf = if len == 0 { None } else { Some(vec![0u8; len]) };
            let ptr = match buf.as_mut() {
                Some(b) => b.as_mut_ptr(),
                None => std::ptr::null_mut(),
            };
            // SAFETY: In libc we trust.
            let n = unsafe {
                nix::libc::llistxattr(path.as_ptr(), ptr as *mut _ as *mut nix::libc::c_char, len)
            };
            let n = if n == -1 {
                return Err(Errno::last());
            } else {
                n as usize
            };
            if let Some(buf) = buf {
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(&buf[..n], req.data.args[1] as usize, request)?;
            }
            #[allow(clippy::cast_possible_wrap)]
            Ok(request.return_syscall(n as i64))
        },
    )
}

fn sys_removexattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "removexattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::ENOENT)?;
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: In libc we trust.
            if unsafe { nix::libc::removexattr(path.as_ptr(), name) } == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_fremovexattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: Some(0),
        path: None,
        null: false,
        empty: true,
        resolve: true,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "fremovexattr",
        argv,
        sandbox_path_1,
        |_paths, _cwd, request, proc| {
            let req = request.get_request();
            let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
            let name = if req.data.args[1] != 0 {
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(CStr::from_bytes_until_nul(&buf).map_err(|_| Errno::E2BIG)?)
            } else {
                None
            };
            let name = if let Some(name) = name {
                name.as_ptr()
            } else {
                std::ptr::null()
            };
            let fd = proc.get_fd(req.data.args[0] as RawFd)?;
            // SAFETY: Valid FD
            let fd = unsafe { OwnedFd::from_raw_fd(fd) };
            // SAFETY: In libc we trust.
            if unsafe { nix::libc::fremovexattr(fd.as_raw_fd(), name) } == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

fn sys_lremovexattr(request: &UNotifyEventRequest) -> ScmpNotifResp {
    let argv = NonEmpty::new(SyscallPathArgument {
        dirfd: None,
        path: Some(0),
        null: false,
        empty: false,
        resolve: false,
        dotlast: None,
    });
    syscall_path_handler(
        request,
        "lremovexattr",
        argv,
        sandbox_path_1,
        |_fds, paths, request, proc| {
            let req = request.get_request();
            let path = paths.first();
            let path = CString::new(path.as_os_str().as_bytes()).map_err(|_| Errno::EINVAL)?;
            let name = if req.data.args[1] != 0 {
                let mut buf = vec![0u8; nix::libc::PATH_MAX as usize];
                #[allow(clippy::cast_possible_truncation)]
                proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
                Some(buf)
            } else {
                None
            };
            let name = if let Some(ref name) = name {
                CStr::from_bytes_until_nul(name)
                    .map_err(|_| Errno::E2BIG)?
                    .as_ptr()
            } else {
                std::ptr::null()
            };
            // SAFETY: In libc we trust.
            if unsafe { nix::libc::lremovexattr(path.as_ptr(), name) } == 0 {
                Ok(request.return_syscall(0))
            } else {
                Err(Errno::last())
            }
        },
    )
}

/// Handles syscalls related to paths, reducing code redundancy and ensuring a uniform way of dealing with paths.
///
/// # Parameters
///
/// - `request`: User notification request from seccomp.
/// - `syscall_name`: The name of the syscall being handled, used for logging and error reporting.
/// - `arg_mappings`: Non-empty list of argument mappings containing dirfd and path indexes, if applicable.
/// - `handler`: Closure that processes the constructed canonical paths and performs additional syscall-specific operations.
///
/// # Returns
///
/// - `ScmpNotifResp`: Response indicating the result of the syscall handling.
fn syscall_path_handler<C, H>(
    request: &UNotifyEventRequest,
    syscall_name: &str,
    path_argv: NonEmpty<SyscallPathArgument>,
    checker: C,
    handler: H,
) -> ScmpNotifResp
where
    C: Fn(
        &NonEmpty<PathBuf>,
        Capability,
        &str,
        &UNotifyEventRequest,
        &RemoteProcess,
    ) -> Result<(), Errno>,
    H: Fn(
        &NonEmpty<OwnedFd>,
        &NonEmpty<PathBuf>,
        &UNotifyEventRequest,
        &RemoteProcess,
    ) -> Result<ScmpNotifResp, Errno>,
{
    syscall_handler!(request, |req: &ScmpNotifReq, proc: &RemoteProcess| {
        // Determine system call capabilities.
        let caps = Capability::try_from((proc, request, req, syscall_name))?;

        // If sandboxing for all the selected capabilities is off, return immediately.
        // Exception: CAP_EXEC is available and we have patterns in exec/kill list.
        let mut check = false;
        let sandbox = request.get_sandbox(false);
        if caps.contains(Capability::CAP_EXEC) && sandbox.has_exec_kill() {
            check = true;
        } else {
            for cap in [
                Capability::CAP_READ,
                Capability::CAP_STAT,
                Capability::CAP_WRITE,
                Capability::CAP_EXEC,
                Capability::CAP_CONNECT,
                Capability::CAP_BIND,
            ] {
                if sandbox.enabled(cap) {
                    check = true;
                    break;
                }
            }
        }
        drop(sandbox); // release the read lock.
        if !check {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU,
            // however since sandboxing for the respective Capability
            // is off, this is safe here.
            return unsafe { Ok(request.continue_syscall()) };
        }

        let mut paths: Vec<PathBuf> = Vec::new();
        for arg in path_argv.iter() {
            paths.push(proc.read_path(request, arg)?);
        }

        // At this point, we're absolutely sure the
        // path vectors have at least a single
        // element. Hence, using `unwrap` is fine
        // here.
        #[allow(clippy::disallowed_methods)]
        let non_empty_paths = NonEmpty::from_vec(paths).unwrap();
        checker(&non_empty_paths, caps, syscall_name, request, proc)?;

        // SAFETY: We have already resolved symbolic
        // links as necessary, from this point on we
        // do not resolve any longer for safety as
        // the filesystem may change which may
        // result in a TOCTOU.
        let mut how = openat2::OpenHow::new(nix::libc::O_PATH, 0);
        how.resolve = openat2::ResolveFlags::NO_SYMLINKS;

        let mut fds = Vec::new();
        for path in &non_empty_paths {
            let path = match path.parent() {
                Some(path) => {
                    how.flags |= nix::libc::O_DIRECTORY as u64;
                    path
                }
                None => path,
            };
            let fd = openat2::openat2(None, path, &how)
                .map_err(|e| Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL)))?;
            fds.push(unsafe { OwnedFd::from_raw_fd(fd) });
        }
        #[allow(clippy::disallowed_methods)]
        let non_empty_fds = NonEmpty::from_vec(fds).unwrap();

        handler(&non_empty_fds, &non_empty_paths, request, proc)
    })
}

#[allow(clippy::cognitive_complexity)]
fn syscall_open_handler(request: &UNotifyEventRequest, syscall_name: &str) -> ScmpNotifResp {
    syscall_handler!(request, |req: &ScmpNotifReq, proc: &RemoteProcess| {
        #[allow(clippy::cast_possible_truncation)]
        let (mut flags, mode) = match syscall_name {
            "creat" => {
                let flags = OFlag::O_CREAT | OFlag::O_WRONLY | OFlag::O_TRUNC;
                let mode = Mode::from_bits_truncate(req.data.args[1] as nix::libc::mode_t);
                (flags, mode)
            }
            "open" => {
                let flags = OFlag::from_bits_truncate(req.data.args[1] as nix::libc::c_int);
                let mode = Mode::from_bits_truncate(req.data.args[2] as nix::libc::mode_t);
                (flags, mode)
            }
            "openat" => {
                let flags = OFlag::from_bits_truncate(req.data.args[2] as nix::libc::c_int);
                let mode = Mode::from_bits_truncate(req.data.args[3] as nix::libc::mode_t);
                (flags, mode)
            }
            "openat2" => {
                let open_how = proc.remote_ohow(
                    req.data.args[2] as usize,
                    req.data.args[3] as usize,
                    request,
                )?;
                let flags = OFlag::from_bits_truncate(open_how.flags as nix::libc::c_int);
                let mode = Mode::from_bits_truncate(open_how.mode as nix::libc::mode_t);
                (flags, mode)
            }
            _ => unreachable!(),
        };

        if flags.contains(OFlag::O_PATH) {
            // SAFETY: O_PATH obtains a file descriptor for
            // low-level access. The file is not opened and
            // other operations will fail.  Hence we consider
            // this secure, and continue the system call here.
            return unsafe { Ok(request.continue_syscall()) };
        }

        let cap = if flags.contains(OFlag::O_CREAT) {
            Capability::CAP_WRITE
        } else {
            match flags & OFlag::O_ACCMODE {
                OFlag::O_RDONLY => Capability::CAP_READ,
                _ => Capability::CAP_WRITE,
            }
        };

        let sandbox = request.get_sandbox(false);
        if !sandbox.enabled(cap) {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU.
            // Since the sandbox is off, this is fine here.
            return unsafe { Ok(request.continue_syscall()) };
        }
        drop(sandbox);

        #[allow(clippy::cast_possible_wrap)]
        let pid = Pid::from_raw(req.pid as i32);
        let resolve =
            !(flags.contains(OFlag::O_NOFOLLOW) || flags.contains(OFlag::O_CREAT | OFlag::O_EXCL));

        // Step 1: Read the remote path.
        let (path, arg) = match syscall_name {
            "open" | "creat" => {
                let arg = SyscallPathArgument {
                    dirfd: None,
                    path: Some(0),
                    null: false,
                    empty: false,
                    dotlast: None,
                    resolve,
                };
                (remote_path_n!(proc, req, 0, &request)?, arg)
            }
            _ => {
                // "openat" | "openat2"
                let arg = SyscallPathArgument {
                    dirfd: Some(0),
                    path: Some(1),
                    null: false,
                    empty: false,
                    dotlast: None,
                    resolve,
                };
                (remote_path_n!(proc, req, 1, &request)?, arg)
            }
        };
        let path = PathBuf::from(OsStr::from_bytes(path.to_bytes()));

        // Step 2: Handle the special /dev/syd paths.
        if path.strip_prefix(MAGIC_PREFIX).is_ok() {
            let sandbox = request.get_sandbox(false);
            #[allow(clippy::cast_possible_wrap)]
            let is_lock = sandbox.locked_for_pid(req.pid as nix::libc::pid_t);
            if is_lock {
                return Err(Errno::ENOENT);
            }
            let content = if cap.contains(Capability::CAP_READ) {
                Some(serde_json::to_string(&*sandbox).map_err(|_| Errno::EINVAL)?)
            } else {
                None
            };
            drop(sandbox);

            let fd = if let Some(content) = content {
                let fd = open(
                    ".",
                    OFlag::O_RDWR | OFlag::O_TMPFILE,
                    Mode::S_IRUSR | Mode::S_IWUSR,
                )?;
                write(fd, content.as_bytes())?;
                lseek(fd, 0, Whence::SeekSet)?;
                fd
            } else {
                open("/dev/null", flags, mode)?
            };

            // Send the file descriptor to the process and return the fd no.
            let result = request.add_fd(fd, true);
            let _ = close(fd);
            return result.map(|fd| request.return_syscall(i64::from(fd)));
        }

        // Step 3: Canonicalize the path
        //
        // Step 3.1: Ensure path is not empty, normalize.
        if path.is_empty() && !flags.contains(OFlag::O_TMPFILE) {
            return Err(Errno::ENOENT);
        }
        let path = normalize_path(path);
        // Step 3.2: Ensure the path is absolute.
        let abspath = if path.is_absolute() {
            path.clone()
        } else if path == PathBuf::from(".") {
            // SAFETY: Add a trailing slash to the directory
            // to assert it must be a directory.
            proc.read_directory(request, &arg)?.join("")
        } else {
            // SAFETY: We join path even if it's empty.
            // This way e.g. a final dot adds a trailing
            // slash to the path, ensuring the system call
            // to fail with ENOTDIR as necessary.
            proc.read_directory(request, &arg)?.join(&path)
        };

        // Step 3.3: Check for magic paths and replace.
        let abspath = replace_self_with_pid(abspath, pid);

        // Step 3.4: Canonicalize the resulting path.
        let abspath =
            match crate::fs::canonicalize(&abspath, resolve, crate::fs::MissingHandling::Missing) {
                Ok(path) => {
                    trace!("ctx": "path",
                        "act": "resolve",
                        "arg": arg,
                        "path_fini": format!("{}", abspath.display()),
                        "path_orig": format!("{}", path.display()),
                        "pid": pid.as_raw());
                    path
                }
                Err(error) => {
                    debug!("ctx": "path",
                        "act": "resolve",
                        "arg": arg,
                        "path_fini": format!("{}", abspath.display()),
                        "path_orig": format!("{}", path.display()),
                        "err": error.to_string(),
                        "pid": pid.as_raw());
                    return Err(error);
                }
            };

        // Step 3.5: Check for magic paths, and replace.
        // SAFETY: We have to do this one more time to ensure
        // relative paths are properly handled.
        let abspath = replace_self_with_pid(abspath, pid);

        // Step 3.6: Add O_NOFOLLOW to flags.
        // SAFETY: We have already resolved the symbolic
        // links in the path as necessary, to prevent a
        // time-of-check to time-of-use vector we add
        // O_NOFOLLOW to flags here.
        flags.insert(OFlag::O_NOFOLLOW);

        // Step 4: Check for access.
        let sandbox = request.get_sandbox(false);
        let action = sandbox.check_path(cap, &abspath);
        drop(sandbox);

        // Step 5: Perform action: allow->emulate, deny->log.
        match action {
            Action::Allow => {
                // The system call is allowlisted.
                // To prevent TOCTOU, we open the file ourselves,
                // and put the file descriptor to the process'
                // address space with SECCOMP_IOCTL_NOTIF_ADDFD.

                // SAFETY: Hold the lock early to ensure sanity.
                // The lock will be released when it goes out of
                // scope.
                let _lock = request.umask_lock();

                // SAFETY: We have already resolved symbolic
                // links as necessary, from this point on we
                // do not resolve any longer for safety as
                // the filesystem may change which may
                // result in a TOCTOU.
                let mut how = openat2::OpenHow::new(nix::libc::O_PATH, 0);
                how.resolve = openat2::ResolveFlags::NO_SYMLINKS;
                let parent = abspath.parent().unwrap_or(&abspath);
                let name = crate::path_filename(&abspath);
                let fd = openat2::openat2(None, parent, &how)
                    .map_err(|e| Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL)))?;
                // SAFETY: Valid FD.
                let fd = unsafe { OwnedFd::from_raw_fd(fd) };

                // SAFETY: Refuse to open the file is not one of:
                // 1. Regular file
                // 2. Directory
                // 3. Socket
                // 4. Character device
                // This leaves out block devices and FIFOs above all.
                // Deny with ENOSYS signaling these file types are unsupported.
                if let Ok(stat) = fstatat(fd.as_raw_fd(), &name, AtFlags::empty()) {
                    match SFlag::from_bits_truncate(stat.st_mode & SFlag::S_IFMT.bits()) {
                        SFlag::S_IFREG | SFlag::S_IFSOCK | SFlag::S_IFCHR => {}
                        SFlag::S_IFDIR => {
                            // SAFETY: Opening a directory for write must open with EISDIR.
                            if flags.contains(
                                OFlag::O_WRONLY | OFlag::O_RDWR | OFlag::O_CREAT | OFlag::O_TRUNC,
                            ) {
                                return Err(Errno::EISDIR);
                            }
                        }
                        _ => return Err(Errno::ENOSYS),
                    }
                }

                // SAFETY: We must respect process' umask here.
                let prev_umask = umask(proc_umask(Some(pid))?);
                let result = openat(fd.as_raw_fd(), &name, flags, mode);
                let _ = umask(prev_umask);

                let fd_this = match result {
                    Ok(fd) => {
                        trace!("ctx": "open",
                            "pid": pid.as_raw(),
                            "path_orig": format!("{}", path.display()),
                            "path_fini": format!("{}", abspath.display()),
                            "sys": syscall_name,
                            "flags": crate::oflag_fmt(flags),
                            "mode": mode.bits());
                        fd
                    }
                    Err(error)
                        if error == Errno::EEXIST
                            || error == Errno::ENOENT
                            || error == Errno::ENOTDIR =>
                    {
                        debug!("ctx": "open",
                            "pid": pid.as_raw(),
                            "err": error.to_string(),
                            "path_orig": format!("{}", path.display()),
                            "path_fini": format!("{}", abspath.display()),
                            "sys": syscall_name,
                            "flags": crate::oflag_fmt(flags),
                            "mode": mode.bits());
                        return Err(error);
                    }
                    Err(error) => {
                        info!("ctx": "open",
                            "pid": pid.as_raw(),
                            "err": error.to_string(),
                            "path_orig": path.to_string_lossy(),
                            "path_fini": abspath,
                            "sys": syscall_name,
                            "flags": crate::oflag_fmt(flags),
                            "mode": mode.bits());
                        return Err(error);
                    }
                };
                let result = request.add_fd(fd_this, flags.contains(OFlag::O_CLOEXEC));
                let _ = close(fd_this);
                result.map(|fd| request.return_syscall(i64::from(fd)))
            }
            Action::Filter | Action::Kill => Err(Errno::EACCES),
            Action::Deny => {
                // Report access violation.
                warn!("ctx": "access",
                    "cap": cap,
                    "orig": format!("{}", path.display()),
                    "path": format!("{}", abspath.display()),
                    "flags": oflag_fmt(flags),
                    "mode": mode.bits(),
                    "pid": pid.as_raw(),
                    "sys": syscall_name);
                Err(Errno::EACCES)
            }
        }
    })
}

#[allow(clippy::cognitive_complexity)]
fn syscall_stat_handler(
    request: &UNotifyEventRequest,
    syscall_name: &str,
    path_argv: NonEmpty<SyscallPathArgument>,
) -> ScmpNotifResp {
    syscall_handler!(request, |req: &ScmpNotifReq, proc: &RemoteProcess| {
        // If sandboxing for CAP_STAT is off, and magic lock is set, return immediately.
        let sandbox = request.get_sandbox(false);
        #[allow(clippy::cast_possible_wrap)]
        let is_lock = sandbox.locked_for_pid(req.pid as nix::libc::pid_t);
        let is_stat = sandbox.enabled(Capability::CAP_STAT);
        if is_lock && !is_stat {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU.
            // However, since stat sandboxing is disabled here this is
            // safe to call.
            return unsafe { Ok(request.continue_syscall()) };
        }
        drop(sandbox); // release the read lock.

        let arg = path_argv.first();
        let path = proc.read_path(request, arg)?;
        trace!("ctx": "syscall", "sys": syscall_name, "cap": Capability::CAP_STAT, "path": format!("{}", path.display()));

        let (path, fd, flags) = if let Ok(path) = path.strip_prefix(MAGIC_PREFIX) {
            // Handle magic prefix (ie /dev/syd)
            let sandbox = request.get_sandbox(false);
            #[allow(clippy::cast_possible_wrap)]
            if sandbox.locked_for_pid(req.pid as nix::libc::pid_t) {
                // Magic commands locked, return ENOENT.
                return Ok(request.fail_syscall(nix::libc::ENOENT));
            }
            drop(sandbox);

            // Execute magic command.
            let mut sandbox = request.get_sandbox(true);
            if path.is_empty() {
                sandbox.config("")?;
            } else {
                // SAFETY: Conversion from PathBuf to String is OK here,
                // since sandbox config function does not work on the
                // filesystem, rather treats the Strings as sandbox
                // commands.
                match sandbox.config(path.to_string_lossy().as_ref()) {
                    Ok(_) => {
                        if let Err(error) = sandbox.build_globsets() {
                            error!("ctx": "config", "pid": req.pid, "sys": syscall_name, "cfg": format!("{}", path.display()), "err": error.to_string());
                        } else {
                            warn!("ctx": "config", "pid": req.pid, "sys": syscall_name, "cfg": format!("{}", path.display()));
                        }
                    }
                    Err(error) => {
                        error!("ctx": "config", "pid": req.pid, "sys": syscall_name, "cfg": format!("{}", path.display()), "err": error.to_string());
                        return Err(error);
                    }
                };
            }
            drop(sandbox);

            // SAFETY: We trust the contents of the MAGIC_FILE.
            let path = unsafe { CString::from_vec_unchecked(MAGIC_FILE.into()) };

            (path, nix::libc::AT_FDCWD, 0)
        } else if is_stat {
            let non_empty_paths = nonempty![path.clone()];
            sandbox_path_1(
                &non_empty_paths,
                Capability::CAP_STAT,
                syscall_name,
                request,
                proc,
            )?;

            let is_proc = GLOB_PROC.is_match(&path);
            let fd = if syscall_name == "fstat"
                || (is_proc && matches!(syscall_name, "newfstatat" | "statx"))
            {
                proc.get_fd(req.data.args[0] as RawFd)?
            } else {
                // SAFETY: We have already resolved symbolic
                // links as necessary, from this point on we
                // do not resolve any longer for safety as
                // the filesystem may change which may
                // result in a TOCTOU.
                let mut how = openat2::OpenHow::new(nix::libc::O_PATH, 0);
                if !arg.resolve {
                    how.flags |= nix::libc::O_NOFOLLOW as u64;
                }
                how.resolve = openat2::ResolveFlags::NO_SYMLINKS;
                openat2::openat2(None, path, &how)
                    .map_err(|e| Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL)))?
            };

            match syscall_name {
                "stat" => (CString::default(), fd, nix::libc::AT_EMPTY_PATH),
                "lstat" => (
                    CString::default(),
                    fd,
                    nix::libc::AT_EMPTY_PATH | nix::libc::AT_SYMLINK_NOFOLLOW,
                ),
                "fstat" => (CString::default(), fd, nix::libc::AT_EMPTY_PATH),
                "statx" | "newfstatat" => {
                    if is_proc {
                        (CString::default(), fd, nix::libc::AT_EMPTY_PATH)
                    } else {
                        let flags = if arg.resolve {
                            nix::libc::AT_EMPTY_PATH
                        } else {
                            nix::libc::AT_EMPTY_PATH | nix::libc::AT_SYMLINK_NOFOLLOW
                        };
                        (CString::default(), fd, flags)
                    }
                }
                _ => unreachable!(),
            }
        } else {
            // Continue system call normally.
            // SAFETY: This is unsafe due to vulnerability to TOCTOU,
            // however if we're here stat sandboxing is disabled,
            // so this is safe to call.
            return unsafe { Ok(request.continue_syscall()) };
        };

        // SAFETY: Wrap the fd into an OwnedFd so it gets closed properly.
        let _guard = if fd != nix::libc::AT_FDCWD {
            // SAFETY: Valid FD.
            Some(unsafe { OwnedFd::from_raw_fd(fd) })
        } else {
            None
        };

        match syscall_name {
            "statx" => {
                let mut statx = MaybeUninit::<crate::compat::statx>::uninit();
                // SAFETY: In libc we trust.
                #[allow(clippy::cast_possible_truncation)]
                let ret = unsafe {
                    nix::libc::syscall(
                        nix::libc::SYS_statx,
                        fd,
                        path.as_ptr(),
                        flags,
                        req.data.args[3] as nix::libc::c_int,
                        statx.as_mut_ptr(),
                    )
                };
                let errno = Errno::last();
                if ret == 0 {
                    trace!("ctx": "stat",
                        "path": &path,
                        "pipe": flags & nix::libc::AT_EMPTY_PATH != 0,
                        "sys": syscall_name);
                } else if matches!(errno, Errno::EPERM | Errno::ENOENT | Errno::ENOTDIR) {
                    debug!("ctx": "stat",
                        "path": format!("{}", path.to_string_lossy()),
                        "pipe": flags & nix::libc::AT_EMPTY_PATH != 0,
                        "sys": syscall_name,
                        "err": errno.to_string());
                    return Err(errno);
                } else {
                    info!("ctx": "stat",
                        "path": format!("{}", path.to_string_lossy()),
                        "pipe": flags & nix::libc::AT_EMPTY_PATH != 0,
                        "sys": syscall_name,
                        "err": errno.to_string());
                    return Err(errno);
                }

                // SAFETY: The following block creates an immutable byte slice representing the memory of `statx`.
                // We ensure that the slice covers the entire memory of `statx` using `std::mem::size_of_val`.
                // Since `statx` is a stack variable and we're only borrowing its memory for the duration of the slice,
                // there's no risk of `statx` being deallocated while the slice exists.
                // Additionally, we ensure that the slice is not used outside of its valid lifetime.
                let statx = unsafe {
                    std::slice::from_raw_parts(
                        statx.as_ptr() as *const u8,
                        std::mem::size_of_val(&statx),
                    )
                };
                #[allow(clippy::cast_possible_truncation)]
                proc.write_mem(statx, req.data.args[4] as usize, request)?;
            }
            _ => {
                // "stat" | "fstat" | "lstat" | "newfstatat"
                let mut stat = MaybeUninit::<nix::libc::stat>::uninit();
                Errno::clear();
                // SAFETY: In libc we trust.
                let ret =
                    unsafe { nix::libc::fstatat(fd, path.as_ptr(), stat.as_mut_ptr(), flags) };
                let errno = Errno::last();
                if ret == 0 {
                    trace!("ctx": "stat",
                        "path": format!("{}", path.to_string_lossy()),
                        "pipe": flags & nix::libc::AT_EMPTY_PATH != 0,
                        "sys": syscall_name);
                } else if matches!(errno, Errno::EPERM | Errno::ENOENT | Errno::ENOTDIR) {
                    debug!("ctx": "stat",
                        "path": format!("{}", path.to_string_lossy()),
                        "pipe": flags & nix::libc::AT_EMPTY_PATH != 0,
                        "sys": syscall_name,
                        "err": errno.to_string());
                    return Err(errno);
                } else {
                    info!("ctx": "stat",
                        "path": format!("{}", path.to_string_lossy()),
                        "pipe": flags & nix::libc::AT_EMPTY_PATH != 0,
                        "sys": syscall_name,
                        "err": errno.to_string());
                    return Err(errno);
                }

                // SAFETY: stat returned success, stat struct is properly populated.
                unsafe { stat.assume_init() };
                // SAFETY: The following block creates an immutable byte slice representing the memory of `stat`.
                // We ensure that the slice covers the entire memory of `stat` using `std::mem::size_of_val`.
                // Since `stat` is a stack variable and we're only borrowing its memory for the duration of the slice,
                // there's no risk of `stat` being deallocated while the slice exists.
                // Additionally, we ensure that the slice is not used outside of its valid lifetime.
                let stat = unsafe {
                    std::slice::from_raw_parts(
                        std::ptr::addr_of!(stat) as *const u8,
                        std::mem::size_of_val(&stat),
                    )
                };
                #[allow(clippy::cast_possible_truncation)]
                let addr = req.data.args[if syscall_name == "newfstatat" { 2 } else { 1 }] as usize;
                if addr != 0 {
                    proc.write_mem(stat, addr, request)?;
                }
            }
        }

        // stat system call successfully emulated.
        Ok(request.return_syscall(0))
    })
}

/// A helper function to handle network-related syscalls.
///
/// This function abstracts the common logic involved in handling network syscalls such as `bind`,
/// `connect`, `recvfrom`, and `sendto` in a seccomp-based sandboxing environment. It reduces code
/// duplication across different syscall handler functions.
///
/// # Parameters
///
/// - `request`: Reference to `UNotifyEventRequest`, the user notification request from seccomp.
/// - `arg_addr`: Index of the argument representing the address in the syscall arguments.
/// - `arg_len`: Index of the argument representing the length of the address in the syscall arguments.
/// - `syscall_name`: A string slice holding the name of the syscall being handled, used for error reporting.
/// - `handler`: A closure that handles the processing of the `SockaddrStorage` object and is responsible
///   for logging or any other required operations, specific to the syscall being handled.
///
/// # Safety
///
/// This function contains unsafe blocks due to direct memory access and potential TOCTOU (Time-of-Check
/// Time-of-Use) vulnerabilities, which are acceptable in this context as per requirements. The invoked
/// unsafe blocks are commented with detailed explanations of safety considerations.
///
/// # Errors
///
/// This function returns an error if there is an issue reading memory, converting the address types,
/// or any error returned by the passed `handler` closure. All errors are wrapped in `ScmpNotifResp`
/// and are returned as syscall failures.
///
/// # Returns
///
/// Returns `ScmpNotifResp` indicating the result of the syscall handling:
/// - If successful, it contains a continued syscall.
/// - If an error occurs, it contains a failed syscall with an `EACCES` error code.
#[allow(clippy::cognitive_complexity)]
fn syscall_network_handler<F>(
    request: &UNotifyEventRequest,
    arg_addr: usize,
    arg_len: usize,
    len_ptr: bool,
    syscall_name: &str,
    handler: F,
) -> ScmpNotifResp
where
    F: Fn(
        &RemoteProcess,
        &UNotifyEventRequest,
        &ScmpNotifReq,
        &SockaddrStorage,
        &str,
        Capability,
    ) -> Result<(), Errno>,
{
    syscall_handler!(request, |req: &ScmpNotifReq, proc: &RemoteProcess| {
        let cap = match syscall_name {
            "bind" => Capability::CAP_BIND,
            _ => Capability::CAP_CONNECT,
        };

        let sandbox = request.get_sandbox(false);
        let (check, allow_successful_bind, allow_unsupported_socket_families) = (
            sandbox.enabled(cap),
            sandbox.allow_successful_bind(),
            sandbox.allow_unsupported_socket_families(),
        );
        drop(sandbox);

        // Return immediately if sandboxing is not enabled for current capability,
        if !check {
            // SAFETY: This is unsafe due to vulnerability to TOCTOU,
            // however since the sandboxing for the requested capability
            // is disabled this is safe here.
            return unsafe { Ok(request.continue_syscall()) };
        }

        #[allow(clippy::cast_possible_truncation)]
        let (addr_remote, addr_len) = (
            req.data.args[arg_addr] as usize,
            get_addr_len(req, arg_len, len_ptr, proc, request)?,
        );

        if addr_remote == 0 || addr_len == 0 {
            if matches!(syscall_name, "recvfrom" | "sendto") {
                // SAFETY: Connection mode socket, safe to continue.
                return Ok(unsafe { request.continue_syscall() });
            } else {
                return Err(Errno::EFAULT);
            }
        }

        let addr = get_addr(proc, request, addr_remote, addr_len)?;

        let family = match addr.family() {
            Some(AddressFamily::Unix) => AddressFamily::Unix,
            Some(AddressFamily::Inet) => AddressFamily::Inet,
            Some(AddressFamily::Inet6) => AddressFamily::Inet6,
            _ if allow_unsupported_socket_families => {
                return unsafe { Ok(request.continue_syscall()) }
            }
            _ => return Err(Errno::EAFNOSUPPORT),
        };

        handler(proc, request, req, &addr, syscall_name, cap)?;

        // SAFETY: We're sure we pass a valid fd to OwnedFd here.
        let fd = unsafe { OwnedFd::from_raw_fd(proc.get_fd(req.data.args[0] as RawFd)?) };
        match syscall_name {
            "bind" => handle_bind(&fd, &addr, allow_successful_bind, family, request),
            "connect" => handle_connect(&fd, &addr, request),
            "recvfrom" => handle_recvfrom(&fd, req, proc, request, addr_len, family),
            "sendto" => handle_sendto(&fd, req, proc, request, &addr),
            _ => unreachable!(),
        }
    })
}

fn handle_bind(
    fd: &OwnedFd,
    addr: &SockaddrStorage,
    allow_successful_bind: bool,
    family: AddressFamily,
    request: &UNotifyEventRequest,
) -> Result<ScmpNotifResp, Errno> {
    let mut save_umask = None;
    let mut save_ulock = None;
    if family == AddressFamily::Unix {
        // SAFETY: We must respect process' umask here.
        let req = request.get_request();
        #[allow(clippy::cast_possible_wrap)]
        let pid = Pid::from_raw(req.pid as i32);
        let lock = request.umask_lock();
        let mask = umask(proc_umask(Some(pid))?);
        save_umask = Some(mask);
        save_ulock = Some(lock);
    }
    let result = bind(fd.as_raw_fd(), addr);
    if let Some(mask) = save_umask {
        let _ = umask(mask);
        // We're sure save_ulock is some here.
        #[allow(clippy::disallowed_methods)]
        drop(save_ulock.unwrap());
    }
    if let Err(error) = result {
        return Err(error);
    } else if !allow_successful_bind {
        return Ok(request.return_syscall(0));
    }

    // Handle allow_successful_bind
    let cmd = match family {
        AddressFamily::Unix => {
            let addr = addr.as_unix_addr().ok_or(Errno::EINVAL)?;
            match (addr.path(), addr.as_abstract()) {
                (Some(path), _) => {
                    let path = path.as_os_str().as_bytes();
                    let null = path.iter().position(|&b| b == 0).unwrap_or(path.len());
                    let path = PathBuf::from(OsStr::from_bytes(&path[..null]));
                    Some(format!("allowlist/net/connect+{}", path.display()))
                }
                (_, Some(path)) => {
                    let null = path.iter().position(|&b| b == 0).unwrap_or(path.len());
                    let path = PathBuf::from(OsStr::from_bytes(&path[..null]));
                    Some(format!("allowlist/net/connect+{}", path.display()))
                }
                _ => {
                    // Unnamed UNIX socket
                    None
                }
            }
        }
        AddressFamily::Inet => {
            let addr = addr.as_sockaddr_in().ok_or(Errno::EINVAL)?;
            let mut port = addr.port();
            let addr = IpAddr::V4(Ipv4Addr::from(addr.ip()));
            if port == 0 {
                port = getsockname::<SockaddrStorage>(fd.as_raw_fd())?
                    .as_sockaddr_in()
                    .ok_or(Errno::EINVAL)?
                    .port();
            }
            Some(format!("allowlist/net/connect+{addr}@{port}"))
        }
        AddressFamily::Inet6 => {
            let addr = addr.as_sockaddr_in6().ok_or(Errno::EINVAL)?;
            let mut port = addr.port();
            let addr = IpAddr::V6(addr.ip());
            if port == 0 {
                port = getsockname::<SockaddrStorage>(fd.as_raw_fd())?
                    .as_sockaddr_in6()
                    .ok_or(Errno::EINVAL)?
                    .port();
            }
            Some(format!("allowlist/net/connect+{addr}@{port}"))
        }
        _ => None,
    };
    if let Some(cmd) = cmd {
        let mut sandbox = request.get_sandbox(true);
        let result = sandbox.config(&cmd);
        let pid = request.get_request().pid;
        match result {
            Ok(_) => {
                if let Err(error) = sandbox.build_globsets() {
                    error!("ctx": "config", "pid": pid, "sys": "bind", "cfg": cmd, "err": error.to_string());
                } else {
                    warn!("ctx": "config", "pid": pid, "sys": "bind", "cfg": cmd);
                }
            }
            Err(error) => {
                error!("ctx": "config", "pid": pid, "sys": "bind", "cmd": cmd, "err": error.to_string());
            }
        };
        drop(sandbox);
    }
    Ok(request.return_syscall(0))
}

fn handle_connect(
    fd: &OwnedFd,
    addr: &SockaddrStorage,
    request: &UNotifyEventRequest,
) -> Result<ScmpNotifResp, Errno> {
    connect(fd.as_raw_fd(), addr)?;
    Ok(request.return_syscall(0))
}

fn handle_recvfrom(
    fd: &OwnedFd,
    req: &ScmpNotifReq,
    proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    addr_len: usize,
    addr_family: AddressFamily,
) -> Result<ScmpNotifResp, Errno> {
    // SAFETY: The length argument to the recvfrom call
    // must not be fully trusted, it can be overly large,
    // and allocating a Vector of that capacity may overflow.
    #[allow(clippy::cast_possible_truncation)]
    let len = req.data.args[2] as usize;
    let len = len.min(1000000); // Cap count at 1mio
    let mut buf = vec![0u8; len];
    let (n, addr) = recvfrom::<SockaddrStorage>(fd.as_raw_fd(), &mut buf)?;
    #[allow(clippy::cast_possible_truncation)]
    proc.write_mem(&buf[..n], req.data.args[1] as usize, request)?;
    if let Some(addr) = addr {
        #[allow(clippy::cast_possible_truncation)]
        let addr_remote = req.data.args[4] as usize;
        match addr_family {
            AddressFamily::Inet => {
                // SAFETY: TODO
                let addr = unsafe {
                    any_as_u8_slice(addr.as_sockaddr_in().ok_or(Errno::EINVAL)?.as_ref())
                };
                proc.write_mem(addr, addr_remote, request)?;
            }
            AddressFamily::Inet6 => {
                // SAFETY: TODO
                let addr = unsafe {
                    any_as_u8_slice(addr.as_sockaddr_in6().ok_or(Errno::EINVAL)?.as_ref())
                };
                proc.write_mem(addr, addr_remote, request)?;
            }
            AddressFamily::Unix => {
                // SAFETY: TODO
                let addr =
                    unsafe { any_as_u8_slice(addr.as_unix_addr().ok_or(Errno::EINVAL)?.as_ref()) };
                proc.write_mem(addr, addr_remote, request)?;
            }
            _ => unreachable!(),
        }
        let addr_len = unsafe { any_as_u8_slice(&addr_len) };
        #[allow(clippy::cast_possible_truncation)]
        proc.write_mem(addr_len, req.data.args[5] as usize, request)?;
    }
    #[allow(clippy::cast_possible_wrap)]
    Ok(request.return_syscall(n as i64))
}

fn handle_sendto(
    fd: &OwnedFd,
    req: &ScmpNotifReq,
    proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    addr: &SockaddrStorage,
) -> Result<ScmpNotifResp, Errno> {
    // SAFETY: The length argument to the sendto call
    // must not be fully trusted, it can be overly large,
    // and allocating a Vector of that capacity may overflow.
    #[allow(clippy::cast_possible_truncation)]
    let len = req.data.args[2] as usize;
    let len = len.min(1000000); // Cap count at 1mio
    #[allow(clippy::cast_possible_truncation)]
    let flags = MsgFlags::from_bits_truncate(req.data.args[3] as nix::libc::c_int);
    let mut buf = vec![0u8; len];
    #[allow(clippy::cast_possible_truncation)]
    proc.read_mem(&mut buf, req.data.args[1] as usize, request)?;
    let n = sendto(fd.as_raw_fd(), &buf, addr, flags)?;
    #[allow(clippy::cast_possible_wrap)]
    Ok(request.return_syscall(n as i64))
}

fn get_addr(
    proc: &RemoteProcess,
    request: &UNotifyEventRequest,
    addr_remote: usize,
    addr_len: usize,
) -> Result<SockaddrStorage, Errno> {
    let mut addr = vec![0u8; addr_len];
    proc.read_mem(&mut addr, addr_remote, request)?;

    // SAFETY: Invoking `SockaddrStorage::from_raw` is safe because:
    // 1. The memory location of `sockaddr_ptr` is valid, correctly aligned,
    //    and readable for `addr_len` bytes, preventing segmentation faults.
    // 2. The memory is allocated based on a valid `sockaddr` structure.
    // 3. The `addr_len` is correctly representing the size of the actual `sockaddr` structure.
    // 4. There are no concurrent writes to the memory location while reading.
    match unsafe {
        #[allow(clippy::cast_ptr_alignment)]
        #[allow(clippy::cast_possible_truncation)]
        SockaddrStorage::from_raw(
            addr.as_ptr() as *const nix::libc::sockaddr,
            Some(addr_len as u32),
        )
    } {
        Some(addr) => Ok(addr),
        None => {
            // Invalid socket address of length `addr_len`.
            Err(Errno::EINVAL)
        }
    }
}

fn get_addr_len(
    req: &ScmpNotifReq,
    arg_len: usize,
    len_ptr: bool,
    proc: &RemoteProcess,
    request: &UNotifyEventRequest,
) -> Result<usize, Errno> {
    let addr_len: usize = if !len_ptr {
        req.data.args[arg_len]
            .try_into()
            .map_err(|_| Errno::EINVAL)?
    } else if req.data.args[arg_len] == 0 {
        // Careful here:
        // recvfrom with NULL address length is required for connection-mode sockets.
        // assert!(syscall_name == "recvfrom");
        0
    } else {
        let mut l: nix::libc::socklen_t = 0;
        // SAFETY:
        // We are using the `any_as_u8_mut_slice` function to obtain a mutable byte slice from
        // `l`. We must ensure that the lifetime of this slice does not exceed `l`, and that
        // the memory representation of `l` is correctly interpreted when accessed as bytes.
        // Also, as we're getting a mutable slice, we need to guarantee that no other parts of
        // the program will simultaneously read or write to `l` or the memory area we're
        // interfacing with to prevent data races and undefined behavior.
        let l_slice = unsafe { any_as_u8_mut_slice(&mut l) };
        #[allow(clippy::cast_possible_truncation)]
        proc.read_mem(l_slice, req.data.args[arg_len] as usize, request)?;
        l as usize
    };

    Ok(addr_len)
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn glob_proc() {
        assert!(GLOB_PROC.is_match("/proc/256677/fd/pipe:[2379377]"));
        assert!(GLOB_PROC.is_match("/proc/256677/task/123/fd/pipe:[2379377]"));
        assert!(GLOB_PROC.is_match("/proc/256677/fd/foo (deleted)"));
    }
}

/*
#[cfg(test)]
mod tests {
    use std::{ffi::CStr, time::Duration};

    use nix::sys::signal::{kill, Signal::SIGKILL};

    use super::*;

    #[test]
    fn smoke_test_sleep() {
        fn openat_handler(req: &UNotifyEventRequest) -> libseccomp::ScmpNotifResp {
            let path = req.get_request().data.args[1];
            let remote = RemoteProcess::new(Pid::from_raw(req.request.pid as i32)).unwrap();
            let mut buf = [0u8; 256];
            remote.read_mem(&mut buf, path as usize).unwrap();
            eprintln!("open (read from remote): {:?}", buf);
            let path = CStr::from_bytes_until_nul(&buf).unwrap();
            if !req.is_valid() {
                return req.fail_syscall(libc::EACCES);
            }
            eprintln!("open (path CStr): {:?}", path);
            unsafe { req.continue_syscall() }
        }

        let mut supervisor = Supervisor::new(2).unwrap();
        supervisor.insert_handler(ScmpSyscall::new("openat"), openat_handler);
        let mut cmd = Command::new("/bin/sleep");
        let cmd = cmd.arg("1");
        let (pid, thread_handle, pool) = supervisor.spawn(cmd).unwrap();
        let status = Supervisor::wait(pid, thread_handle, pool).unwrap();
        assert!(status.success());
    }

    #[test]
    fn smoke_test_whoami() {
        fn geteuid_handler(req: &UNotifyEventRequest) -> libseccomp::ScmpNotifResp {
            req.return_syscall(0)
        }

        let mut supervisor = Supervisor::new(2).unwrap();
        supervisor.insert_handler(ScmpSyscall::new("geteuid"), geteuid_handler);
        let mut cmd = Command::new("/usr/bin/whoami");
        let cmd = cmd.stdout(Stdio::piped());
        let (pid, thread_handle, pool) = supervisor.spawn(cmd).unwrap();
        let status = Supervisor::wait(pid, thread_handle, pool).unwrap();
        assert!(status.success());
        let whoami_stdout = child.stdout.as_mut().unwrap();
        let mut buf = String::new();
        whoami_stdout.read_to_string(&mut buf).unwrap();
        assert_eq!(buf.trim(), "root");
    }

    #[test]
    fn test_sleep_blocking_syscall() {
        fn clock_nanosleep_handler(req: &UNotifyEventRequest) -> libseccomp::ScmpNotifResp {
            // sleep for extra 60s
            // Please note that it may bring A LOT OF PROBLEMS if you try using pthread_cancel
            // So here we just use the easy way: check valid in the loop
            let (tx, rx) = std::sync::mpsc::channel();
            let handler = std::thread::spawn(move || {
                for _ in 0..60 {
                    if rx.try_recv().is_ok() {
                        break;
                    }
                    std::thread::sleep(Duration::from_secs(1));
                }
            });
            // while handler is running, check valid in the loop
            loop {
                if !req.is_valid() {
                    // cancel the thread
                    eprintln!("canceling thread as req is invalid now");
                    tx.send(()).unwrap();
                    break;
                }
                std::thread::sleep(Duration::from_millis(100));
            }
            handler.join().unwrap();
            unsafe { req.continue_syscall() }
        }

        let mut supervisor = Supervisor::new(2).unwrap();
        supervisor.insert_handler(ScmpSyscall::new("clock_nanosleep"), clock_nanosleep_handler);
        let mut cmd = Command::new("/bin/sleep");
        let cmd = cmd.arg("120");
        let (pid, thread_handle, pool) = supervisor.spawn(cmd).unwrap();
        std::thread::spawn(move || {
            std::thread::sleep(Duration::from_secs(1));
            // kill the child process
            kill(pid, SIGKILL).unwrap();
        });
        let _ = Supervisor::wait(pid, thread_handle, pool).unwrap();
    }

    #[test]
    fn test_new_fd() {
        fn openat_handler(req: &UNotifyEventRequest) -> libseccomp::ScmpNotifResp {
            let path = req.get_request().data.args[1];
            let remote = RemoteProcess::new(Pid::from_raw(req.request.pid as i32)).unwrap();
            let mut buf = [0u8; 256];
            remote.read_mem(&mut buf, path as usize).unwrap();
            eprintln!("open (read from remote): {:?}", buf);
            let path = CStr::from_bytes_until_nul(&buf).unwrap();
            if !req.is_valid() {
                return req.fail_syscall(libc::EACCES);
            }
            eprintln!("open (path CStr): {:?}", path);
            if path.to_str().unwrap() == "/etc/passwd" {
                // open /etc/resolv.conf instead
                let file = File::open("/etc/resolv.conf").unwrap();
                let fd = file.as_raw_fd();
                let remote_fd = req.add_fd(fd).unwrap();
                req.return_syscall(remote_fd as i64)
            } else {
                unsafe { req.continue_syscall() }
            }
        }

        let mut supervisor = Supervisor::new(2).unwrap();
        supervisor.insert_handler(ScmpSyscall::new("openat"), openat_handler);
        let mut cmd = Command::new("/bin/cat");
        let cmd = cmd.arg("/etc/passwd").stdout(Stdio::piped());
        let (pid, thread_handle, pool) = supervisor.spawn(cmd).unwrap();
        let status = Supervisor::wait(pid, thread_handle, pool).unwrap();
        assert!(status.success());
        let cat_stdout = child.stdout.as_mut().unwrap();
        let mut buf = String::new();
        cat_stdout.read_to_string(&mut buf).unwrap();
        assert!(buf.contains("nameserver"));
    }
}
*/
